package uibk.mtk.draw2d.base;

import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;

import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.math.Interval;

/**
 * Eine 2D-Szene ist im wesentlichen ein rechteckiger Bereich eines kartesischen
 * Koordinatensystens (gegeben durch ein <CODE>CoordinateRect2D</CODE>
 * -Objekt)mit Funktionen zum Umrechnen von Punkten in Weltkoordinaten (Punkte
 * die sich in dem kartesischen Koordinatensystem befinden) und den
 * Ger?tekoordinaten am Asugabeger?t (Bildschirm).
 * 
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */

public class Scene2D {

	/**
	 * Rand der freigelassen werden soll
	 */
	private int edge = 15;

	boolean ensurezerox = false;

	boolean ensurezeroy = false;

	/** Skalierungsfaktor x-Achse */
	double scalex = 1;

	/** Skalierungsfaktor y-Achse */
	double scaley = 1;

	/** Setzt fest ob der Sklaierungsfaktor >=1 sein m?ssen */
	boolean zoomforbidden = true;

	/** Rand der f?r die Beschriftung usw. freigelassen wird */
	private Insets insets = new Insets(0, 0, 0, 0);

	private MathPanel2D mathpanel2d;

	/** aktuell angeforderter Rand */
	private Insets requestedInsets = new Insets(0, 0, 0, 0);

	/**
	 * Die Grenzen des Koordinatensystems, nach Anwenden des Runden,
	 * Skalierens,...
	 */
	CoordinateRect2D limits = new CoordinateRect2D(-1, 1, -1, 1);

	/** Die Originalgrenzen */
	CoordinateRect2D originallimits = new CoordinateRect2D(-1, 1, -1, 1);

	/**
	 * diese Variable bestimmt ob die Grenzen des Koordinatensystems so
	 * angepasst werden sollen dass eine Einheit in vertikaler Richtung und eine
	 * Einheit in Horizontaler Richutng immer gleich vielen Pixeln entsprechen.
	 * Dabei werden die Grenzen immer vergr?ssert und nie verkleinert. Falls
	 * auch das Runden der Grenzen aktiviert ist wird zuerst gerundet und dann
	 * die Achsen angegelichen
	 *  
	 */
	boolean equalize = false;

	/** gibt an ob die x-Achse gerundet werden soll */
	boolean roundx = true;

	/** gibt an ob die y-Achse gerundet werden soll */
	boolean roundy = true;

	/** gibt an um wieviel die Achsen gerundet werden sollen */
	double roundingpercent = 20;

	/** Gibt den Skalierungsfaktor f?r die x-Achse zur?ck */
	public double getScaleFactorX() {
		return scalex;
	}

	/** Gibt den Skalierungsfaktor f?r die y-Achse zur?ck */
	public double getScaleFactorY() {
		return scaley;
	}

	/** Setzt den Skalierungsfaktor f?r die x-Achse */
	public void setScaleFactorX(double scale) {
		if (scale <= 0)
			throw new IllegalArgumentException(""); //$NON-NLS-1$
		if (scale >= 1 || (scale < 1 && !zoomforbidden))
			scalex = scale;

	}

	/** Setzt den Skalierungsfaktor f?r die y-Achse */
	public void setScaleFactorY(double scale) {
		if (scale <= 0)
			throw new IllegalArgumentException(""); //$NON-NLS-1$
		if (scale >= 1 || (scale < 1 && !zoomforbidden))
			scaley = scale;

	}

	/**
	 * Standardkonstruktor
	 * 
	 * Erzeugen eines Koordinatensystem mit den Standardwerten -1,1,-1,1
	 */
	public Scene2D(MathPanel2D mathpanel2d) {
		this(mathpanel2d, new CoordinateRect2D());

	}

	/** Setzt die Grenzen des Koordinatensystems auf -1,1,-1,1 zur?ck */
	public void reset() {
		this.setLimits(-1, 1, -1, 1);
	}

	public Rectangle getCoordinatePixelRectangle() {
		Rectangle rect = new Rectangle();
		rect.x = xToPixel(limits.xmin);
		rect.y = yToPixel(limits.ymax);
		rect.width = xToPixel(limits.xmax) - xToPixel(limits.xmin);
		rect.height = yToPixel(limits.ymin) - yToPixel(limits.ymax);

		return rect;

	}

	/**
	 * Erzeugen eines Koordinatensystem mit den ?bergebenen Grenzen
	 * 
	 * @param mathpanel
	 *            MathPanel2D in dem sich die Szene befindet
	 * 
	 * @param rect
	 *            Grenzen
	 */
	public Scene2D(MathPanel2D mathpanel, CoordinateRect2D rect) {
		this.mathpanel2d = mathpanel;
		setLimits(rect.xmin, rect.xmax, rect.ymin, rect.ymax);
	}

	/**
	 * Gibt <CODE>true</CODE> zur?ck, falls die Skala zu klein ist und
	 * verschiedene Pixel auf gleiche Wert abgebildet werden
	 * 
	 *  
	 */
	boolean isScaleTooSmall() {
		if (pixelToX(0) - pixelToX(1) == 0 || pixelToY(0) - pixelToY(1) == 0) {
			return true;
		}
		return false;

	}

	/** Setzt das x-Intervall f?r die Szene */
	public void setXInterval(Interval interval) {
		setLimits(interval.a, interval.b, limits.ymin, limits.ymax);

	}

	/** Setzt das y-Intervall f?r die Szene */
	public void setYInterval(Interval interval) {
		setLimits(limits.xmin, limits.xmax, interval.a, interval.b);
	}

	/** aktiviert das Runden der Grenzen */
	public void enableRoundX(boolean b) {
		roundx = b;
	}

	/** aktiviert das Runden der Grenzen */
	public void enableRound(boolean b) {
		roundx = b;
		roundy = b;
	}

	/** aktiviert das Runden der Grenzen */
	public void enableRoundY(boolean b) {
		roundy = b;
	}

	/** aktiviert das Runden der Grenzen */
	public boolean isEnabledRoundX() {
		return roundx;
	}

	/** aktiviert das Runden der Grenzen */
	public boolean isEnabledRoundY() {
		return roundy;
	}

	/** Setzt um wieviel Prozent bei aktivertem Runden gerundet wird */
	public void setRoundingPercent(double percent) {
		if (percent < 0)
			throw new IllegalArgumentException();
		this.roundingpercent = percent;
	}

	/**
	 * Diese Funktion vergr?ssert die Grenzen des Koordinatensystems so, dass 1
	 * Pixel in horizontaler (d.h. Richtung der x-Achse) dem gleichen Abstand in
	 * Koordinaten entspricht wie ein Pixel in vertikaler Richtung. D.h. nicht
	 * unbeding das die Skalierung gleich sein muss, da Breite und H?he des
	 * Fensters nicht gleich sein m?ssen. Dies ist vorallem dann n?tzlich, falls
	 * man sicher stellen will, das die Objekte unverzerrt, gezeichnet
	 * werden,z.B. ein Kreis immer als Kreis gezeichnet wird.
	 */
	private void equalizeCoords() {

		int width = mathpanel2d.getWidth() - 2 * edge - insets.left
				- insets.right;
		int height = mathpanel2d.getHeight() - 2 * edge - insets.top
				- insets.bottom;
		double newxmin = limits.xmin, newxmax = limits.xmax;
		double newymin = limits.ymin, newymax = limits.ymax;
		double scaleX = limits.getXRange() / width;
		double scaleY = limits.getYRange() / height;
		double adjustY, adjustX;

		int area = getCoordinates().getArea();

		if (scaleX < scaleY) {
			adjustX = scaleY * width - limits.getXRange();
			if (area == CoordinateRect2D.AREA_UPPERRIGHT
					|| area == CoordinateRect2D.AREA_RIGHT
					|| area == CoordinateRect2D.AREA_LOWERRIGHT) {
				newxmax = limits.xmax + adjustX;

			}

			if (area == CoordinateRect2D.AREA_UPPER
					|| area == CoordinateRect2D.AREA_ALL
					|| area == CoordinateRect2D.AREA_LOWER) {
				newxmin = limits.xmin - adjustX / 2;
				newxmax = limits.xmax + adjustX / 2;
			}

			if (area == CoordinateRect2D.AREA_UPPERLEFT
					|| area == CoordinateRect2D.AREA_LEFT
					|| area == CoordinateRect2D.AREA_LOWERLEFT) {
				newxmin = limits.xmin - adjustX;

			}

		} else {
			adjustY = scaleX * height - limits.getYRange();
			if (area == CoordinateRect2D.AREA_UPPERRIGHT
					|| area == CoordinateRect2D.AREA_UPPER
					|| area == CoordinateRect2D.AREA_UPPERLEFT) {
				newymax = limits.ymax + adjustY;
			}

			if (area == CoordinateRect2D.AREA_LEFT
					|| area == CoordinateRect2D.AREA_ALL
					|| area == CoordinateRect2D.AREA_RIGHT) {
				newymax = limits.ymax + adjustY / 2;
				newymin = limits.ymin - adjustY / 2;

			}

			if (area == CoordinateRect2D.AREA_LOWERRIGHT
					|| area == CoordinateRect2D.AREA_LOWER
					|| area == CoordinateRect2D.AREA_LOWERLEFT) {
				newymin = limits.ymin - adjustY;
			}

		}

		checkAndSetLimits(newxmin, newxmax, newymin, newymax, limits);

	}

	/** Gibt die Breite des Randes zur?ck */
	public int getEdge() {
		return edge;
	}

	public Insets getInsets() {
		return insets;
	}

	/**
	 * Gibt die maximale x-Koordinate zur?ck
	 */
	public double getXmax() {
		return limits.xmax;
	}

	/**
	 * Gibt die minimale x-Koordinate zur?ck
	 */
	public double getXmin() {
		return limits.xmin;
	}

	/**
	 * gibt zur?ck wie gro? die Streche ist, die einem Pixel entspricht
	 */

	public double getXPixelScale() {
		return limits.getXRange()
				/ (mathpanel2d.getWidth() - 2 * edge - insets.left - insets.right);
	}

	/**
	 * Gibt die maximale y-Koordinate zur?ck
	 */
	public double getYmax() {
		return limits.ymax;
	}

	/**
	 * Gibt die minimale y-Koordinate zur?ck
	 */
	public double getYmin() {
		return limits.ymin;
	}

	public double getYPixelScale() {
		return limits.getYRange()
				/ (mathpanel2d.getHeight() - 2 * edge - insets.top - insets.bottom);
	}

	/**
	 * Wandelt eine x-Koordinate (Ger?tekoordinate) in die entsprechende
	 * Weltkoordinate um
	 */
	public double pixelToX(int x) {

		return limits.xmin
				+ ((x - edge - insets.left) * (limits.xmax - limits.xmin))
				/ (mathpanel2d.getWidth() - 2 * edge - insets.left
						- insets.right - 1);
	}

	/**
	 * Wandelt eine y-Koordinate (Ger?tekoordinate) in die entsprechende
	 * Weltkoordinate um
	 */
	public double pixelToY(int y) {
		return limits.ymax
				- ((y - edge - insets.top) * (limits.ymax - limits.ymin))
				/ (mathpanel2d.getHeight() - 2 * edge - insets.top
						- insets.bottom - 1);
	}

	/** Runden der Koordinaten */
	private void roundCoords() {

		int adjust = 0;

		if (roundx)
			adjust = CoordinateRect2D.ROUND_X;
		if (roundy)
			adjust = CoordinateRect2D.ROUND_Y;
		if (roundx && roundy)
			adjust = CoordinateRect2D.ROUND_XY;

		limits.roundCoords(roundingpercent, adjust);

	}

	/** Setzt den Rand, der frei bleiben soll */
	public void setEdge(int edge) {
		if (edge < 0)
			throw new IllegalArgumentException(""); //$NON-NLS-1$
		this.edge = edge;

	}

	/**
	 * Setzt die Grenzen des Koordinatensystems. Falls aktiviert wird
	 * <ul>
	 * <li>sichergestellt, dass sich die 0 auf der x-Achse bzw. y-Achse
	 * befindet</li>
	 * <li>gerundet</li>
	 * <li>gleiche Skalierung auf beiden Achsen hergestellt</li>
	 * </ul>
	 */
	public void setLimits(CoordinateRect2D coordinates) {
		setLimits(coordinates.xmin, coordinates.xmax, coordinates.ymin,
				coordinates.ymax);
	}

	/**
	 * Setzt die Grenzen des Koordinatensystems. Falls aktiviert wird
	 * <ul>
	 * <li>sichergestellt, dass sich die 0 auf der x-Achse bzw. y-Achse
	 * befindet</li>
	 * <li>gerundet</li>
	 * <li>gleiche Skalierung auf beiden Achsen hergestellt</li>
	 * </ul>
	 */
	public void setLimits(double xmin, double xmax, double ymin, double ymax) {

		checkAndSetLimits(xmin, xmax, ymin, ymax, originallimits);
		//Skalierung auf 1 zur?ck
		this.scalex = 1;
		this.scaley = 1;
		this.prepareScene();

	}

	/** Anwenden der aktuellen Skalierung, Rundung usw. */
	public void prepareScene() {
		limits = (CoordinateRect2D) originallimits.clone();

		if (ensurezerox)
			ensureZeroX();
		if (ensurezeroy)
			ensureZeroY();

		calcRescaledLimits();

		if (roundx || roundy)
			roundCoords();

		if (equalize) {
			equalizeCoords();
		}

	}

	public void ensureZero(boolean enable) {
		ensurezerox = enable;
		ensurezeroy = enable;
	}

	/**
	 * Stellt sicher, dass sich die druch <CODE>rect</CODE> geg. Grenzen in
	 * der Szene befinden (falls sie sich nicht darin befinden wird das
	 * Koordinatenrechteck vergr??ert)
	 *  
	 */
	public void ensureLimits(CoordinateRect2D rect) {

		CoordinateRect2D combined = limits.combine(rect);
		setLimits(combined.xmin, combined.xmax, combined.ymin, combined.ymax);

	}

	private void ensureZeroY() {

		if (limits.ymin > 0) {
			checkAndSetLimits(limits.xmin, limits.xmax, 0, limits.ymax, limits);
			return;
		}
		if (limits.ymax < 0) {
			checkAndSetLimits(limits.xmin, limits.xmax, limits.ymin, 0, limits);
			return;
		}

	}

	private void ensureZeroX() {

		if (limits.xmin > 0) {
			checkAndSetLimits(0, limits.xmax, limits.ymin, limits.ymax, limits);
			return;
		}
		if (limits.xmax < 0) {
			checkAndSetLimits(limits.xmin, 0, limits.ymin, limits.ymax, limits);
			return;
		}

	}

	/**
	 * Legt fest, dass beim Setzen der Koordinaten die Grenzen so ge?ndert
	 * werden, dass die x- und y- "gleich" skaliert sind
	 *  
	 */
	public void axesequal(boolean b) {

		equalize = b;

	}

	/** Gibt zur?ck ob die x und y-Achse gleich skaliert werden */
	public boolean isAxesEqualEnabled() {
		return equalize;

	}

	/**
	 * Setzt die Grenzen des Koordinatensystems. Leere Koordinatensysteme (d.h.
	 * xmin==xmax oder ymin==ymax werden abgefangen)
	 * 
	 * @param xmin
	 *            minimale x-Koordinate
	 * @param xmax
	 *            maximale y-Koordinate
	 * @param ymin
	 *            minimale y-Koordinate
	 * @param ymax
	 *            maximale y-Koordinate
	 */
	private void checkAndSetLimits(double xmin, double xmax, double ymin,
			double ymax, CoordinateRect2D rect) {

		if (xmin == xmax) {
			rect.xmin = xmin - 1;
			rect.xmax = xmin + 1;
		} else {
			rect.xmin = xmin;
			rect.xmax = xmax;
		}

		if (ymin == ymax) {
			rect.ymin = ymin - 1;
			rect.ymax = ymin + 1;
		} else {
			rect.ymin = ymin;
			rect.ymax = ymax;
		}

	}

	/**
	 * Legt fest, dass beim Setzen der Koordinaten das x-Intervall gegebenfalls
	 * so angepasst wird das sich die 0 im Intervall befindet.
	 *  
	 */
	public void ensureZeroX(boolean ensure) {
		this.ensurezerox = ensure;
	}

	/**
	 * Legt fest, dass beim Setzen der Koordinaten das y-Intervall gegebenfalls
	 * so angepasst wird das sich die 0 im Intervall befindet.
	 *  
	 */
	public void ensureZeroY(boolean ensure) {
		this.ensurezeroy = ensure;
	}

	/** Setzt den <CODE>MathPanel2D</CODE> in dem sich die Szene befindet */
	public void setPanel(MathPanel2D mathpanel2d) {
		this.mathpanel2d = mathpanel2d;
	}

	/** Gibt die Grenzen der Szene in einem String zur?ck */
	public String toString() {
		return limits.toString();
	}

	/**
	 * Wandelt eine x-Koordinate (Weltkoordinate) in die entsprechende
	 * Ger?tekoordinate um
	 */
	public int xToPixel(double x) {

		int xInt = edge
				+ insets.left
				+ (int) Math.round((x - limits.xmin)
						/ (limits.xmax - limits.xmin)
						* (mathpanel2d.getWidth() - 2 * edge - insets.left
								- insets.right - 1));

		return xInt;
	}

	/**
	 * Wandelt eine y-Koordinate (Weltkoordinate) in die entsprechende
	 * Ger?tekoordinate um
	 */
	public int yToPixel(double y) {

		int yInt = edge
				+ insets.top
				+ (int) Math.round((limits.ymax - y)
						/ (limits.ymax - limits.ymin)
						* (mathpanel2d.getHeight() - 2 * edge - insets.top
								- insets.bottom - 1));
		return yInt;
	}

	/**
	 * Berechnet die Ger?tekoordinaten f?r den Punkt <CODE>a</CODE>
	 * 
	 * @param a
	 *            Punkt in Weltkoordinaten
	 * 
	 * @return Ger?tekoordinaten des Punktes
	 */
	public Point project(Punkt2D a) {
		return (new Point(xToPixel(a.x), yToPixel(a.y)));

	}

	/**
	 * Berechnet die Ger?tekoordinaten f?r den Punkt <CODE>a</CODE>
	 * 
	 * @param a
	 *            Punkt in Weltkoordinaten
	 * 
	 * @return Ger?tekoordinaten des Punktes
	 */
	public Punkt2D projectdouble(Punkt2D a) {
		return (new Punkt2D(xToPixel(a.x), yToPixel(a.y)));

	}

	/** Gibt eine Kopie der Grenzen des Koordinatensystems zur?ck */
	public CoordinateRect2D getCoordinates() {
		return (CoordinateRect2D) limits.clone();
	}

	/** Gibt die L?nge des x-Intervalls des Koordinatensystems zur?ck */
	public double getXRange() {
		return limits.getXRange();
	}

	/** Gibt die L?nge des y-Intervalls des Koordinatensystems zur?ck */
	public double getYRange() {
		return limits.getYRange();
	}

	boolean processRequests() {

		if (requestedInsets.left != 0 || requestedInsets.right != 0
				|| requestedInsets.top != 0 || requestedInsets.bottom != 0) {

			insets.left = insets.left + requestedInsets.left;
			insets.right = insets.right + requestedInsets.right;
			insets.top = insets.top + requestedInsets.top;
			insets.bottom = insets.bottom + requestedInsets.bottom;

			requestedInsets.left = 0;
			requestedInsets.right = 0;
			requestedInsets.top = 0;
			requestedInsets.bottom = 0;

			return true;
		}
		return false;

	}

	public void requestSpaceBottom(int pixel) {
		if (pixel >= mathpanel2d.getHeight() - edge
				&& Math.abs(pixel - mathpanel2d.getHeight() + edge) > requestedInsets.bottom) {
			requestedInsets.bottom = Math.abs(pixel - mathpanel2d.getHeight()
					+ edge);
		}
	}

	public void requestSpaceLeft(int pixel) {
		if (pixel <= edge && Math.abs(edge - pixel) > requestedInsets.left) {
			requestedInsets.left = Math.abs(edge - pixel);
		}
	}

	public void requestSpaceRight(int pixel) {
		if (pixel >= mathpanel2d.getWidth() - edge
				&& Math.abs(pixel - mathpanel2d.getWidth() + edge) > requestedInsets.right) {
			requestedInsets.right = Math.abs(pixel - mathpanel2d.getWidth()
					+ edge);
		}
	}

	public void requestSpaceTop(int pixel) {

		if (pixel <= edge && Math.abs(edge - pixel) > requestedInsets.top) {
			requestedInsets.top = Math.abs(edge - pixel);
		}
	}

	void resetInsets() {
		insets.bottom = 0;
		insets.top = 0;
		insets.left = 0;
		insets.right = 0;
		requestedInsets.bottom = 0;
		requestedInsets.top = 0;
		requestedInsets.left = 0;
		requestedInsets.right = 0;

	}

	/** umskalierte Grenzen berechnen */
	private void calcRescaledLimits() {

		calcRescaledLimitsX(scalex);
		calcRescaledLimitsY(scaley);

	}

	/** umskalierte Grenzen x-Achse berechnen */
	private void calcRescaledLimitsX(double factor) {

		double xmin = limits.xmin;
		double xmax = limits.xmax;
		double ymin = limits.ymin;
		double ymax = limits.ymax;
		double newxmin = xmin, newxmax = xmax, newymin = ymin, newymax = ymax;

		int area = limits.getArea();

		if (area == CoordinateRect2D.AREA_UPPERRIGHT
				|| area == CoordinateRect2D.AREA_RIGHT
				|| area == CoordinateRect2D.AREA_LOWERRIGHT) {
			newxmax = xmin + getXRange() * factor;
		}

		if (area == CoordinateRect2D.AREA_UPPER
				|| area == CoordinateRect2D.AREA_ALL
				|| area == CoordinateRect2D.AREA_LOWER) {
			newxmax = limits.xmax * factor;
			newxmin = limits.xmin * factor;
		}

		if (area == CoordinateRect2D.AREA_UPPERLEFT
				|| area == CoordinateRect2D.AREA_LEFT
				|| area == CoordinateRect2D.AREA_LOWERLEFT) {
			newxmin = xmax - getXRange() * factor;
		}

		limits.xmin = newxmin;
		limits.xmax = newxmax;
		limits.ymin = newymin;
		limits.ymax = newymax;

	}

	/** umskalierte Grenzen y-Achse berechnen */
	private void calcRescaledLimitsY(double factor) {

		double xmin = limits.xmin;
		double xmax = limits.xmax;
		double ymin = limits.ymin;
		double ymax = limits.ymax;
		double newxmin = xmin, newxmax = xmax, newymin = ymin, newymax = ymax;

		int area = limits.getArea();

		if (area == CoordinateRect2D.AREA_UPPERRIGHT
				|| area == CoordinateRect2D.AREA_UPPER
				|| area == CoordinateRect2D.AREA_UPPERLEFT) {
			newymax = ymin + getYRange() * factor;
		}

		if (area == CoordinateRect2D.AREA_LEFT
				|| area == CoordinateRect2D.AREA_ALL
				|| area == CoordinateRect2D.AREA_RIGHT) {
			newymax = ymax * factor;
			newymin = ymin * factor;

		}

		if (area == CoordinateRect2D.AREA_LOWERRIGHT
				|| area == CoordinateRect2D.AREA_LOWER
				|| area == CoordinateRect2D.AREA_LOWERLEFT) {
			newymin = ymax - getYRange() * factor;
		}
		limits.xmin = newxmin;
		limits.xmax = newxmax;
		limits.ymin = newymin;
		limits.ymax = newymax;

	}

	/** Gibt die Grenzen ohne Skalierung,Rundung usw. zur?ck */
	public CoordinateRect2D getOriginalLimits() {
		return (CoordinateRect2D) originallimits.clone();
	}

}