package uibk.mtk.draw2d.util;

import java.util.Iterator;
import java.util.Vector;

import uibk.mtk.draw2d.objects.Graph2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.geom3d.Punkt3D;
import uibk.mtk.lang.Messages;
import uibk.mtk.math.SingularityChecker;
import uibk.mtk.math.functions.Function1D;
import uibk.mtk.math.functions.ParametricCurve2D;
import uibk.mtk.math.functions.ParametricCurve3D;
import uibk.mtk.util.IntervalIterator;

/**
 * Utilityklasse zum Berechnen von Graphen.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public final class GraphUtil {

	private static final String BUNDLE_NAME = "uibk.mtk.draw2d.util.messages";//$NON-NLS-1$

	private GraphUtil() {
		super();

	}

	/**
	 * Berechnet den Graph einer 2d-parametrischen Kurven mit <CODE>n</CODE>
	 * Punkten den Parameterwerten <CODE>paramstart</CODE> und <CODE>paramend
	 * </CODE> und gibt sie in einem Vektor bestehend aus <CODE>Punkt2D</CODE>
	 * zur?ck zur?ck Falls ?berlauf/Division durch 0 auftritt wird Exception
	 * geworfen.
	 */
	public static Vector compute(ParametricCurve2D curve2d, double paramstart,
			double paramend, int n) throws Exception {

		Vector points = new Vector();

		IntervalIterator it = new IntervalIterator(paramstart, paramend, n);

		//Berechnen der Punkte
		while (it.hasNext()) {

			double t = it.nextdouble();

			Punkt2D p = curve2d.calcPoint(t);
			if (!p.isReal()) {
				throw new Exception(Messages.getString(BUNDLE_NAME,
						"GraphUtil.0"));} //$NON-NLS-1$

			points.add(p);

		}

		return points;

	}

	/**
	 * Berechnet den Graph einer 3d-parametrischen Kurven mit <CODE>n</CODE>
	 * Punkten zw. den Parameterwerten <CODE>paramstart</CODE> und <CODE>
	 * paramend</CODE> und gibt sie in einem Vektor bestehend aus <CODE>
	 * Punkt2D</CODE> zur?ck zur?ck Falls ?berlauf/Division durch 0 auftritt
	 * wird Exception geworfen.
	 */
	public static Vector compute(ParametricCurve3D curve3d, double paramstart,
			double paramend, int n) throws Exception {

		Vector points = new Vector();

		IntervalIterator it = new IntervalIterator(paramstart, paramend, n);

		//Berechnen der Punkte
		while (it.hasNext()) {

			double t = it.nextdouble();

			Punkt3D p = curve3d.calcPoint(t);
			if (!p.isReal()) {
				throw new Exception(Messages.getString(BUNDLE_NAME,
						"GraphUtil.0"));} //$NON-NLS-1$

			points.add(p);

		}
		return points;

	}

	/**
	 * Berechnet <CODE>n</CODE> Punkte auf dem Graphen der Funktion zw. den
	 * Funktionswerten <CODE>xstart</CODE> und <CODE>xend</CODE> und gibt
	 * sie in einem Vektor bestehend aus <CODE>Punkt2D</CODE> zur?ck zur?ck.
	 * Falls ?berlauf/Division durch 0 auftritt wird Exception geworfen.
	 */
	public static Vector compute(Function1D function, double xstart,
			double xend, int n) throws Exception {

		Vector points = new Vector();

		IntervalIterator it = new IntervalIterator(xstart, xend, n);

		//Berechnen der Punkte
		while (it.hasNext()) {

			double t = it.nextdouble();

			double y = function.getValue(t);

			if (Double.isNaN(y) || Double.isInfinite(y)) {
				throw new Exception(Messages.getString(BUNDLE_NAME,
						"GraphUtil.0"));} //$NON-NLS-1$

			points.add(new Punkt2D(t, y));

		}
		return points;

	}

	/**
	 * Berechnet <CODE>n</CODE> Punkte auf dem Graphen der Funktion zw. den
	 * Funktionswerten <CODE>xstart</CODE> und <CODE>xend</CODE> und gibt
	 * sie in einem Vektor bestehend aus <CODE>Punkt2D</CODE> zur?ck zur?ck.
	 *  
	 */
	public static Vector computewithsing(Function1D function, double xstart,
			double xend, int n) {

		double min = Double.MAX_VALUE;
		double max = Double.MIN_VALUE;
		//Vector v = new Vector();
		IntervalIterator it1 = new IntervalIterator(xstart, xend, n);
		while (it1.hasNext()) {
			double t = it1.nextdouble();
			double y = function.getValue(t);
			if (y > max)
				max = y;
			if (y < min)
				min = y;
			//v.add(new Punkt2D(t, y));
		}

		Vector points = new Vector();
		SingularityChecker sing = new SingularityChecker(min, max);
		IntervalIterator it = new IntervalIterator(xstart, xend, n);
		double last = 0;
		//Berechnen der Punkte
		while (it.hasNext()) {

			double t = it.nextdouble();
			if (it.getPos() > 1) {
				boolean s = sing.hasSingularity(function, last, t);

				if (s) {
					points.add(null);
				} else {
					double y = function.getValue(t);

					points.add(new Punkt2D(t, y));

				}

			}
			last = t;

		}
		return points;

	}

	public static Vector convertPointsToGraphs(Vector v) {
		Vector graphs = new Vector();

		Iterator it = v.iterator();

		Graph2D g = new Graph2D();
		while (it.hasNext()) {
			Punkt2D p = (Punkt2D) it.next();
			if (p == null) {
				if (g.getNumberPoints() > 0) {
					graphs.add(g);
				}
				g = new Graph2D();

			}

			g.addPoint(p);

		}
		if (g.getNumberPoints() > 0) {
			graphs.add(g);
		}

		return graphs;
	}

}