package uibk.mtk.draw2d.util;

import java.awt.Point;
import java.awt.Rectangle;

import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;

/**
 * Klippen von Linien in der Ebene an einem Rechteck nach dem Cohen-Sutherland
 * Algorithmus.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 *  
 */
public final class LineClipping2D {

	private LineClipping2D() {
	}

	private static final byte EMPTY = 0x0;

	/** Punkt liegt links vom Rechteck */
	private static final byte LEFT = 0x1;

	/** Punkt liegt rechts vom Rechteck */
	private static final byte RIGHT = 0x2;

	/** Punkt liegt ?ber dem Rechteck */
	private static final byte BOTTOM = 0x4;

	/** Punkt liegt unter dem Rechteck */
	private static final byte TOP = 0x8;

	/**
	 * gibt ein byte zur?ck das mit Hilfe der Konstanen die Lage des Punktes (
	 * <CODE>px</CODE>,<CODE>px</CODE>) im Bezug zum Rechteck <CODE>rect
	 * </CODE> angibt
	 */
	private static byte regioncode(long px, long py, Rectangle rect) {
		byte code = EMPTY;
		int xmin = rect.x;
		int xmax = rect.x + rect.width;
		int ymin = rect.y;
		int ymax = rect.y + rect.height;

		if (px < xmin)
			code = LEFT;
		if (px > xmax)
			code = RIGHT;
		if (py > ymax)
			code |= TOP;
		if (py < ymin)
			code |= BOTTOM;
		return code;
	}

	/**
	 * gibt ein byte zur?ck das mit Hilfe der Konstanen die Lage des Punktes
	 * <CODE>p</CODE> im Bezug zum Rechteck <CODE>rect</CODE> angibt
	 */
	private static byte regioncode(Punkt2D p, CoordinateRect2D rect) {

		byte code = EMPTY;

		if (p.x < rect.xmin)
			code = LEFT;
		if (p.x > rect.xmax)
			code = RIGHT;
		if (p.y > rect.ymax)
			code |= TOP;
		if (p.y < rect.ymin)
			code |= BOTTOM;
		return code;
	}

	/**
	 * Klippen der einer durch Anfans- und Endpunkt gegebenen Strecke an einem
	 * Rechteck
	 * 
	 * @param a
	 *            Anfangspunkt der Strecke
	 * @param b
	 *            Endpunkt der Strecke
	 * @param p
	 *            geclippter Anfangspunkt der Strecke (Funktion setzt die
	 *            Koordinaten)
	 * @param q
	 *            geclippter Endpunkt der Strecke (Funktion setzt die
	 *            Koordinaten)
	 * @param rect
	 *            Rechteck an dem geclippt werden soll
	 * @return gibt false zur?ck falls die Gerade komplett au?erhalb des
	 *         Rechtecks liegt, ansonsten true
	 */
	public static boolean cohensutherland(Point a, Point b, Point p, Point q,
			Rectangle rect) {

		int count = 0;

		if (Double.isInfinite(a.x) || Double.isNaN(a.y)
				|| Double.isInfinite(b.x) || Double.isNaN(b.y))
			return false;

		int xmin = rect.x;
		int xmax = rect.x + rect.width;
		int ymin = rect.y;
		int ymax = rect.y + rect.height;

		long px = a.x;
		long py = a.y;
		long qx = b.x;
		long qy = b.y;

		byte c1 = regioncode(px, py, rect);
		byte c2 = regioncode(qx, qy, rect);
		byte c;
		//endliche Steigung
		boolean finiteslope = px != qx ? true : false;

		//zu berechnender Schnittpunkt
		//Point a = new Point();
		long ax = 0;
		long ay = 0;

		//Steigung
		double slope = 0;
		if (finiteslope)
			slope = (qy - py) / (double) (qx - px);

		while ((c1 != EMPTY) || (c2 != EMPTY)) { //solange wir drau?en liegen

			//beide auf der selben Siete au?erhalb
			if ((c1 & c2) != EMPTY)
				return false;

			c = (c1 == EMPTY) ? c2 : c1; //c ist au?erhalb berechne Schnitt
			// mit einer Kante

			//falls c links -> scheiden mit linker Kante
			if ((c & LEFT) != EMPTY) {
				ax = xmin;
				ay = (long) (py + (ax - px) * slope);
			}

			//falls c rechts -> scheiden mit rechter Kante
			if ((c & RIGHT) != EMPTY) {
				ax = xmax;
				ay = (long) (py + (ax - px) * slope);
			}

			//falls c rechts -> scheiden mit rechter Kante
			if ((c & BOTTOM) != EMPTY) {
				ay = ymin;
				if (finiteslope)
					ax = (long) (px + (ay - py) / slope);
				else
					ax = px;
			}

			//falls c oben -> scheiden mit rechter Kante
			if ((c & TOP) != EMPTY) {
				ay = ymax;
				if (finiteslope)
					ax = (long) (px + (ay - py) / slope);
				else
					ax = px;
			}

			if (c == c1) {
				px = ax;
				py = ay;
				c1 = regioncode(px, py, rect);
			} else {
				qx = ax;
				qy = ay;
				c2 = regioncode(qx, qy, rect);
			}

			count++;

		}

		p.x = (int) px;
		p.y = (int) py;
		q.x = (int) qx;
		q.y = (int) qy;

		return true;
	}

	/**
	 * Klippen der einer durch Anfans- und Endpunkt gegebenen Strecke an einem
	 * Rechteck f?r Punkte in doppelter Genauigkeit
	 * 
	 * @param p
	 *            Anfangspunkt der Strecke
	 * @param q
	 *            Endpunkt der Strecke
	 * @param u
	 *            geclippter Anfangspunkt der Strecke (Funktion setzt die
	 *            Koordinaten)
	 * @param v
	 *            geclippter Endpunkt der Strecke (Funktion setzt die
	 *            Koordinaten)
	 * @param rect
	 *            Rechteck an dem geclippt werden soll
	 * @return gibt false zur?ck falls die Gerade komplett au?erhalb des
	 *         Rechtecks liegt, ansonsten true
	 */
	public static boolean cohensutherland(Punkt2D p, Punkt2D q, Punkt2D u,
			Punkt2D v, CoordinateRect2D rect) {

		if (Double.isInfinite(u.x) || Double.isNaN(u.y)
				|| Double.isInfinite(v.x) || Double.isNaN(v.x))
			return false;

		double xmin = rect.xmin;
		double xmax = rect.xmax;
		double ymin = rect.ymin;
		double ymax = rect.ymax;

		u.x = p.x;
		u.y = p.y;
		v.x = q.x;
		v.y = q.y;
		byte c1 = regioncode(u, rect);
		byte c2 = regioncode(v, rect);
		byte c;
		//endliche Steigung
		boolean finiteslope = u.x != v.x ? true : false;

		//zu berechnender Schnittpunkt
		Punkt2D a = new Punkt2D();

		//Steigung
		double slope = 0;
		if (finiteslope)
			slope = (v.y - u.y) / (v.x - u.x);

		while ((c1 != EMPTY) || (c2 != EMPTY)) { //solange wir drau?en liegen

			//beide auf der selben Siete au?erhalb
			if ((c1 & c2) != EMPTY)
				return false;

			//c ist au?erhalb berechne Schnitt mit einer Kante
			c = (c1 == EMPTY) ? c2 : c1;

			//falls c links -> scheiden mit linker Kante
			if ((c & LEFT) != EMPTY) {
				a.x = xmin;
				a.y = (u.y + (a.x - u.x) * slope);
			}

			//falls c rechts -> scheiden mit rechter Kante
			if ((c & RIGHT) != EMPTY) {
				a.x = xmax;
				a.y = (u.y + (a.x - u.x) * slope);
			}

			//falls c rechts -> scheiden mit rechter Kante
			if ((c & BOTTOM) != EMPTY) {
				a.y = ymin;
				if (finiteslope)
					a.x = (u.x + (a.y - u.y) / slope);
				else
					a.x = u.x;
			}

			//falls c oben -> scheiden mit rechter Kante
			if ((c & TOP) != EMPTY) {
				a.y = ymax;
				if (finiteslope)
					a.x = (u.x + (a.y - u.y) / slope);
				else
					a.x = u.x;
			}

			if (c == c1) {
				u.x = a.x;
				u.y = a.y;
				c1 = regioncode(u, rect);
			} else {
				v.x = a.x;
				v.y = a.y;
				c2 = regioncode(v, rect);
			}

		}

		return true;
	}

}