package uibk.mtk.draw3d.base;

import java.awt.Graphics2D;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.image.BufferedImage;

import uibk.mtk.lang.Highlightable;
import uibk.mtk.lang.MathPanel;
import uibk.mtk.lang.PrepaintComputable;

/**
 * Basisklasse f?r alle Aktionen der 3D-Applets.
 * 
 * <p>
 * Der Panel enth?lt eine <CODE>Scene3D</CODE> in der die zu diesem Panel
 * hinzugef?gten <CODE>Drawable3D</CODE> Objekte gezeichnet werden. Der
 * Mathpanel3D verwaltet das Zeichnen dieser Objekte, erlaubt Verschieben mit
 * der Maus von Dragable Objekten und zeigt gegebenfalls Tooltips der Objekte an
 * (siehe Interface <CODE>Tooltipable</CODE>) oder hebt Sie hervor (siehe
 * Interface <CODE>Highlightable</CODE>). Weiters erlaubt er das Rotieren der
 * Szene durch Ziehen mit der Maus (nicht gleichzeitig mit drag & drop)
 * </p>
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class MathPanel3D extends MathPanel {

	/** gibt an ob die Szene durch Ziehen der Maus rotieren kann */
	boolean listenforSceneRot = false;

	/** Ver?nderung des Winkels bei drag&drop */
	private static final double ROT = 0.05;

	/** Z-Buffer */
	protected double[][] zbuffer;

	/** die 3D-Szene die der Panel enth?lt */
	protected Scene3D scene3d;

	/** MausMotionListener f?r das Rotieren der Szene */
	private MyMouseMotionListening mouseMotionListener = new MyMouseMotionListening();

	/** MausListener f?r das Rotieren der Szene */
	private MyMouseListening mouseListener = new MyMouseListening();

	/** alte x-Position der Maus (f?r Rotieren der Szene) */
	int oldy = MOUSEPOS_NOT_INITILIZED;

	/** alte y-Position der Maus (f?r Rotieren der Szene) */
	int oldx = MOUSEPOS_NOT_INITILIZED;

	/** Kennzeichen f?r nicht initialisierte Mausposition */
	private static int MOUSEPOS_NOT_INITILIZED = -1;

	/** Erzeugt den Panel */
	public MathPanel3D() {

		scene3d = new Scene3D();
		scene3d.setPanel(this);
		setListenForSceneRotation(true);

	}

	private class MyMouseListening extends MouseAdapter {

		public void mouseReleased(MouseEvent e) {
			oldx = MOUSEPOS_NOT_INITILIZED;
			oldy = MOUSEPOS_NOT_INITILIZED;
		}

	}

	/** Maus-Ereignisverarbeitung: Rotieren der Szene */
	private class MyMouseMotionListening extends MouseMotionAdapter {

		public void mouseDragged(MouseEvent ev) {

			if (oldx == MOUSEPOS_NOT_INITILIZED
					&& oldy == MOUSEPOS_NOT_INITILIZED) {
				oldx = ev.getX();
				oldy = ev.getY();
				return;
			}

			double max = Math.max(Math.abs(ev.getY() - oldy), Math.abs(ev
					.getX()
					- oldx));
			if (max == 0)
				max = 1;
			double offx = ev.getY() < oldy ? ROT : -ROT;
			offx *= Math.abs(ev.getY() - oldy) / max;

			scene3d.setRotationX(scene3d.getRotationX() + offx);

			oldy = ev.getY();

			double offz = ev.getX() < oldx ? -ROT : ROT;
			offz *= Math.abs(ev.getX() - oldx) / max;

			scene3d.setRotationZ(scene3d.getRotationZ() + offz);

			oldx = ev.getX();

			repaint();

		}
	}

	/** gibt die 3D-Szene die der Panel enth?lt zur?ck */
	public Scene3D getScene3D() {
		return scene3d;

	}

	/** (de)aktivieren des Rotieren der Szene mit der Maus */
	public void setListenForSceneRotation(boolean listenForDrags) {

		if (listenForDrags == this.listenforSceneRot) {
			return;
		}

		this.listenforSceneRot = listenForDrags;

		if (listenForDrags) {

			this.addMouseMotionListener(mouseMotionListener);
			this.addMouseListener(mouseListener);
		} else {

			this.removeMouseMotionListener(mouseMotionListener);
			this.removeMouseListener(mouseListener);

		}

	}

	/**
	 * Hinzuf?gen eines Drawable Objektes zum MathPanel. Falls <CODE>obj
	 * </CODE> nicht vom Typ <CODE>Drawable3D</CODE> oder <CODE>null</CODE>
	 * ist wird ein <CODE>IllegalArgumentException</CODE> geworfen. Falls es
	 * schon hinzugef?gt worden ist, passiert nichts.
	 * 
	 *  
	 */
	public void add(Object obj) {
		if (obj == null || !(obj instanceof Drawable3D))
			throw new IllegalArgumentException(""); //$NON-NLS-1$

		if (!drawItems.contains(obj)) {
			Drawable3D draw = (Drawable3D) obj;

			super.add(draw);
			draw.setMathPanel3D(this);
		}
	}

	/** setzt den z-Buffer zur?ck */
	private void resetZBuffer() {
		for (int i = 0; i < getWidth() - 1; i++) {

			for (int j = 0; j < getHeight() - 1; j++)
				zbuffer[i][j] = Double.POSITIVE_INFINITY;
		}

	}

	/** Anpassen der Buffer */
	protected void preparebuffer() {

		super.preparebuffer();

		int w = getWidth();
		int h = getHeight();

		if (zbuffer == null || w > zbuffer.length || h > zbuffer[0].length) {

			zbuffer = new double[w][h];

		}

		resetZBuffer();

	}

	/** Zeichnen aller Drawable Objekte in dem Panel */
	protected synchronized void drawDrawables(BufferedImage framebuffer,
			Graphics2D g2) {

		//Abbildungsmatrizen neu berechnen
		scene3d.prepareScene();

		options.setRenderingHints(g2);
		int ct = drawItems.size();

		if (repaintrequested) {
			//alle Objekte neu berechnen
			for (int i = 0; i < ct; i++) {
				if (drawItems.elementAt(i) instanceof PrepaintComputable
						&& ((Drawable3D) drawItems.elementAt(i)).isVisible()) {
					try {
						((PrepaintComputable) drawItems.elementAt(i))
								.prepaintcompute();
					} catch (Exception ex) {

						reportError(ex);
					}

				}
			}

		}

		//Drawables zeichnen
		for (int i = 0; i < ct; i++) {
			Drawable3D d = (Drawable3D) drawItems.elementAt(i);
			if (d.isVisible()) {
				if (d == highlighted) {

					((Highlightable) d).highlight(framebuffer, g2);
				} else
					d.draw(framebuffer, g2);
			}
		}

		repaintrequested = false;

	}

}