package uibk.mtk.draw3d.base;

import uibk.mtk.geom.geom3d.Punkt3D;
import uibk.mtk.geom.geom3d.Vector3D;
import uibk.mtk.lang.Messages;

/**
 * 4x4 Matrix zum Berechnen von 3D-Transformationen
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Matrix3D {

	private static final String BUNDLE_NAME = "uibk.mtk.draw3d.base.messages";//$NON-NLS-1$

	/** Die Eintr?ge der Matrix */
	private double[][] coeffs = new double[4][4];

	/** erzeugt die Nullmatrix */
	public Matrix3D() {

	}

	/** Multiplikation von zwei Matrizen */
	public Matrix3D multiply(Matrix3D B) {
		Matrix3D mult = new Matrix3D();

		for (int i = 0; i < 4; i++)
			for (int j = 0; j < 4; j++)
				for (int k = 0; k < 4; k++)
					mult.coeffs[i][j] += coeffs[i][k] * B.coeffs[k][j];
		return mult;
	}

	/** Matrix-Vektor Multiplikation */
	public Vector3D multiply(Punkt3D v) {
		Vector3D mult = new Vector3D();

		mult.x1 = coeffs[0][0] * v.x1 + coeffs[0][1] * v.x2 + coeffs[0][2]
				* v.x3 + coeffs[0][3] * v.h;
		mult.x2 = coeffs[1][0] * v.x1 + coeffs[1][1] * v.x2 + coeffs[1][2]
				* v.x3 + coeffs[1][3] * v.h;
		mult.x3 = coeffs[2][0] * v.x1 + coeffs[2][1] * v.x2 + coeffs[2][2]
				* v.x3 + coeffs[2][3] * v.h;
		mult.h = coeffs[3][0] * v.x1 + coeffs[3][1] * v.x2 + coeffs[3][2]
				* v.x3 + coeffs[3][3] * v.h;

		return mult;
	}

	/** gibt den (i,j)-ten Eintrag zur?ck */
	public double getCoeff(int i, int j) {
		return coeffs[i - 1][j - 1];
	}

	/** erzeugt die Einheitsmatrix */
	public static Matrix3D getIdentityMatrix() {
		Matrix3D matrix = new Matrix3D();
		for (int i = 0; i < 4; i++)
			matrix.coeffs[i][i] = 1;
		return matrix;
	}

	/**
	 * Gibt die Skalierungsmatrix, mit den ?bergebenen Faktoren zur?ck
	 * (Skalierung f?r den homogenen Teil=1)
	 */
	public static Matrix3D getScaleMatrix(double xs, double ys, double zs) {
		Matrix3D matrix = getIdentityMatrix();
		matrix.coeffs[0][0] = xs;
		matrix.coeffs[1][1] = ys;
		matrix.coeffs[2][2] = zs;
		return matrix;
	}

	/** gibt einen String mit den Koeffizienten der Matrix zur?ck */
	public String toString() {
		String str = ""; //$NON-NLS-1$

		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				str = str + coeffs[i][j] + " "; //$NON-NLS-1$
			}
			str += "\n"; //$NON-NLS-1$
		}
		return str;
	}

	/** gibt die Translationsmatrix um den Vektor (xt,yt,zt) zur?ck */
	public static Matrix3D getTranslationMatrix(double xt, double yt, double zt) {
		Matrix3D matrix = getIdentityMatrix();

		matrix.coeffs[0][3] = xt;
		matrix.coeffs[1][3] = yt;
		matrix.coeffs[2][3] = zt;

		return matrix;

	}

	/** gibt die Matrix f?r die Parallelprojektion auf die XZ-Ebene zur?ck */
	public static Matrix3D getParallelProjectionMatrixXZPlane() {
		Matrix3D matrix = Matrix3D.getIdentityMatrix();
		matrix.coeffs[1][1] = 0;
		return (matrix);
	}

	/**
	 * gibt die Matrix f?r die ZNetralprojektion auf die XZ-Ebene mit Auge in
	 * (0,distance,0) zur?ck
	 */
	public static Matrix3D getPerspectiveProjectionMatrixXZPlane(double distance) {
		Matrix3D matrix = Matrix3D.getIdentityMatrix();
		matrix.coeffs[1][1] = 0;
		matrix.coeffs[3][1] = -1 / distance;
		return (matrix);
	}

	/**
	 * gibt die Matrix f?r die Rotation um die x-Achse zur?ck
	 * 
	 * @param angle
	 *            Winkel um den rotiert wird
	 */
	public static Matrix3D getRotationMatrixXAxis(double angle) {
		Matrix3D matrix = Matrix3D.getIdentityMatrix();

		matrix.coeffs[1][1] = Math.cos(angle);
		matrix.coeffs[1][2] = -Math.sin(angle);
		matrix.coeffs[2][1] = Math.sin(angle);
		matrix.coeffs[2][2] = Math.cos(angle);

		return (matrix);
	}

	/**
	 * gibt die Matrix f?r die Rotation um die z-Achse zur?ck
	 * 
	 * @param angle
	 *            Winkel um den rotiert wird
	 */
	public static Matrix3D getRotationMatrixZAxis(double angle) {
		Matrix3D matrix = Matrix3D.getIdentityMatrix();

		matrix.coeffs[0][0] = Math.cos(angle);
		matrix.coeffs[0][1] = -Math.sin(angle);
		matrix.coeffs[1][0] = Math.sin(angle);
		matrix.coeffs[1][1] = Math.cos(angle);

		return (matrix);
	}

	/**
	 * gibt die Matrix f?r die Rotation um die y-Achse zur?ck
	 * 
	 * @param angle
	 *            Winkel um den rotiert wird
	 */
	public static Matrix3D getRotationMatrixYAxis(double angle) {
		Matrix3D matrix = Matrix3D.getIdentityMatrix();

		matrix.coeffs[0][0] = Math.cos(angle);
		matrix.coeffs[0][2] = -Math.sin(angle);
		matrix.coeffs[2][0] = Math.sin(angle);
		matrix.coeffs[2][2] = Math.cos(angle);

		return (matrix);
	}

	/** gibt die transponierte Matrix einer Matrix zur?ck */
	public static Matrix3D transpose(Matrix3D A) {
		Matrix3D T = new Matrix3D();

		for (int i = 0; i < 4; i++)
			for (int j = 0; j < 4; j++)
				T.coeffs[i][j] = A.coeffs[j][i];

		return T;

	}

	/** erzeugt eine Matrix mit den Eintr?gen der ?bergebenen Matrix */
	public Matrix3D(Matrix3D A) {
		for (int i = 0; i < 4; i++)
			for (int j = 0; j < 4; j++)
				coeffs[i][j] = A.coeffs[i][j];

	}

	/**
	 * Berechnet die Inverse einer Matrix. Wirft eine Exception falls die Matrix
	 * singul?r ist
	 */
	public Matrix3D inverse() throws Exception {
		Matrix3D temp = new Matrix3D(this);
		Matrix3D inv = getIdentityMatrix();

		for (int j = 0; j < 4; j++) {

			double e = temp.coeffs[j][j];
			if (e == 0)
				throw new Exception(Messages.getString(BUNDLE_NAME,
						"Matrix3D.3")); //$NON-NLS-1$
			for (int i = 0; i < 4; i++) {
				temp.coeffs[j][i] = temp.coeffs[j][i] / e;
				inv.coeffs[j][i] = inv.coeffs[j][i] / e;
			}

			for (int k = 0; k < 4; k++) {
				if (k != j) {
					e = -temp.coeffs[k][j];
					for (int i = 0; i < 4; i++) {
						temp.coeffs[k][i] = temp.coeffs[k][i] + e
								* temp.coeffs[j][i];
						inv.coeffs[k][i] = inv.coeffs[k][i] + e
								* inv.coeffs[j][i];
					}
				}
			}
		}
		return inv;
	}

}