package uibk.mtk.draw3d.base;

import java.awt.Point;

import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.geom2d.Vector2D;
import uibk.mtk.geom.geom3d.CoordinateBox3D;
import uibk.mtk.geom.geom3d.Punkt3D;
import uibk.mtk.geom.geom3d.Vector3D;

/**
 * Eine 3D-Szene mit Perspektivischer- und Parallel-Projektion.
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Scene3D {

	/** Modelviewmatrix zur Transformation Modell->Weltkoordinaten */
	protected Matrix3D M;

	/** Viewporttransfomationsmatrix */
	protected Matrix3D V;

	/** Projektionsmatrix f?r Parallel/Zentralporj. */
	Matrix3D P;

	/**
	 * Matrix die die Normalvektoren von Modellkoordinaten zu Weltkoord.
	 * transformiert, N=(M^(-1))^T
	 */
	Matrix3D N;

	/** Produkt der 3 Transformationsmatrizen M,V und P */
	Matrix3D T;

	/** MathPanel3D in den die Szene gezeichnet wird */
	MathPanel3D mathpanel3d;

	/**
	 * Zeigt an ob sich etwas Szene ver?ndert hat falls ja wird von MathPanel3d
	 * vor Zeichnen die Funktion prepaintcompute der Drawables aufgerufen
	 */

	/** Skalierungsmodus */
	int scalingmode = SCALE_INDEPENDENT;

	/** Rotation um x-Achse */
	public double rotationx = 0;

	/** Rotation um z-Achse */
	public double rotationz = 0;

	/** Rotation um y-Achse */
	public double rotationy = 0;

	/** Bestimmt Position der Kamera, die sich in (0,camera,0) befindet */
	double distance = 4;

	/** Skalierungsfaktor in x-Richtung */
	double scalex = 1;

	/** Skalierungsfaktor in y-Richtung */
	double scaley = 1;

	/** Skalierungsfaktor in z-Richtung */
	double scalez = 1;

	/** Prozent um die gerundet wird */
	double roundingpercent = 2;

	/** Runden aktiv/nicht aktiv */
	boolean roundingenabled = true;

	/**
	 * Die Grenzen f?r die normierten Koordinaten alles wird so transformiert
	 * das es im W?rfel [-1,-1,-1] ... liegt
	 */
	private CoordinateBox3D nc_window = new CoordinateBox3D();

	/** aktuelle Koordinatengrenzen */
	private CoordinateBox3D boundingbox = new CoordinateBox3D();

	/** Original Koordinatengrenzen */
	private CoordinateBox3D orig_boundingbox = new CoordinateBox3D();

	public static final int SCALE_EQUAL = 0;

	public static final int SCALE_INDEPENDENT = 1;

	static final double DEGREE90 = 90.0 / 360 * 2 * Math.PI;

	static final double DEGREE180 = 180.0 / 360 * 2 * Math.PI;

	//boolean clip = true;
	/** Konstane f?r Perspektive Projektion */
	public static final int PROJECT_PERSPECTIVE = 0;

	/** Konstane f?r Perspektive Parallelprojektion */
	public static final int PROJECT_PARALLEL = 1;

	/**
	 * legt den Typ der Projektion fest, entweder PROJECT_PERSPECTIVE oder
	 * PROJECT_PARALLEL
	 */
	int projectiontype = PROJECT_PERSPECTIVE;

	public void setScalingMode(int mode) {
		this.scalingmode = mode;

	}

	public int getScalingMode() {
		return scalingmode;
	}

	/** aktiviert das Runden der Grenzen */
	void enableRounding(boolean b) {
		roundingenabled = b;
	}

	/** Setzt die Prozent fest um die die Achsen gerundet werden */
	public void setRoundingPercent(double percent) {
		this.roundingpercent = percent;
	}

	/**
	 * Multipliziert einen Punkt in Modellkoordinaten mit der Modelviewmatrix,
	 * d.h. die Funktion berechnet die Weltkoordinaten des Punktes
	 * 
	 * @param p
	 *            Punkt in Modellkoordinaten
	 * @return Punkt in Weltkoordinanten
	 */
	public Punkt3D getWC(Punkt3D p) {
		return M.multiply(p);
	}

	/** Setzt die Werte f?r die Transformationen auf Standardwerte zur?ck */
	public void defaultview() {
		projectiontype = PROJECT_PERSPECTIVE;
		distance = 3;
		rotationx = 0.0;
		rotationz = 1.2;
		rotationy = 0.0;
		scalex = 1;
		scaley = 1;
		scalez = 1;

	}

	/**
	 * setzt die Rotationen um die Achsen so, dass sich die Projektion auf die
	 * XZ-Ebene ergibt
	 */
	public void projectXZPlane() {
		rotationx = 0;
		rotationz = DEGREE180;
		rotationy = 0;

	}

	/**
	 * setzt die Rotationen um die Achsen so, dass sich die Projektion auf die
	 * XY-Ebene ergibt
	 */
	public void projectXYPlane() {
		rotationx = DEGREE90;
		rotationz = 0;
		rotationy = DEGREE180;

	}

	/**
	 * setzt die Rotationen um die Achsen so, dass sich die Projektion auf die
	 * YZ-Ebene ergibt
	 */
	public void projectYZPlane() {
		rotationx = 0;
		rotationz = DEGREE90;
		rotationy = 0;

	}

	/**
	 * gibt den Projektionstyp zur?ck
	 * 
	 * @return entweder PROJECT_PERSPECTIV oder PROJECT_PARALLEL
	 */
	public int getProjectionType() {
		return projectiontype;
	}

	/** legt die Distanz der Kamera zur XZ-Ebene fest */
	public void setDistance(double distance) {
		this.distance = distance;

	}

	/** Setzt den Skalierungsfaktor in x-Richtung */
	public void setXScaleFactor(double scalex) {
		this.scalex = scalex;

	}

	/** Setzt den Skalierungsfaktor in y-Richtung */
	public void setYScaleFactor(double scaley) {
		this.scaley = scaley;

	}

	/** Setzt den Skalierungsfaktor in z-Richtung */
	public void setZScaleFactor(double scalez) {
		this.scalez = scalez;

	}

	/** Gibt denSkalierungsfaktor in x-Richtung zur?ck */
	public double getXScaleFactor() {
		return scalex;
	}

	/** Gibt denSkalierungsfaktor in y-Richtung zur?ck */
	public double getYScaleFactor() {
		return scaley;
	}

	/** Gibt denSkalierungsfaktor in z-Richtung zur?ck */
	public double getZScaleFactor() {
		return scalez;
	}

	/** Erzeugt eine 3D-Szene mit Standardeinstellungen f?r Rotation,... */
	public Scene3D() {
		defaultview();
	}

	/** setzt den Wert der Drehung um die x-Achse */
	public void setRotationX(double rot) {
		this.rotationx = rot;

	}

	/** setzt den Wert der Drehung um die y-Achse */
	public void setRotationY(double rot) {
		this.rotationy = rot;

	}

	/** setzt den Wert der Drehung um die z-Achse */
	public void setRotationZ(double rot) {
		this.rotationz = rot;

	}

	/**
	 * legt den Projektionstyp fest
	 * 
	 * @param type
	 *            PROJECT_PERSPECTIVE f?r Perspektive o der PROJECT_PARALLEL f?r
	 *            Parallelprojektion
	 */
	public void setProjectionType(int type) {
		this.projectiontype = type;

	}

	/**
	 * gibt zur?ck ob geclippt werden soll oder nicht das Clipping m?ssen alle
	 * Objekte selber durchf?hren
	 */
	public boolean isClippingEnabled() {
		return true;
	}

	/** gibt den Wert der Drehung um die x-Achse zur?ck */
	public double getRotationX() {
		return rotationx;
	}

	/** gibt den Wert der Drehung um die y-Achse zur?ck */
	public double getRotationY() {
		return rotationy;
	}

	/** gibt den Wert der Drehung um die z-Achse zur?ck */
	public double getRotationZ() {
		return rotationz;
	}

	/** setzt den MathPanel3D fest in den die Szene gezeichnet wird */
	public void setPanel(MathPanel3D mathpanel3d) {
		this.mathpanel3d = mathpanel3d;
	}

	/**
	 * setzt die BoundingBox f?r die Scene3D, falls die ?bergebnene Boudningbox
	 * leer ist werden die entsprechnenden Grenzen um 1 erh?ht bzw. verkleinert.
	 */
	public void setBoundingBox(CoordinateBox3D box) {

		this.boundingbox = box;
		if (box.getXRange() == 0) {
			boundingbox.xmin = box.xmax - 1;
			boundingbox.xmax = box.xmax + 1;
		}
		if (box.getZRange() == 0) {
			boundingbox.zmin = box.zmax - 1;
			boundingbox.zmax = box.zmax + 1;
		}
		if (box.getYRange() == 0) {
			boundingbox.ymin = box.ymax - 1;
			boundingbox.ymax = box.ymax + 1;
		}
		scalex = 1;
		scaley = 1;
		scalez = 1;
		orig_boundingbox = (CoordinateBox3D) box.clone();

	}

	/**
	 * gibt eine Kopie der aktuellen BoundingBox (mit aktueller Skalierung,...)
	 * zur?ck
	 */
	public CoordinateBox3D getBoundingBox() {
		return (CoordinateBox3D) boundingbox.clone();
	}

	/** gibt eine Kopie der Original-BoundingBox zur?ck */
	public CoordinateBox3D getOriginalBoundingBox() {
		return orig_boundingbox;
	}

	/**
	 * gibt die Ebene zur?ck an der geclippt werden soll (wir klippen an einer
	 * Ebene parallel zur XZ-Ebene also, wird nur die y-Koordinate
	 * zur?ckgegeben)
	 */
	public double getClippingPlane() {
		return distance - 0.15;
	}

	/**
	 * Berechnet die Transformationsmatrizen zu den aktuellen Optionen der
	 * 3D-Szene. Wird vom MathPanel3D der die Szene beinhaltet vor dem Zeichnen
	 * aufgerufen-
	 */
	public void prepareScene() {

		boundingbox = (CoordinateBox3D) orig_boundingbox.clone();

		if (roundingenabled)
			boundingbox.roundCoords(roundingpercent, CoordinateBox3D.ROUND_XYZ);

		setupModelViewMatrix();

		if (projectiontype == PROJECT_PERSPECTIVE)
			P = Matrix3D.getPerspectiveProjectionMatrixXZPlane(distance);
		else
			P = Matrix3D.getParallelProjectionMatrixXZPlane();

		setupViewPortTransformationMatrix();

		//alle Transformationen zusammen
		T = V.multiply(P.multiply(M));

		setupNormalTransMatrix();
	}

	/** gibt die Position der Kamera zur?ck */
	public Punkt3D getCamera() {
		return new Punkt3D(0, distance, 0);
	}

	/**
	 * berechnet die Viewport-Transformationsmatrix,die die Koordinaten nach
	 * Anwendung die nach Anwendung der Modellviewtransformation und Projektion
	 * in Koordinaten innerhalb des Rechtecks ("Window") [-1,1,-1,1] geben sind
	 * auf den Ausgabebereich ("Viewport") des MathPanel3D um
	 */
	private void setupViewPortTransformationMatrix() {

		int vx_min = 0;
		int vy_min = 0;
		int vx_max = mathpanel3d.getWidth();
		int vy_max = mathpanel3d.getHeight();

		//Skalierung Window
		Matrix3D SN = Matrix3D.getScaleMatrix(
				1 / (nc_window.xmax - nc_window.xmin),
				1 / (nc_window.ymax - nc_window.ymin),
				1 / (nc_window.zmax - nc_window.zmin));

		//Skalierung Viewport
		Matrix3D SV = Matrix3D.getScaleMatrix(vx_max - vx_min, 1, vy_max
				- vy_min);

		//Translation Viewport
		Matrix3D TV = Matrix3D.getTranslationMatrix(vx_min, 0, vy_min);

		//Translation Window
		Matrix3D TW = Matrix3D.getTranslationMatrix(-nc_window.xmin,
				-nc_window.ymin, -nc_window.zmin);

		//Verschieben von Window nach (0,0,0), skalieren auf normierte
		// Koordinaten
		//skalieren auf Viewport verschieben auf gew?nschte Pos.
		V = TV.multiply(SV.multiply(SN.multiply(TW)));

	}

	/**
	 * Transformiert einen Punkt in Weltkoodinaten in Ger?tekoordinaten
	 * 
	 * @param a
	 *            ein Punkt in Modellkoordinaten
	 * @return die Projektion des Punktes
	 */
	public Point wctoDevice(Punkt3D a) {
		Matrix3D Z = V.multiply(P);
		Punkt3D p = Z.multiply(a);
		double h = p.h;

		p.x1 /= h;
		p.x3 /= h;

		return new Point(mathpanel3d.getWidth() - (int) p.x1, mathpanel3d
				.getHeight()
				- (int) p.x3);

	}

	/**
	 * f?hrt die gesamte Transformation auf einen Punkt gegeben in
	 * Modellkoordinaten aus, d.h. der Punkt wird auf die Ausgabefl?che
	 * "projeziert".Im Gegensatz zu der Funkion toDevice wird der Punkt nicht
	 * als Point (Koordinaten sind Integers) sondern als Punkt2D zur?ckgeben
	 * (falls weitere Berechnungen notwendig sind)
	 * 
	 * @param a
	 *            ein Punkt in Modellkoordinaten
	 * @return die Projektion des Punktes
	 */
	public Point mcToDevice(Punkt3D a) {
		Punkt3D p = T.multiply(a);
		double h = p.h;
		return new Point(mathpanel3d.getWidth() - (int) (p.x1 / h), mathpanel3d
				.getHeight()
				- (int) (p.x3 / h));
	}

	/**
	 * f?hrt die gesamte Transformation auf einen Punkt gegeben in
	 * Modellkoordinaten aus, d.h. der Punkt wird auf die Ausgabefl?che
	 * "projeziert".Im Gegensatz zu der Funkion toDevice wird der Punkt nicht
	 * als Point (Koordinaten sind Integers) sondern als Punkt2D zur?ckgeben
	 * (falls weitere Berechnungen notwendig sind)
	 * 
	 * @param a
	 *            ein Punkt in Modellkoordinaten
	 * @return die Projektion des Punktes
	 */
	public Punkt2D project(Punkt3D a) {
		Punkt3D p = T.multiply(a);
		double h = p.h;
		return new Punkt2D(mathpanel3d.getWidth() - (int) (p.x1 / h),
				mathpanel3d.getHeight() - (int) (p.x3 / h));
	}

	/**
	 * Tiefe eines Punktes (gegeben in Weltkoordinaten) berechnen. (wird
	 * verwendet f?r den Tiefenbuffer)
	 */
	public double calcdepth(Punkt3D p) {
		return getCamera().sub(p).quadnorm();

	}

	/** Erzeugt die Modellview-Matrix */
	private void setupModelViewMatrix() {

		Punkt3D center;
		if (boundingbox == null)
			center = new Punkt3D(0, 0, 0);
		else
			center = boundingbox.getCenter();

		Matrix3D TR = Matrix3D.getTranslationMatrix(-center.x1, -center.x2,
				-center.x3);

		double max = Math.max(Math.max(boundingbox.getXRange(), boundingbox
				.getYRange()), boundingbox.getZRange());
		Matrix3D SR;
		if (scalingmode == SCALE_EQUAL) {

			SR = Matrix3D.getScaleMatrix(nc_window.getXRange() / max, nc_window
					.getYRange()
					/ max, nc_window.getZRange() / max);
		} else {

			SR = Matrix3D.getScaleMatrix(nc_window.getXRange()
					/ boundingbox.getXRange(), nc_window.getYRange()
					/ boundingbox.getYRange(), nc_window.getZRange()
					/ boundingbox.getZRange());
		}

		//User Skalierung
		Matrix3D SUSER = Matrix3D.getScaleMatrix(scalex * 0.45, scaley * 0.45,
				scalez * 0.45);

		//Sklarung gesamt
		Matrix3D S = SUSER.multiply(SR);

		//Rotationsmatrizen
		Matrix3D rotX = Matrix3D.getRotationMatrixXAxis(rotationx);
		Matrix3D rotZ = Matrix3D.getRotationMatrixZAxis(rotationz);
		Matrix3D rotY = Matrix3D.getRotationMatrixYAxis(rotationy);

		Matrix3D R = rotX.multiply(rotZ.multiply(rotY));

		M = R.multiply(S.multiply(TR));

	}

	/**
	 * berechnet die Matrix mit der Normalvektoren von Modell zu Weltkoordinaten
	 * transformiert werden
	 */
	private void setupNormalTransMatrix() {
		try {
			Matrix3D temp = M.inverse();
			N = Matrix3D.transpose(temp);
		} catch (Exception ex) {

		}

	}

	/** gibt die Entfernung der Kamera von der XZ-Ebene zur?ck */
	public double getDistance() {
		return distance;
	}

	/** Transformiert einen Normalenvektor von Modell- in Weltkoordinaten */
	public Vector3D transformNormal(Vector3D v) {
		return N.multiply(v);
	}

	/**
	 * transformiert einen Richtungsvektor von Modellkoordinaten in
	 * Ger?tekoordinaten
	 */
	public Vector2D projectVector(Punkt3D base, Vector3D mc_dir) {

		Vector3D wc_dir = transformNormal(mc_dir);

		Punkt2D pzero = project(base);
		Vector2D dir;
		if (wc_dir.x2 > 0) {

			mc_dir.scaleself(-getBoundingBox().maxBoundingBox());
			Punkt2D temp = project(base.add(mc_dir));
			dir = temp.sub(pzero).toVector();
			dir.x = -dir.x;
			dir.y = -dir.y;
		} else {
			mc_dir.scaleself(getBoundingBox().maxBoundingBox());
			Punkt2D temp = project(base.add(mc_dir));
			dir = temp.sub(pzero).toVector();
		}
		dir.normalize();
		return dir;
	}
}