package uibk.mtk.draw3d.objects;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.image.BufferedImage;
import java.util.Vector;

import uibk.mtk.draw2d.util.LineClipping2D;
import uibk.mtk.draw3d.base.Drawable3D;
import uibk.mtk.draw3d.base.Scene3D;
import uibk.mtk.draw3d.util.LineClipping3D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.geom3d.Punkt3D;
import uibk.mtk.geom.geom3d.Vector3D;
import uibk.mtk.lang.PrepaintComputable;

/**
 * Klasse f?r (konvexe) Polyeder im Raum mit Entfernung bzw. Zeichnung mit
 * anderem Linienstiel(z.B. strichliert) verdeckter Kanten.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Polyeder3D extends Drawable3D implements PrepaintComputable {

	/** Farbe f?r unsichtbare Kante */
	Color colorhiddenedge;

	/** Farbe f?r sichtbare Kante */
	Color colorvisibleedge;

	/** Linienstil f?r unsichtbare Kante */
	Stroke strokehiddenedge;

	/** Linienstil f?r sichtbare Kante */
	Stroke strokevisibleedge;

	/***************************************************************************
	 * Die Weltkoordinanten der Vertice nach Anwendung der
	 * Modelviewtransformation auf die Modelkoordinaten
	 **************************************************************************/
	protected Punkt3D[] wc_vertices;

	/** setzt fest ob versteckte Linien gezeichnet werden */
	boolean drawhiddenedges = true;

	/** Modelkoordinaten des Polyeders** */
	protected Punkt3D[] mc_vertices;

	/** Ger?tekoordinaten Modelkoordinaten des Polyeders** */
	protected Punkt2D[] dc_vertices;

	/** Array mit Kanten des Polyeders* */
	protected Edge[] edges;

	/** Array mit den Fl?chen des Polyeders * */
	protected Face[] faces;

	/** Hilfsklasse f?r die Kanten des Polyeders */
	public class Edge {

		/** Index des ersten Punkts der Kante in des Arrays vertices/wc_vertices */
		public int p1;

		/**
		 * Index des zweiten Punkts der Kante in des Arrays vertices/wc_vertices
		 */
		public int p2;

		/** gibt an ob die Kante versteckt ist */
		public boolean hidden = true;

		/** Erzeugt eine Kante mit den ?bergebenen Indices der Eckpunkte */
		Edge(int p1, int p2) {
			this.p1 = p1;
			this.p2 = p2;
		}

		/**
		 * ?berpr?ft auf Sichtbarkeit und setzt die Membervariable hidden Eine
		 * Kante ist genau dann sichtbar, falls sie Kante mindestens einer
		 * sichtbaren Fl?che ist
		 */
		void checkVisibility() {
			hidden = true;
			for (int i = 0; i < faces.length; i++) {
				if (faces[i].containsEdge(this) && !faces[i].hidden) {
					hidden = false;
					break;
				}
			}
		}
	} //Ende Klasse Edge

	/**
	 * Hilfsklasse f?r die Fl?chen des Polyeders. Die Fl?chen m?ssen koplanar
	 * sein
	 */
	protected class Face {

		/** gibt an ob die Kante versteckt ist */
		boolean hidden = true;

		/** die Indices der Eckpunkte, die die Fl?che aufbauen */
		Vector indices = new Vector();

		/** f?gt einen Punkt (Index des Punktes) zur Fl?che hinzu */
		void addPoint(int i) {
			indices.add(new Integer(i));
		}

		/** gibt zur?ck ob die F?che die ?bergebene Kante enth?lt */
		public boolean containsEdge(Edge e) {
			for (int i = 0; i < indices.size(); i++) {
				int j = i < indices.size() - 1 ? i + 1 : 0;
				int index_p1 = ((Integer) indices.elementAt(i)).intValue();
				int index_p2 = ((Integer) indices.elementAt(j)).intValue();
				if ((index_p1 == e.p1 && index_p2 == e.p2)
						|| (index_p1 == e.p2 && index_p2 == e.p1))
					return true;
			}
			return false;
		}

		/**
		 * berechnet die Normale auf die Fl?che in Weltkoordinaten (zur
		 * Berechnung werden die ersten 3 Punkte der Fl?che verwendet)
		 */
		public Vector3D getSurfaceNormalWC() {
			Punkt3D p1 = wc_vertices[((Integer) indices.elementAt(0))
					.intValue()];
			Punkt3D p2 = wc_vertices[((Integer) indices.elementAt(1))
					.intValue()];
			Punkt3D p3 = wc_vertices[((Integer) indices.elementAt(2))
					.intValue()];
			Vector3D n = p1.sub(p2).crossProd(p3.sub(p2));
			n.normalize();
			return n;
		}

		/**
		 * berechnet die Normale auf die Fl?che in Modelkoordinaten (zur
		 * Berechnung werden die ersten 3 Punkte der Fl?che verwendet)
		 */
		public Vector3D getSurfaceNormalMC() {
			Punkt3D p1 = mc_vertices[((Integer) indices.elementAt(0))
					.intValue()];
			Punkt3D p2 = mc_vertices[((Integer) indices.elementAt(1))
					.intValue()];
			Punkt3D p3 = mc_vertices[((Integer) indices.elementAt(2))
					.intValue()];
			Vector3D n = p1.sub(p2).crossProd(p3.sub(p2));
			n.normalize();
			return n;
		}

		/**
		 * ?berpr?ft die Sichtbarkeit der Fl?che und setzt die Membervariable
		 * hidden (die Sichtbarkeit wird mit Hilfe der Fl?chennormalen bestimmt)
		 * 
		 * @param camera
		 *            Position der Kamera der Scene3D
		 */
		void checkVisibility(Punkt3D camera) {
			Vector3D n = getSurfaceNormalWC();
			Punkt3D surfacepoint = wc_vertices[((Integer) indices.elementAt(0))
					.intValue()];
			if (scene3d.getProjectionType() == Scene3D.PROJECT_PERSPECTIVE)
				hidden = camera.sub(surfacepoint).dotProd(n) < 0 ? true : false;
			else
				hidden = n.x2 < 0 ? true : false;
		}
	} //Ende der Klasse Face

	/**
	 * Erzeugt den konvexen Polyeder durch Angabe seiner Eckpunkte, Kanten,
	 * Fl?chen und seines Zentrums. Ruft auch Polyeder3D() um die Standardwerte
	 * zum Zeichnen zu setzten
	 * 
	 * @param vertices
	 *            die Eckpunkte in Modellkoordinaten
	 * @param edges
	 *            die Kanten des Polyeders
	 * @param faces
	 *            die Fl?chen des Polyeders
	 */
	Polyeder3D(Punkt3D[] vertices, Edge[] edges, Face[] faces) {
		this();
		this.mc_vertices = vertices;
		this.edges = edges;
		this.faces = faces;

	}

	Polyeder3D(double[][] vertices, int[][] e, int[][] f) {

		this();

		int nedges = e.length;
		edges = new Edge[nedges];
		for (int i = 0; i < nedges; i++)
			edges[i] = new Edge(e[i][0], e[i][1]);

		int nfaces = f.length;
		faces = new Face[nfaces];

		for (int i = 0; i < nfaces; i++) {

			faces[i] = new Face();
			for (int j = 0; j < f[i].length; j++)
				faces[i].addPoint(f[i][j]);
		}

		int nvertices = vertices.length;
		mc_vertices = new Punkt3D[nvertices];
		for (int i = 0; i < nvertices; i++)
			mc_vertices[i] = new Punkt3D(vertices[i][0], vertices[i][1],
					vertices[i][2]);

	}

	/**
	 * Erzeugt einen leeren Polyeder. Setzt Standardwerte zum Zeichnen der
	 * Linien.
	 */
	Polyeder3D() {
		colorhiddenedge = new Color(100, 100, 100, 100);
		colorvisibleedge = Color.black;
		strokevisibleedge = new BasicStroke();
		final float[] dash = { 10.0f };
		strokehiddenedge = new BasicStroke(1.0f, BasicStroke.CAP_BUTT,
				BasicStroke.JOIN_MITER, 10.0f, dash, 0.0f);
	}

	/**
	 * Vorabberechnung der Elemente des Polyeders: Berechnung der
	 * Weltkoordinaten, Bestimmung der Sichtbarkeit der Fl?chen und Kanten
	 */
	public void prepaintcompute() {

		if (wc_vertices == null)
			wc_vertices = new Punkt3D[mc_vertices.length];
		if (dc_vertices == null)
			dc_vertices = new Punkt2D[mc_vertices.length];

		//Modelkoordinaten -> Weltkoordinaten

		for (int i = 0; i < mc_vertices.length; i++) {
			wc_vertices[i] = scene3d.getWC(mc_vertices[i]);
		}

		//Modelkoordinaten -> Ger?tekoordinaten

		for (int i = 0; i < mc_vertices.length; i++) {
			dc_vertices[i] = scene3d.project(mc_vertices[i]);
		}

		//Sichtbarkeit Fl?chen bestimmen
		Punkt3D camera = scene3d.getCamera();

		for (int i = 0; i < faces.length; i++) {
			faces[i].checkVisibility(camera);
		}

		//Sichtbarkeit Kanten bestimmen
		for (int i = 0; i < edges.length; i++)
			edges[i].checkVisibility();

	}

	/**
	 * Zeichnen des Polyeders, Linien werden an der in der Klasse Scene3D
	 * definierten ClippingPlane geclippt
	 */
	public void draw(BufferedImage framebuffer, Graphics2D g2) {

		Stroke old = g2.getStroke();

		Rectangle clippingrect = new Rectangle(0, 0, panel.getWidth(), panel
				.getHeight());

		double yc = scene3d.getClippingPlane();

		Punkt3D ca = new Punkt3D();
		Punkt3D cb = new Punkt3D();

		Point p = new Point();
		Point q = new Point();

		for (int i = 0; i < edges.length; i++) {

			Edge e = edges[i];

			if (e.hidden && !drawhiddenedges)
				continue;

			if (e.hidden) {

				g2.setStroke(strokehiddenedge);
				g2.setColor(colorhiddenedge);
			} else {
				g2.setStroke(strokevisibleedge);
				g2.setColor(colorvisibleedge);

			}

			Punkt3D a = wc_vertices[e.p1];
			Punkt3D b = wc_vertices[e.p2];

			if (this.scene3d.isClippingEnabled()) {

				if (LineClipping3D.clipline3d(a, b, ca, cb, yc)) {

					Point p1 = scene3d.wctoDevice(ca);
					Point p2 = scene3d.wctoDevice(cb);

					if (LineClipping2D.cohensutherland(p1, p2, p, q,
							clippingrect))
						g2.drawLine(p.x, p.y, q.x, q.y);

				}
			} else {
				Point p1 = scene3d.wctoDevice(a);
				Point p2 = scene3d.wctoDevice(b);

				g2.drawLine(p1.x, p1.y, p2.x, p2.y);

			}

		}
		g2.setStroke(old);

	}

	/** setzt den Linienstil f?r sichtbare Kanten */
	public void setSrokeVisibleEdge(Stroke stroke) {
		strokevisibleedge = stroke;
	}

	/** setzt den Linienstil f?r unsichtbare Kanten */
	public void setSrokeHiddenEdge(Stroke stroke) {
		strokehiddenedge = stroke;

	}

	/** setzt die Farbe f?r sichtbare Kanten */
	public void setColorVisibleEdge(Color color) {
		colorvisibleedge = color;
	}

	/** setzt die Farbe f?r unsichtbare Kanten */
	public void setColorHiddenEdge(Color color) {
		colorhiddenedge = color;
	}

	/** legt fest ob die versteckten Kanten gezeichnet werden sollen */
	public void drawHiddenEdges(boolean draw) {
		drawhiddenedges = draw;
	}

	public void setVertices(Punkt3D[] vertices) {
		this.mc_vertices = vertices;
	}

}