package uibk.mtk.draw3d.objects.surface3d;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.Collections;
import java.util.Vector;

import uibk.mtk.draw3d.base.Rasterable3D;
import uibk.mtk.draw3d.rasterizer.Rasterizer;
import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom3d.CoordinateBox3D;
import uibk.mtk.geom.geom3d.Punkt3D;
import uibk.mtk.geom.geom3d.Vector3D;
import uibk.mtk.lang.PrepaintComputable;

/**
 * Klasse zum Visualisieren einer Fl?che die durch z=f(x,y) gegeben ist. Die
 * Fl?che kann entweder als Gittermodel oder als durch Dreiecke approximierte
 * Fl?che gezeichnet werden. Um die "richtige" Sichtbarkeit zu erreichen werden
 * die Dreiecke von "hinten" nach "vorne",also beginnend mit dem am weites
 * entfernten, gezeichnet. Um die "richtige" Sichtbarkeit beim Gittermodel zu
 * erreichen werden die Dreiecke ebenfalls von hinten nach vorne gezeichnet und
 * in der Hintergrundfarbe (weiss)gef?rbt. Die Fl?che kann entweder nach
 * Position im Raum gef?rbt werden, oder es werden die Vorder und Hinterfl?che
 * in zwei verschiedenen Farben gezeichnet.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */

public class Surface3D extends Rasterable3D implements PrepaintComputable {

	/** Netz aus dem die Fl?che aufgebaut ist */
	Mesh mesh;

	/** Klasse zum zus?tzlichen Rendern der Dreiecke der Fl?che */
	AdditionalRendering addrendering;

	/** Farbe zum Zeichnen des Gitters, falls nur Gitter gezeichnet wird */
	static final Color COLORMESH = new Color(255, 0, 0, 120);

	/**
	 * Farbe zum Zeichnen des Gitters Fl?che (d.h. die Dreiecke) gezeichnet
	 * werden
	 */
	static final Color COLORPATCHMESH = new Color(0, 0, 0, 120);

	/**
	 * Farbee f?r die Vorderseite (falls als als colormode FRONTBACK gesetzt
	 * ist)
	 */
	static final Color COLORFRONT = new Color(255, 0, 0);

	/**
	 * Farbee f?r die R?ckseiteVorderseite (falls als als colormode FRONTBACK
	 * gesetzt ist)
	 */
	static final Color COLORBACK = new Color(0, 0, 255);

	/** Farbmodell zum F?rben der Vertice */
	ColorModel colormodel = new ColorModelXYZ();

	/** HintergrundFarbee des Panels der Fl?che enth?lt */
	static final Color COLORBACKGROUND = Color.white;

	/**
	 * setzt den Farbemodus zum F?llen der Fl?che fest: entweder FRONTBACK oder
	 * VERTEXCOLORS
	 */
	protected int colormode = VERTEXCOLORS;

	/** Fl?che wird an Vorder und R?ckseite mit verschiedenen Farbeen gef?llt */
	public static final int FRONTBACK = 0;
	
	/** Variable fr die Art der Flche(Treppe,Funktionsflche) */
	private boolean step = false;

	/**
	 * Fl?che wird nach Postion im Raum gef?llt, genaues Farbeschema wird durch
	 * die Variable coloring festgesetzt
	 */
	public static final int VERTEXCOLORS = 1;

	/** setzt den Zeichenmodus f?r die Fl?che fest (eine der DRAW- Konstanten) */
	int drawmode = DRAW_PATCH_WIREFRAME;

	/** es wird das Gittermodell der Fl?che gezeichnet */
	public static final int DRAW_WIREFRAME = 0;

	/** es wird der sichtbare Teil des Gittermodells der Fl?che gezeichnet */
	public static final int DRAW_VISIBLE_WIREFRAME = 1;

	/** die Fl?che wird gef?llt und das Gittermodell gezeichnet */
	public static final int DRAW_PATCH_WIREFRAME = 2;

	/** die Fl?che wird gef?llt dargestellt */
	public static final int DRAW_PATCH = 3;

	/** Setzt das Gitter aus der die Fl?che aufgebaut ist */
	public void setMesh(Mesh mesh) {
		this.mesh = mesh;
		this.calcColors();
	}

	/** Erzeugt eine leere Fl?che */
	public Surface3D() {
		super();

	}

	/**
	 * Setzt den Zeichenmodus fest
	 * 
	 * @param mode
	 *            DRAW_WIREFRAME,DRAW_PATCH_WIREFRAME,DRAW_WIREFRAME oder
	 *            DRAW_VISIBLE_WIREFRAME
	 */
	public void setDrawMode(int mode) {
		this.drawmode = mode;
	}

	/**
	 * Liefert den Zeichenmodus eine der Konstanten
	 * DRAW_WIREFRAME,DRAW_PATCH_WIREFRAME,DRAW_WIREFRAME oder
	 * DRAW_VISIBLE_WIREFRAME
	 */
	public int getDrawMode() {
		return drawmode;
	}

	/** Zeichnen der Fl?che */
	public void draw(BufferedImage framebuffer, Graphics2D g2) {

		if (mesh == null)
			return;

		rasterizer.enableZBuffer(false);

		switch (drawmode) {
		case DRAW_WIREFRAME:
			drawmesh(g2);
			break;
		case DRAW_VISIBLE_WIREFRAME:
			drawVisibleMesh(g2);
			break;
		case DRAW_PATCH_WIREFRAME:
			drawpatchwireframe(g2);
			break;
		case DRAW_PATCH:
			drawpatch(g2);
			break;
		default:
			break;
		}
	}

	/** F?llen der Fl?che */
	private void drawpatch(Graphics2D g2) {

		Vector triangles = mesh.getTriangles();

		switch (colormode) {

		case VERTEXCOLORS:
			//Farbee durch Vertice bestimmet

			for (int i = 0; i < triangles.size(); i++) {
				Triangle tri = (Triangle) triangles.elementAt(i);

				rasterizer.setColor(tri.p1.color); //Farbe setzen falls
				// FLAT_SHADING aktiviert
				rasterizer.renderTriangle(tri);
				additionalRendering(tri, g2);

			}
			break;

		case FRONTBACK:
			//Vorder-R?ckseite
			int savemode = rasterizer.getShadingMode();
			rasterizer.setShadingMode(Rasterizer.FLAT);
			for (int i = 0; i < triangles.size(); i++) {
				Triangle tri = (Triangle) triangles.elementAt(i);
				//Vorder oder R?ckseite bestimmen
				Punkt3D cam = panel.getScene3D().getCamera();
				if (cam.sub(tri.p1.wc).dotProd(tri.normal) < 0)
					rasterizer.setColor(COLORFRONT);
				else
					rasterizer.setColor(COLORBACK);
				//Dreicke rendern
				rasterizer.renderTriangle(tri);
				//
			}
			rasterizer.setShadingMode(savemode);
			break;
		default:
			break;
		}
	}

	/** Fl?che f?llen und Gitter zeichnen */
	void drawpatchwireframe(Graphics2D g2) {
		Vector triangles = mesh.getTriangles();
		switch (colormode) {

		case VERTEXCOLORS:

			for (int i = 0; i < triangles.size(); i++) {
				Triangle tri = (Triangle) triangles.elementAt(i);
				rasterizer.setColor(tri.p1.color);
				rasterizer.renderTriangle(tri);
				rasterizer.setColor(COLORPATCHMESH);
				rasterizer.drawline(tri.p1, tri.p2);
				rasterizer.drawline(tri.p3, tri.p1);

				additionalRendering(tri, g2);
			}
			break;

		case FRONTBACK:
			int savemode = rasterizer.getShadingMode();
			rasterizer.setShadingMode(Rasterizer.FLAT);

			for (int i = 0; i < triangles.size(); i++) {
				Triangle tri = (Triangle) triangles.elementAt(i);
				Punkt3D cam = panel.getScene3D().getCamera();
				if (cam.sub(tri.p1.wc).dotProd(tri.normal) < 0)
					rasterizer.setColor(COLORFRONT);
				else
					rasterizer.setColor(COLORBACK);
				//Dreieck rendern
				rasterizer.renderTriangle(tri);
				rasterizer.setColor(COLORPATCHMESH);

				rasterizer.drawline(tri.p1, tri.p2);
				rasterizer.drawline(tri.p3, tri.p1);

				additionalRendering(tri, g2);
			}
			rasterizer.setShadingMode(savemode);
			break;
		default:
			break;

		}

	}

	/** Berechnet die Normalen auf die Dreieck */
	private void computenormals() {
		Punkt3D p1, p2, p3;
		Vector triangles = mesh.getTriangles();
		for (int i = 0; i < triangles.size(); i++) {
			Triangle tri = (Triangle) triangles.elementAt(i);
			p1 = tri.p1.wc;
			p2 = tri.p2.wc;
			p3 = tri.p3.wc;
			Vector3D n = p2.sub(p1).crossProd(p3.sub(p2));
			tri.normal = n;
		}
	}

	/**
	 * Zeichnet das sichtbare Gittermodel durch F?llen der Dreicke in
	 * HintergrundFarbee
	 */
	void drawVisibleMesh(Graphics2D g2) {
		int savemode = rasterizer.getShadingMode();
		rasterizer.setShadingMode(Rasterizer.FLAT);

		rasterizer.setColor(COLORBACKGROUND);
		Vector triangles = mesh.getTriangles();

		for (int i = 0; i < triangles.size(); i++) {
			Triangle tri = (Triangle) triangles.elementAt(i);
			rasterizer.setColor(COLORBACKGROUND);
			//Dreiecke f?llen
			rasterizer.renderTriangle(tri);
			//Gitter zeichnen
			rasterizer.setColor(COLORMESH);
			rasterizer.drawline(tri.p1, tri.p2);
			rasterizer.drawline(tri.p3, tri.p1);
			additionalRendering(tri, g2);
		}
		rasterizer.setShadingMode(savemode);
	}

	/**
	 * Berechnen der Welt- und Ger?tekoordinaten, der Fl?chennormalen und der
	 * Tiefe der Dreiecke
	 */
	public void prepaintcompute() {
		if (mesh == null)
			return;

		Vector triangles = mesh.getTriangles();
		Vector vertices = mesh.getVertices();

		//Welt- und Ger?tekoordinaten der Vertice berechnen
		for (int i = 0; i < vertices.size(); i++) {
			Vertex3D vertex = (Vertex3D) vertices.elementAt(i);
			vertex.dc = scene3d.mcToDevice(vertex.mc);
			vertex.wc = scene3d.getWC(vertex.mc);
		}

		//Tiefe der Vertice berechnen
		for (int i = 0; i < vertices.size(); i++) {
			Vertex3D vertex = (Vertex3D) vertices.elementAt(i);
			vertex.distance = scene3d.calcdepth(vertex.wc);
		}

		//Tiefe der Dreiecke berechnen
		for (int i = 0; i < triangles.size(); i++) {
			Triangle tri = (Triangle) triangles.elementAt(i);
			Punkt3D a = tri.p1.wc;
			Punkt3D b = tri.p2.wc;
			Punkt3D c = tri.p3.wc;
			Punkt3D center=new Punkt3D();
			if (step) {
				center=b.add(c);
				center.scaleself(1.0/2);
			} else {
				center = a.add(b).add(c);
				center.scaleself(1.0 / 3);
			}
			tri.distance = scene3d.calcdepth(center);
		}

		//Falls Vorder/R?ckseite-F?rbung Berechnung der Normalen
		if (colormode == FRONTBACK)
			computenormals();

		if (drawmode != DRAW_WIREFRAME)
			Collections.sort(triangles);

	}

	/**
	 * Setzt fest wie die Farbeen der Vertice berechnet werden
	 *  
	 */
	public void setVertexColorModel(ColorModel model) {
		this.colormodel = model;

		calcColors();
	}

	/**
	 * setzt den Farbemodus zum F?llen der Fl?che fest (VERTEXCOLORS oder
	 * FRONTBACK)
	 */
	public void setColorMode(int mode) {
		this.colormode = mode;

		if (colormode == VERTEXCOLORS)
			calcColors();
	}
	
	/**
	 * setzt je nach Art der Flche(Funktionsflche,Treppenfunktion)
	 * die Variable fr die Sichtbarkeit
	 */
	public void setVisibility(boolean g) {
		step = g;
	}

	/**
	 * gibt den Farbemodus zum F?llen der Fl?che zur?ck
	 * 
	 *  
	 */
	public int getColorMode() {
		return colormode;

	}

	/**
	 * gibt den Farbemodus zum F?llen der Fl?che zur?ck
	 * 
	 *  
	 */
	public ColorModel getColorModel() {
		return colormodel;

	}

	/** Berechnet die Farbeen f?r die Vertice */
	protected void calcColors() {
		if (mesh == null)
			return;

		CoordinateBox3D limits = mesh.getBoundingBox();
		Vector vertices = mesh.getVertices();
		colormodel.colorVertices(vertices.iterator(), limits);

	}

	/** Zeichnet das Gittermodell */
	private void drawmesh(Graphics2D g2) {

		Vector triangles = mesh.getTriangles();

		for (int i = 0; i < triangles.size(); i++) {
			Triangle tri = (Triangle) triangles.elementAt(i);
			rasterizer.setColor(COLORMESH);
			//Gitter zeichnen
			rasterizer.drawline(tri.p1, tri.p2);
			rasterizer.drawline(tri.p3, tri.p1);

			additionalRendering(tri, g2);

		}

	}

	/** Gibt die Koordinategrenzen der Fl?che zur?ck */
	public CoordinateBox3D getBoundingBox() {
		if (mesh == null)
			return null;
		else
			return mesh.getBoundingBox();
	}

	public CoordinateRect2D getXYLimits() {
		if (mesh == null)
			return null;
		else {
			CoordinateBox3D cube = mesh.getBoundingBox();
			return new CoordinateRect2D(cube.xmin, cube.xmax, cube.ymin,
					cube.ymax);

		}
	}

	/** Setzt die Klasse zum zus?tzlichen Rendern der Dreicke */
	protected void setAdditionalRendering(AdditionalRendering add) {
		this.addrendering = add;
	}

	/**
	 * Zus?tzliches Rendern durchf?hren (falls mittels <CODE>
	 * setAdditionalRendering(AdditionalRendering)</CODE> Klasse zum
	 * erweiterten Rendern gesetzt worden ist
	 */
	private void additionalRendering(Triangle tri, Graphics2D g2) {
		if (addrendering != null)
			addrendering.render(tri, g2);
	}
}