package uibk.mtk.geom.geom2d;

import uibk.mtk.math.Rounder;

/**
 * Ein Rechteck zur Festlegung der Grenzen eines Koordinatensystems durch Angabe
 * der minimalen und maximalen x-und y Koordinaten.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class CoordinateRect2D implements Cloneable {

	/** minimale x-Koordinate */
	public double xmin;

	/** maximale x-Koordinate */
	public double xmax;

	/** minimale y-Koordinate */
	public double ymin;

	/** maximale y-Koordinate */
	public double ymax;

	/** Code f?r Runden des x-Intervalls */
	public static final int ROUND_X = 0x1;

	/** Code f?r Runden des y-Intervalls */
	public static final int ROUND_Y = 0x2;

	/** Code f?r Runden des x und y-Intervalls */
	public static final int ROUND_XY = 0x3;

	/** Code das, das Koordinatenrechteck im 1. Quadranten liegt */
	public static final int AREA_UPPERRIGHT = 0x1;

	/** Code das, das Koordinatenrechteck im 2. Quadranten liegt */
	public static final int AREA_UPPERLEFT = 0x2;

	/** Code das, das Koordinatenrechteck im 3. Quadranten liegt */
	public static final int AREA_LOWERLEFT = 0x4;

	/** Code das, das Koordinatenrechteck im 4. Quadranten liegt */
	public static final int AREA_LOWERRIGHT = 0x8;

	/** Code das, das Koordinatenrechteck im 1. und 2. Quadranten liegt */
	public static final int AREA_UPPER = 0x3;

	/** Code das, das Koordinatenrechteck im 3. und 4. Quadranten liegt */
	public static final int AREA_LOWER = 0x12;

	/** Code das, das Koordinatenrechteck im 1. und 4. Quadranten liegt */
	public static final int AREA_RIGHT = 0x9;

	/** Code das, das Koordinatenrechteck im 2. und 3. Quadranten liegt */
	public static final int AREA_LEFT = 0x6;

	/** Code das, das Koordinatenrechteck in allen 4 Quadranten liegt */
	public static final int AREA_ALL = 0x15;

	/** Erzeugt das Koordinatenrechteck mit den angegebenen Grenzen */
	public CoordinateRect2D(double xmin, double xmax, double ymin, double ymax) {
		if (xmax < xmin || ymax < ymin)
			throw new IllegalArgumentException(""); //$NON-NLS-1$

		this.xmin = xmin;
		this.xmax = xmax;
		this.ymin = ymin;
		this.ymax = ymax;

	}

	/** Erzeugt das Koordinatenrechteck mit den Standardgrenzen (-1,1,-1,1) */
	public CoordinateRect2D() {
		this(-1, 1, -1, 1);
	}

	/** gibt <CODE>true</CODE> falls <CODE>xmin <=x <=xmax</CODE> gilt */
	public boolean xcontains(double x) {
		if (x <= xmax && x >= xmin)
			return true;

		return false;
	}

	/** gibt <CODE>true</CODE> falls sich <CODE>ymin <=y <=ymax</CODE> gilt */
	public boolean ycontains(double y) {
		if (y <= ymax && y >= ymin)
			return true;

		return false;
	}

	/**
	 * gibt <CODE>true</CODE> falls das Koordinatenrechteck das
	 * Koordinatenrechteck <CODE>rect</CODE> enth?lt
	 *  
	 */
	public boolean contains(CoordinateRect2D rect) {
		if (rect.xmin < xmin || rect.ymin < ymin || rect.xmax > xmax
				|| rect.ymax > ymax)
			return false;

		return true;
	}

	/**
	 * gibt <CODE>true</CODE> falls das Koordinatenrechteck den Punkt <CODE>p
	 * </CODE> enth?lt
	 *  
	 */
	public boolean contains(Punkt2D p) {
		return xcontains(p.x) && ycontains(p.y);
	}

	/**
	 * gibt <CODE>true</CODE> falls das Koordinatenrechteck den Punkt <CODE>
	 * (x,y)</CODE> enth?lt
	 *  
	 */
	public boolean contains(double x, double y) {
		return xcontains(x) && ycontains(y);
	}

	/** gibt <CODE>xmax-xmin</CODE> zur?ck */
	public double getXRange() {
		return xmax - xmin;
	}

	/** gibt <CODE>ymax-ymin</CODE> zur?ck */
	public double getYRange() {
		return ymax - ymin;
	}

	/** gibt eine Zeichenkette mit den Grenzen zur?ck */
	public String toString() {
		return "[xmin=" //$NON-NLS-1$
				+ xmin + ", xmax=" //$NON-NLS-1$
				+ xmax + ", ymin=" //$NON-NLS-1$
				+ ymin + ", ymax=" //$NON-NLS-1$
				+ ymax + "]"; //$NON-NLS-1$
	}

	/** Erzeugt eine Kopie des Objekts */
	public Object clone() {
		return new CoordinateRect2D(xmin, xmax, ymin, ymax);
	}

	/**
	 * Kombiniert das Rechteck mit einem zweiten Rechteck <CODE>rect</CODE> zu
	 * einem neuen Rechteck, sodass beide in dem neuen enthalten sind.
	 */
	public CoordinateRect2D combine(CoordinateRect2D rect) {

		double newxmin = this.xmin < rect.xmin ? this.xmin : rect.xmin;
		double newxmax = this.xmax > rect.xmax ? this.xmax : rect.xmax;
		double newymin = this.ymin < rect.ymin ? this.ymin : rect.ymin;
		double newymax = this.ymax > rect.ymax ? this.ymax : rect.ymax;
		return new CoordinateRect2D(newxmin, newxmax, newymin, newymax);
	}

	/**
	 * Runden der Grenzen des Koordinatenw?rfels
	 * 
	 * @param percent
	 *            Prozent der jeweiligen Intervalll?nge um die gerundet werden
	 *            soll
	 * @param adjust
	 *            eine der ROUND-Konstanten, gibt an welche Intervalle
	 *            gerundetet werden sollen.
	 */
	public void roundCoords(double percent, int adjust) {

		double newymin = ymin;
		double newymax = ymax;
		double newxmin = xmin;
		double newxmax = xmax;

		double distx = Math.abs(((xmax - xmin) / 100.0) * percent);
		double disty = Math.abs(((ymax - ymin) / 100.0) * percent);

		//Runden y-Intervall
		if ((adjust & ROUND_Y) == ROUND_Y) {
			newymax = Rounder.getUpperRoundValue(ymax, disty);
			newymin = Rounder.getLowerRoundValue(ymin, disty);

		}

		//Runden x-Intervall
		if ((adjust & ROUND_X) == ROUND_X) {
			newxmax = Rounder.getUpperRoundValue(xmax, distx);
			newxmin = Rounder.getLowerRoundValue(xmin, distx);

		}

		//nach null suchen y-Achse
		if ((adjust & ROUND_Y) == ROUND_Y) {
			if (ymin > 0 && ymin - disty < 0) {
				newymin = 0;
			}
			if (ymax < 0 && ymax + disty > 0) {
				newymax = 0;
			}
		}

		//nach null suchen x-Achse
		if ((adjust & ROUND_X) == ROUND_X) {
			if (xmin > 0 && xmin - distx < 0) {
				newxmin = 0;
			}
			if (xmax < 0 && xmax + distx > 0) {
				newxmax = 0;
			}
		}

		ymin = newymin;
		ymax = newymax;
		xmin = newxmin;
		xmax = newxmax;

	}

	/**
	 * Gibt zur?ck in welchem/welchen Quadranten sich das Rechteck befindet
	 * 
	 * @return eine der AREA Konstanten
	 */
	public int getArea() {
		if (ymin >= 0) {
			if (xmin >= 0) {
				return AREA_UPPERRIGHT;
			} else {
				if (xmax <= 0) {
					return AREA_UPPERLEFT;
				} else {
					return AREA_UPPER;
				}
			}

		} else {
			if (ymax <= 0) {
				if (xmin >= 0) {
					return AREA_LOWERRIGHT;
				} else {
					if (xmax <= 0) {
						return AREA_LOWERLEFT;
					} else {
						return AREA_LOWER;
					}
				}

			} else if (xmax <= 0) {
				return AREA_LEFT;
			} else {
				if (xmin >= 0) {
					return AREA_RIGHT;
				} else {
					return AREA_ALL;
				}
			}
		}

	}

	/**
	 * Vergrert die Lngen des Rechteck um das x-fache
	 *  
	 */
	public CoordinateRect2D mult(double x) {
		double x1=-getXRange()/2*(x-1)+xmin;
		double x2=getXRange()/2*(x-1)+xmax;
		double y1=-getYRange()/2*(x-1)+ymin;
		double y2=getYRange()/2*(x-1)+ymax;
		CoordinateRect2D mult=new CoordinateRect2D(x1,x2,y1,y2);
		return mult;
	}

}