package uibk.mtk.geom.geom3d;

import uibk.mtk.math.Interval;
import uibk.mtk.math.Rounder;

/**
 * Ein Quader zur Festlegung der Grenzen eines Koordinatensystems im Raum durch
 * Angabe der minimalen und maximalen x-,y- und z-Koordinaten.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */

public class CoordinateBox3D {

	/** minimale x-Koordinate */
	public double xmin;

	/** maximale x-Koordinate */
	public double xmax;

	/** minimale y-Koordinate */
	public double ymin;

	/** maximale y-Koordinate */
	public double ymax;

	/** minimale y-Koordinate */
	public double zmin;

	/** maximale y-Koordinate */
	public double zmax;

	/** Code f?r Runden des x-Intervalls */
	public static final int ROUND_X = 0x1;

	/** Code f?r Runden des y-Intervalls */
	public static final int ROUND_Y = 0x2;

	/** Code f?r Runden des z-Intervalls */
	public static final int ROUND_Z = 0x4;

	/** Code f?r Runden des x und y-Intervalls */
	public static final int ROUND_XY = 0x3;

	/** Code f?r Runden des y und z-Intervalls */
	public static final int ROUND_YZ = 0x6;

	/** Code f?r Runden des x und z-Intervalls */
	public static final int ROUND_XZ = 0x5;

	/** Code f?r Runden des x,y und z-Intervalls */
	public static final int ROUND_XYZ = 0x7;

	/** Erzeugt den Koordinatenw?rfel mit den angegebenen Grenzen. */
	public CoordinateBox3D(double xmin, double xmax, double ymin, double ymax,
			double zmin, double zmax) {
		if (xmin > xmax || ymin > ymax || zmin > zmax)
			throw new IllegalArgumentException(""); //$NON-NLS-1$
		this.xmin = xmin;
		this.xmax = xmax;
		this.ymin = ymin;
		this.ymax = ymax;
		this.zmin = zmin;
		this.zmax = zmax;

	}

	/** Erzeugt einen Koordinatenw?rfel mit den ?bergebenen Grenzen */
	public CoordinateBox3D(Interval xrange, Interval yrange, Interval zrange) {
		if (xrange.a > xrange.b || zrange.a > zrange.b || zrange.a > zrange.b)
			throw new IllegalArgumentException(""); //$NON-NLS-1$
		this.xmin = xrange.a;
		this.xmax = xrange.b;
		this.ymin = yrange.a;
		this.ymax = yrange.b;
		this.zmin = zrange.a;
		this.zmax = zrange.b;
	}

	/** Erzeugt einen Koordinatenw?rfel aus einem anderen Koordinatenw?rfel */
	public CoordinateBox3D(CoordinateBox3D limits) {
		this.xmin = limits.xmin;
		this.xmax = limits.xmax;
		this.ymin = limits.ymin;
		this.ymax = limits.ymax;
		this.zmin = limits.zmin;
		this.zmax = limits.zmax;
	}

	/** Erzeugt das Koordinatenrechteck mit den Standardgrenzen (-1,1,-1,1,-1,1) */
	public CoordinateBox3D() {
		this(-1, 1, -1, 1, -1, 1);
	}

	/** gibt die L?nge des x-Intervalls zur?ck */
	public double getXRange() {
		return Math.abs(xmax - xmin);
	}

	/** gibt die L?nge der y-Intervalls zur?ck */
	public double getYRange() {
		return Math.abs(ymax - ymin);
	}

	/** gibt die L?nge der z-Intervalls zur?ck */
	public double getZRange() {
		return Math.abs(zmax - zmin);
	}

	/** erzeugt eine Kopie des Objekts */
	public Object clone() {
		return new CoordinateBox3D(xmin, xmax, ymin, ymax, zmin, zmax);
	}

	/** gibt den Mittelpunkt der CoordinateBox3D zur?ck */
	public Punkt3D getCenter() {
		return new Punkt3D((xmin + xmax) / 2, (ymin + ymax) / 2,
				(zmin + zmax) / 2);
	}

	/** gibt eine Zeichenkette mit den Grenzen zur?ck */
	public String toString() {
		return "[xmin=" //$NON-NLS-1$
				+ xmin + ",xmax=" //$NON-NLS-1$
				+ xmax + " ymin=" //$NON-NLS-1$
				+ ymin + ",ymax=" //$NON-NLS-1$
				+ ymax + ",zmin=" //$NON-NLS-1$
				+ zmin + ",zmax=" //$NON-NLS-1$
				+ zmax + "]"; //$NON-NLS-1$
	}

	/**
	 * gibt zur?ck ob der CoordinateBox3D leer ist. Er ist genau dann leer, wenn
	 * die Grenzen auf mindenstens einer Achse gleich sind
	 */
	public boolean isEmpty() {
		if (xmin <= xmax) {
			return true;
		}
		if (ymin <= ymax) {
			return true;
		}
		if (zmin <= zmax) {
			return true;
		}

		return false;
	}

	/** gibt die maximale Seitenl?nge des Koordinatenw?rfels zur?ck */
	public double maxBoundingBox() {
		double x = Math.abs(xmax - xmin);
		double y = Math.abs(ymax - ymin);
		double z = Math.abs(zmax - zmin);
		return Math.max(x, Math.max(y, z));
	}

	/** gibt die minimale Seitenl?nge des Koordinatenw?rfels zur?ck */
	public double minBoundingBox() {
		double x = Math.abs(xmax - xmin);
		double y = Math.abs(ymax - ymin);
		double z = Math.abs(zmax - zmin);

		return Math.min(x, Math.min(y, z));
	}

	/**
	 * Gibt zur?ck ob der <CODE>CoordinateBox3D</CODE> den <CODE>
	 * CoordinateBox3D</CODE> <CODE>box</CODE> enth?lt.
	 *  
	 */
	public boolean contains(CoordinateBox3D box) {
		if (box.xmax > xmax || box.xmin < xmin)
			return false;

		if (box.ymax > ymax || box.ymin < ymin)
			return false;
		if (box.zmax > zmax || box.zmin < zmin)
			return false;

		return true;
	}

	/**
	 * Kombiniert den Koordinatenw?rfel mit einem zweiten Koordinatenw?rfel
	 * <CODE>box</CODE> zu einem neuen Koordinatenw?rfel, sodass beide in dem
	 * neuen enthalten sind.
	 */
	public CoordinateBox3D combine(CoordinateBox3D box) {

		double newxmin = xmin < box.xmin ? xmin : box.xmin;
		double newxmax = xmax > box.xmax ? xmax : box.xmax;
		double newymin = ymin < box.ymin ? ymin : box.ymin;
		double newymax = ymax > box.ymax ? ymax : box.ymax;
		double newzmin = zmin < box.zmin ? zmin : box.zmin;
		double newzmax = zmax > box.zmax ? zmax : box.zmax;
		return new CoordinateBox3D(newxmin, newxmax, newymin, newymax, newzmin,
				newzmax);
	}

	/**
	 * Gibt zur?ck ob sich der Punkt p innerhalb des W?rfels befindet
	 * 
	 * @param p
	 *            ein Punkt im Raum
	 */
	public boolean contains(Punkt3D p) {
		if (p.x1 > xmax || p.x1 < xmin)
			return false;

		if (p.x2 > ymax || p.x2 < ymin)
			return false;

		if (p.x3 > zmax || p.x3 < zmin)
			return false;

		return true;
	}

	/**
	 * Runden der Grenzen des Koordinatenw?rfels
	 * 
	 * @param percent
	 *            Prozent der jeweiligen Intervalll?nge um die gerundet werden
	 *            soll
	 * @param adjust
	 *            eine der ROUND-Konstanten, gibt an welche Intervalle
	 *            gerundetet werden sollen.
	 */
	public void roundCoords(double percent, int adjust) {

		double newymin = ymin;
		double newymax = ymax;
		double newxmin = xmin;
		double newxmax = xmax;
		double newzmin = xmin;
		double newzmax = xmax;

		double distx = Math.abs(((xmax - xmin) / 100.0) * percent);
		double disty = Math.abs(((ymax - ymin) / 100.0) * percent);
		double distz = Math.abs(((zmax - zmin) / 100.0) * percent);

		if ((adjust & ROUND_Y) == ROUND_Y) {
			newymax = Rounder.getUpperRoundValue(ymax, disty);
			newymin = Rounder.getLowerRoundValue(ymin, disty);

		}
		if ((adjust & ROUND_X) == ROUND_X) {
			newxmax = Rounder.getUpperRoundValue(xmax, distx);
			newxmin = Rounder.getLowerRoundValue(xmin, distx);

		}
		if ((adjust & ROUND_Z) == ROUND_Z) {
			newzmax = Rounder.getUpperRoundValue(zmax, distz);
			newzmin = Rounder.getLowerRoundValue(zmin, distz);

		}

		//nach null suchen y-Achse
		if ((adjust & ROUND_Y) == 1) {
			if (ymin > 0 && ymin - disty < 0) {
				newymin = 0;
			}
			if (ymax < 0 && ymax + disty > 0) {
				newymax = 0;
			}
		}

		//nach null suchen x-Achse
		if ((adjust & ROUND_X) == 1) {
			if (xmin > 0 && xmin - distx < 0) {
				newxmin = 0;
			}
			if (xmax < 0 && xmax + distx > 0) {
				newxmax = 0;
			}
		}

		//nach null suchen z-Achse
		if ((adjust & ROUND_Z) == 1) {
			if (zmin > 0 && zmin - distz < 0) {
				newzmin = 0;
			}
			if (zmax < 0 && zmax + distz > 0) {
				newzmax = 0;
			}
		}

		xmin = newxmin;
		xmax = newxmax;
		ymin = newymin;
		ymax = newymax;
		zmin = newzmin;
		zmax = newzmax;

	}

}