package uibk.mtk.geom.util;

import java.util.Vector;

import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.geom3d.CoordinateBox3D;
import uibk.mtk.geom.geom3d.Punkt3D;

/**
 * Utilityklasse f?r geometrische Funktionen.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public final class GeomUtil {

	private GeomUtil() {
	}

	/** Code f?r leeren Schnitt */
	public static final int INTERSECTION_EMPTY = 0;

	/** Code f?r Schnitt ist ein Punkt* */

	public static final int INTERSECTION_POINT = 1;

	/** Code f?r Schnitt ist eine Strecke* */
	public static final int INTESECTION_LINE = 2;

	/***************************************************************************
	 * Schneidet die Gerade p+lambda*dir mit der durch den Anfanspunkt <CODE>u
	 * </CODE> und dem Endpunkt <CODE>v</CODE> gebebenen Strecke.
	 * 
	 * @param p
	 *            Punkt auf der Geraden
	 * @param dir
	 *            Richtung der Geraden
	 * @param u
	 *            Anfangspunkt der Strecke
	 * @param v
	 *            Endpunkt der Strecke
	 * @param s
	 *            Schnittpunkt (Koordinaten werden gesetzt falls Durchschnitt
	 *            ein Punkt ist)
	 * @return eine der INTERSECTION-Konstanten
	 */
	public static int intersectLines(Punkt2D p, Punkt2D dir, Punkt2D u,
			Punkt2D v, Punkt2D s) {

		//Steigung der Geraden AB
		//double slope1 = (B.y - A.y) / (B.x - A.x);
		double slope1 = dir.y / dir.x;
		//Steigung der Strecke CD
		double slope2 = (v.y - u.y) / (v.x - u.x);

		//falls beide vertikal
		if (Double.isInfinite(slope1) && Double.isInfinite(slope2)) {
			if (p.x == u.x) { //Schnitt ist die Strecke CD
				return INTESECTION_LINE;
			} else
				//leerer Schnitt
				return INTERSECTION_EMPTY;
		}

		//falls Gerade vertikal und Steigung der Strecke endlcih
		if (Double.isInfinite(slope1) && !Double.isInfinite(slope2)) {
			double d2 = u.y - slope2 * u.x;
			double sy = slope2 * p.x + d2;

			if ((sy <= u.y && sy >= v.y) || (sy >= u.y && sy <= v.y)) {
				s.x = p.x;
				s.y = sy;
				return INTERSECTION_POINT;
			} else
				return INTERSECTION_EMPTY;

		}

		//falls Strecke vertikal und Steigung der Geraden endlcih
		if (!Double.isInfinite(slope1) && Double.isInfinite(slope2)) {
			double d1 = p.y - slope1 * p.x;
			double sy = slope1 * u.x + d1;

			if ((sy <= u.y && sy >= v.y) || (sy >= u.y && sy <= v.y)) {
				s.x = u.x;
				s.y = sy;
				return INTERSECTION_POINT;
			} else
				return INTERSECTION_EMPTY;

		}

		//falls beide Steigungen endlich
		if (!Double.isInfinite(slope1) && !Double.isInfinite(slope2)) {
			double d1 = p.y - slope1 * p.x;
			double d2 = u.y - slope2 * u.x;

			if (slope1 == slope2) {
				if (d1 == d2) {
					s.setLocation(u);
					return INTESECTION_LINE;
				} else
					return INTERSECTION_EMPTY;
			} else {

				double sx = (d2 - d1) / (slope1 - slope2);
				double sy = slope2 * sx + d2;
				if (((sy <= u.y && sy >= v.y) || (sy >= u.y && sy <= v.y))
						&& ((sx <= u.x && sx >= v.x) || (sx >= u.x && sx <= v.x))) {
					s.x = sx;
					s.y = sy;
					return INTERSECTION_POINT;
				} else
					return INTERSECTION_EMPTY;
			}
		}

		return INTERSECTION_EMPTY;
	}

	/**
	 * Berechnet das <b>QUADRAT </b> des Abstandes eines Punktes von einer
	 * Strecke in der Ebene. Falls der Fu?punkt des Lotes innerhalb der Strecke
	 * liegt wird der Normalabstand zur?ckgegeben, falls nicht der k?rzere
	 * Abstand des Punktes zu einem der Eckpunkte.
	 * 
	 * @param x
	 *            x-Koordinate des Punktes
	 * @param y
	 *            y-Koordinate des Punktes
	 * @param p1x
	 *            x-Koordinate des 1. Punktes der Strecke
	 * @param p1y
	 *            y-Koordinate des 1. Punktes der Strecke
	 * @param p2x
	 *            x-Koordinate des 2. Punktes der Strecke
	 * @param p2y
	 *            y-Koordinate des 2. Punktes der Strecke
	 */
	public static double distPointLine(double x, double y, double p1x,
			double p1y, double p2x, double p2y) {

		//Berechnung des Fu?punktes des Lotes
		double ax = x;
		double ay = y;

		//Orthogonalbasis der "Strecke" berechnen (= normierter
		// Richtungsvektor)
		double lx = p2x - p1x;
		double ly = p2y - p1y;
		double norm = Math.sqrt(lx * lx + ly * ly);
		lx /= norm;
		ly /= norm;

		//Translation
		ax -= p1x;
		ay -= p1y;

		//Koordinate (bzgl. ONB) des Fu?punktes
		double s = ax * lx + ay * ly;

		//Fu?punkt berechnen
		lx *= s;
		ly *= s;
		//Translation
		lx += p1x;
		ly += p1y;

		//?berpr?fen ob Fu?punkt auf der Strecke liegt
		boolean containsx = (lx >= p1x && lx <= p2x)
				|| (lx <= p1x && lx >= p2x);
		boolean containsy = (ly >= p1y && ly <= p2y)
				|| (ly <= p1y && ly >= p2y);

		if (containsx && containsy) {
			lx -= x;
			ly -= y;
			return lx * lx + ly * ly;
		} else { //Fu?punkt des Lotes liegt nicht innerhalb der Strecke

			//geben Abstand zu n?hren Endpunkt der Strecke zur?ck
			//Abstand zu ersten Endpunkt
			double dist1 = (x - p1x) * (x - p1x) + (y - p1y) * (y - p1y);
			//Abstand zu ersten Endpunkt
			double dist2 = (x - p2x) * (x - p2x) + (y - p2y) * (y - p2y);

			return Math.min(dist1, dist2);
		}

	}

	/**
	 * Berechnet die minimalen/maximalen Koordinaten eines Vektors aus <CODE>
	 * Punkt3D</CODE> Objekten
	 *  
	 */
	public static CoordinateBox3D getLimits3D(Vector points) {

		double maximumx = Double.NEGATIVE_INFINITY;
		double maximumy = Double.NEGATIVE_INFINITY;
		double maximumz = Double.NEGATIVE_INFINITY;
		double minimumx = Double.POSITIVE_INFINITY;
		double minimumy = Double.POSITIVE_INFINITY;
		double minimumz = Double.POSITIVE_INFINITY;

		for (int i = 0; i < points.size(); i++) {

			Punkt3D p = (Punkt3D) points.elementAt(i);
			if (p == null)
				continue;

			if (p.x1 > maximumx)
				maximumx = p.x1;
			if (p.x2 > maximumy)
				maximumy = p.x2;
			if (p.x3 > maximumz)
				maximumz = p.x3;
			if (p.x1 < minimumx)
				minimumx = p.x1;
			if (p.x2 < minimumy)
				minimumy = p.x2;
			if (p.x3 < minimumz)
				minimumz = p.x3;

		}

		//Grenzen zur?ckgeben
		return new CoordinateBox3D(minimumx, maximumx, minimumy, maximumy,
				minimumz, maximumz);
	}

	/**
	 * Berechnet die minimalen/maximalen Koordinaten eines Vektors aus <CODE>
	 * Punkt2D</CODE> Objekten
	 */
	public static CoordinateRect2D getLimits2D(Vector points) {

		double maximumx = Double.NEGATIVE_INFINITY;
		double maximumy = Double.NEGATIVE_INFINITY;

		double minimumx = Double.POSITIVE_INFINITY;
		double minimumy = Double.POSITIVE_INFINITY;

		for (int i = 0; i < points.size(); i++) {
			Punkt2D p = (Punkt2D) points.elementAt(i);
			if (p == null)
				continue;
			if (p.x > maximumx)
				maximumx = p.x;
			if (p.y > maximumy)
				maximumy = p.y;
			if (p.x < minimumx)
				minimumx = p.x;
			if (p.y < minimumy)
				minimumy = p.y;

		}
		//Grenzen zur?ckgeben
		return new CoordinateRect2D(minimumx, maximumx, minimumy, maximumy);
	}

}