package uibk.mtk.lang;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.image.BufferedImage;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Vector;

import uibk.mtk.swing.base.MPanel;

/**
 * Basisklasse f?r alle Aktionen der 2D-Applets.
 * 
 * <p>
 * Zu diesem Panel k?nnen Objekte der Klasse <CODE>Drawable</CODE> hinzugef?gt
 * werden. Der Mathpanel (bzw. die von ihm abgeleiten Klassen <CODE>MathPanel2D
 * </CODE> und <CODE>MathPanel3D</CODE>) verwaltet das Zeichnen dieser
 * Objekte, erlaubt Verschieben mit der Maus von Dragable Objekten, zeigt
 * gegebenfalls Tooltips der Objekte (siehe Interface <CODE>Tooltipable</CODE>)
 * an oder hebt Sie hervor (siehe Interface <CODE>Highlightable</CODE>)<br>
 * Der Panel beachtet keine per setBorder() gesetzten Rahmen sondern verwendet,
 * immer seine ganze Fl?che zum Zeichnen des Hintergrundes und der Objekte. (um
 * einen Rahmem zu bekommen -> hinzuf?gen zu einem anderen Panel mit Rahmen)
 * 
 * 
 * </p>
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
abstract public class MathPanel extends MPanel implements ComponentListener {

	/** h?lt fest ob beim n?chsten Zeichnen alle Objekte neu berechnet werden */
	protected boolean repaintrequested = false;

	/** aktuelles Objekt das gezogen wird */
	protected Highlightable highlighted = null;

	/** MausListener f?r das Hervorheben */
	private HighlightListening highlightListener = new HighlightListening();

	/** MausListener f?r das Drag & Drop */
	private MyMouseDragListening mouseListenerDrag = new MyMouseDragListening();

	/** MausListener f?r KLicke */
	private MyMouseClickListening mouseListenerClicks = new MyMouseClickListening();

	/** MausMotionListener f?r das Drag & Drop */
	private MyMouseMotionListening mouseMotionListener = new MyMouseMotionListening();

	/**
	 * gibt an ob Objekte die das <CODE>Highlightable</CODE> implementieren
	 * hervorgehoben werden sollen
	 */
	boolean highlighting_enabled = false;

	/** Hintergrundfarbe */
	private final static Color BACKGROUND = Color.white;

	/** aktuelles Objekt das gezogen wird */
	Dragable dragged = null;

	boolean listenforclicks = false;

	/** ist drag & drop aktiv */
	boolean listenForDrags = false;

	/** Framebuffer */
	protected BufferedImage framebuffer;
	
	/** Erzeugt den MathPanel */
	public MathPanel() {
		this.setBackground(BACKGROUND);
		this.addComponentListener(this);
	}

	/** L?schte das Objekt <CODE>drawable</CODE> aus dem Panel */
	public void delete(Drawable drawable) {
		Iterator it = drawItems.iterator();

		while (it.hasNext()) {
			Object obj = it.next();
			if (obj == drawable) {
				it.remove();
				requestRecompute();
				return;

			}
		}
	}

	/**
	 * L?scht alle sich im MathPanel befindlichen Drawables die von der Klasse
	 * <CODE>classname</CODE> sind oder von dieser Klasse abgeleitet sind.
	 */
	public void delete(Class tclass) {
		ListIterator it = drawItems.listIterator();

		while (it.hasNext()) {
			Object obj = it.next();

			Class c = obj.getClass();
			while (c != null) {
				if (c == tclass)
					it.remove();
				c = c.getSuperclass();

			}

		}

	}

	/**
	 * Gibt in einem <CODE>ListIterator</CODE> alle sich im MathPanel
	 * befindlichen Drawables, die von der Klasse <CODE>classname</CODE> sind
	 * oder von dieser Klasse abgeleitet sind, zur?ck
	 */
	public ListIterator get(Class tclass) {
		ListIterator it = drawItems.listIterator();

		Vector v = new Vector();

		while (it.hasNext()) {
			Object obj = it.next();

			Class c = obj.getClass();
			while (c != null) {
				if (c == tclass)
					v.add(obj);
				c = c.getSuperclass();

			}

		}

		return v.listIterator();
	}

	/** (de)aktivieren des Hervorhebens */
	public void enableHighLighting(boolean enable) {
		if (this.highlighting_enabled == enable) {
			return;
		}

		this.highlighting_enabled = enable;

		if (highlighting_enabled) {

			this.addMouseMotionListener(highlightListener);
		} else {
			this.removeMouseMotionListener(highlightListener);

		}
	}

	/**
	 * Bestimmt explizit, das dass <CODE>Highlightable</CODE> <CODE>high
	 * </CODE> hervorgehoben werden soll. Falls sich das Objekt <CODE>high
	 * </CODE> nicht im Panel befindet, passiert nichts.
	 */
	public void setHighlightable(Highlightable high) {
		this.highlighted = high;
	}

	/**
	 * Gibt das aktuell hervorgehobene <CODE>Highlightable</CODE> zur?ck (kann
	 * <CODE>null</CODE> sein)
	 * 
	 */
	public Highlightable getHighLighted() {
		return highlighted;
	}

	/**
	 * Maus-Ereignisverarbeitung: Hervorheben der Objekte das <CODE>
	 * Highlightable</CODE> implementieren und ?ber denen sich die Maus
	 * befindet
	 */
	protected class HighlightListening extends MouseMotionAdapter {

		public void mouseMoved(MouseEvent ev) {

			Iterator it = drawItems.iterator();
			Highlightable old = highlighted;
			highlighted = null;
			while (it.hasNext()) {
				Object d = it.next();
				if (d instanceof Highlightable
						&& ((Highlightable) d).highlightcontains(ev.getX(), ev
								.getY())) {

					highlighted = (Highlightable) d;
					informHighlightListeners(d);

				}
			}

			if (old != highlighted)
				repaint();

		}
	}

	/**
	 * Kann aufgerufen werden um beim n?chsten Zeichnen ein neu berechnen aller
	 * Objekte zu veranlassen (im Gegensatz zu repaint()) wird nach aufrufen
	 * dieser Funktion nicht gleich neu gezeichnet
	 * 
	 */
	public void requestRecompute() {
		repaintrequested = true;
	}

	/** Neu Berechnen und Zeichnen aller Objekte in diesem Panel */
	public void repaint() {
		repaintrequested = true;
		super.repaint();
	}

	/**
	 * gibt, falls Tooltips aktiviert sind (mit setToolTipText("")) den Tooltip
	 * des Objektes zur?ck ?ber dem sich die Maus befindet
	 */
	public String getToolTipText(MouseEvent event) {

		int ct = drawItems.size();
		for (int i = 0; i < ct; i++) {
			if (drawItems.elementAt(i) instanceof ToolTipable
					&& ((ToolTipable) drawItems.elementAt(i))
							.isToolTipEnabled()
					&& ((ToolTipable) drawItems.elementAt(i)).tooltipcontains(
							event.getX(), event.getY())) {
				return ((ToolTipable) drawItems.elementAt(i))
						.getToolTipText(event);
			}
		}
		return null;

	}

	/** Vektor f?r die Drawable-Objekte */
	protected Vector drawItems = new Vector();

	/** Reporter in dem die Meldungen der Objekte angezeigt werden */
	private Reporter reporter = null;

	/** leere Implementierung */
	public void componentShown(ComponentEvent ev) {
	}

	/** leere Implementierung */
	public void componentHidden(ComponentEvent ev) {
	}

	/** leere Implementierung */
	public void componentMoved(ComponentEvent ev) {
	}

	/** Neu Berechnen und Zeichnen veranlassen */
	public void componentResized(ComponentEvent ev) {

		repaint();
	}

	/** Gibt eine normale Meldung (in schwarz) in das Label aus */
	public void report(String message) {
		reporter.report(message);

	}

	/**
	 * Gibt die Message dieser Fehlermeldung im Label aus
	 * 
	 */
	public void reportError(Exception ex) {

		// f?r leichtere Fehlersuche
		// ex.printStackTrace();

		reportError(ex.getMessage());
	}

	/**
	 * Gibt eine Fehlermeldung (in rot) in das Label aus
	 * 
	 * @param title
	 *            Titel der Fehlermeldung
	 * @param ex
	 *            Exception deren message ausgegben wird
	 */
	public void reportError(String title, Exception ex) {
		// f?r leichte Fehlersuche
		// ex.printStackTrace();
		reportError(title + ": " + ex.getMessage()); //$NON-NLS-1$
	}

	/** Gibt eine Fehlermeldung (in rot) in das Label aus */
	public void reportError(String message) {
		reporter.reportError(message);
	}

	/** Gibt eine Erfolgsmeldung (in gr?n) in das Label aus */
	public void reportSuccess(String message) {
		reporter.reportSuccess(message);

	}

	/** setzt das Label in das die (Fehler)meldungen ausgegeben werden */
	public void setReporter(Reporter rep) {
		reporter = rep;
	}

	// TODO:

	protected Options options;

	public void setOptions(Options options) {
		this.options = options;
	}

	public Options getOptions() {
		return options;
	}

	/**
	 * Hinzuf?gen eines Drawable Objektes zum MathPanel. Falls <CODE>obj
	 * </CODE> nicht vom Typ <CODE>Drawable</CODE> oder <CODE>null</CODE>
	 * ist wird ein <CODE>IllegalArgumentException</CODE> geworfen. Falls es
	 * schon hinzugef?gt worden ist, passiert nichts.
	 * 
	 * 
	 */
	public void add(Object obj) {
		if (obj == null || !(obj instanceof Drawable))
			throw new IllegalArgumentException(""); //$NON-NLS-1$

		if (!drawItems.contains(obj)) {
			Drawable draw = (Drawable) obj;

			drawItems.addElement(draw);
		}
	}

	/** Maus-Ereignisverarbeitung: Klicken */
	protected class MyMouseClickListening extends MouseAdapter {

		public void mousePressed(MouseEvent ev) {
			for (int i = drawItems.size() - 1; i >= 0; i--) {
				if (drawItems.elementAt(i) instanceof Clickable) {
					if (((Clickable) drawItems.elementAt(i)).clickcontains(ev
							.getX(), ev.getY())) {
						((Clickable) drawItems.elementAt(i)).click(ev);
						informClickListeners(drawItems.elementAt(i));
					}
				}
			}
		}
	}

	/**
	 * Vektor mit</CODE> ActionListener</CODE> die benachrichtigt werden,falls
	 * ein Drawable gelickt worden ist
	 */
	Vector clickactions = new Vector();

	/**
	 * Vektor mit</CODE> ActionListener</CODE> die benachrichtigt werden,falls
	 * ein Drawable gedraggt wird
	 */
	Vector dragactions = new Vector();

	/**
	 * Vektor mit</CODE> ActionListener</CODE> die benachrichtigt werden,falls
	 * ein Drawable hervorgehoben wird
	 */
	Vector highlightactions = new Vector();

	/**
	 * f?gt</CODE> ActionListener</CODE> hinzu der benachrichtigt wird, falls
	 * ein Drawable gelickt worden ist (mit actioncommand ="click")
	 */
	public void addClickListener(ActionListener listener) {
		clickactions.add(listener);
	}

	/**
	 * f?gt</CODE> ActionListener</CODE> hinzu der benachrichtigt wird, falls
	 * ein Drawable gezogen wird (mit actioncommand ="drag")
	 */
	public void addDragListener(ActionListener listener) {
		dragactions.add(listener);
	}

	/**
	 * f?gt</CODE> ActionListener</CODE> hinzu der benachrichtigt wird, falls
	 * ein Drawable hervorgehoben wird (mit actioncommand ="drag")
	 */
	public void addHighlightListener(ActionListener listener) {
		highlightactions.add(listener);
	}

	/** Informieren aller ActionListener die auf Klick warten */
	private void informClickListeners(Object obj) {
		Iterator it = clickactions.iterator();
		while (it.hasNext()) {
			ActionEvent ev = new ActionEvent(obj, 0, "click"); //$NON-NLS-1$
			((ActionListener) it.next()).actionPerformed(ev);
		}
	}

	/** Informieren aller ActionListener die auf Drag warten */
	private void informDragListeners(Object obj) {
		Iterator it = dragactions.iterator();
		while (it.hasNext()) {
			ActionEvent ev = new ActionEvent(obj, 0, "drag"); //$NON-NLS-1$
			((ActionListener) it.next()).actionPerformed(ev);
		}
	}

	/** Informieren aller ActionListener die auf Highlight warten */
	private void informHighlightListeners(Object obj) {
		Iterator it = highlightactions.iterator();
		while (it.hasNext()) {
			ActionEvent ev = new ActionEvent(obj, 0, "highlight"); //$NON-NLS-1$
			((ActionListener) it.next()).actionPerformed(ev);
		}
	}

	/** Maus-Ereignisverarbeitung: Drag & Drop der Dragable Objekte */
	protected class MyMouseDragListening extends MouseAdapter {

		public void mousePressed(MouseEvent ev) {
			for (int i = drawItems.size() - 1; i >= 0; i--) {
				if (drawItems.elementAt(i) instanceof Dragable
						&& ((Dragable) drawItems.elementAt(i)).dragcontains(ev
								.getX(), ev.getY())) {
					dragged = (Dragable) drawItems.elementAt(i);
					drawItems.remove(i);
					drawItems.add(dragged);
				}
			}
		}

		public void mouseReleased(MouseEvent ev) {
			dragged = null;
		}

	}

	/** Maus-Ereignisverarbeitung: Drag & Drop der Dragable Objekte */
	protected class MyMouseMotionListening extends MouseMotionAdapter {

		public void mouseDragged(MouseEvent ev) {
			if (dragged == null || !dragged.isDragEnabled()) {
				return;
			}
			dragged.drag(ev);
			informDragListeners(dragged);
			repaint();
		}

	}

	public void setListenForClicks(boolean listen) {
		if (listenforclicks == listen) {
			return;
		}

		this.listenforclicks = listen;
		if (listenforclicks) {
			this.addMouseListener(mouseListenerClicks);

		} else {
			this.removeMouseListener(mouseListenerClicks);

		}
	}

	/** (de)aktiviert das Drag & Drop */
	public void setListenForDragAndDrop(boolean listenForDrags) {
		if (listenForDrags == this.listenForDrags) {
			return;
		}

		this.listenForDrags = listenForDrags;
		if (listenForDrags) {
			this.addMouseListener(mouseListenerDrag);
			this.addMouseMotionListener(mouseMotionListener);
		} else {
			this.removeMouseListener(mouseListenerDrag);
			this.removeMouseMotionListener(mouseMotionListener);
		}
	}

	/** Vorbereiten des BufferdImage (anpassen seiner Gr??e) */
	protected void preparebuffer() {

		int w = getWidth();
		int h = getHeight();

		if (framebuffer != null
				&& (framebuffer.getWidth() != w || framebuffer.getHeight() != h))
			requestRecompute();

		if (framebuffer == null || w > framebuffer.getWidth()
				|| h > framebuffer.getHeight()) {

			framebuffer = (BufferedImage) createImage(w, h);
		}
		
	}

	/** Zeichnen des Panels und der Drawables */
	public synchronized void paint(Graphics g) {

		super.paint(g);
		
		// falls nicht sichbar oder Breite/L?nge<100 nichts zeichnen
		if (!isVisible() || getWidth() < 100 || getHeight() < 100)
			return;

		// Buffer vorbereiten
		preparebuffer();

		Graphics2D g2 = framebuffer.createGraphics();
		
		// Hintergrund zeichnen
		g2.setColor(Color.white);
		g2.fillRect(0, 0, framebuffer.getWidth(), framebuffer.getHeight());
		// Drawables zeichnen
		drawDrawables(framebuffer, g2);
		// auf Schirm zeichnen
		g.drawImage(framebuffer, 0, 0, null);

	}

	/**
	 * R?ckt ein Drawable in den Hintergrund. D.h. das Drawable wird als erstes
	 * gezeichnet.
	 * 
	 * @param drawable
	 *            <CODE>Drawable</CODE> das in den Hintergrund soll
	 */
	public void background(Drawable drawable) {
		Iterator it = drawItems.iterator();

		while (it.hasNext()) {
			Object obj = it.next();
			if (obj == drawable) {
				it.remove();

				drawItems.add(0, drawable);
				requestRecompute();
				return;
			}
		}
	}

	/**
	 * R?ck ein Drawable in den Vordergrund. D.h. das Drawable wird als letztes
	 * gezeichnet.
	 * 
	 * @param drawable
	 *            <CODE>Drawable</CODE> das in den Vordergrund soll
	 */
	public void foreground(Drawable drawable) {
		Iterator it = drawItems.iterator();

		while (it.hasNext()) {
			Object obj = it.next();
			if (obj == drawable) {
				it.remove();

				drawItems.add(drawable);
				requestRecompute();
				return;
			}
		}
	}
	
	public BufferedImage getBufferedImage() {
		return framebuffer;
	}
	
	/** Zeichnen der Drawable Objekte */
	abstract protected void drawDrawables(BufferedImage framebuffer,
			Graphics2D g2);

}