package uibk.mtk.math;

import uibk.mtk.geom.geom2d.Punkt2D;

/**
 * Stellt statische Funktionen zum Berechnen von Maximimum, Minimum und linearer
 * Interpolation zur Verf?gung
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public final class MathUtil {

	private MathUtil() {
	}

	/**
	 * gibt das maximale Element des ?bergebenen Arrays zur?ck
	 * 
	 * @param t
	 *            Array aus double Werten
	 * @return maximales Element des Arrays
	 */
	public static double max(double[] t) {
		double maximum = t[0];
		for (int i = 1; i < t.length; i++) {
			if (t[i] > maximum) {
				maximum = t[i];
			}
		}
		return maximum;
	}

	/**
	 * gibt das minimale Element des ?bergebenen Arrays zur?ck
	 * 
	 * @param t
	 *            Array aus double Werten
	 * @return maximales Element des Arrays
	 */
	public static double min(double[] t) {
		double minimum = t[0];
		for (int i = 1; i < t.length; i++) {
			if (t[i] < minimum) {
				minimum = t[i];
			}
		}
		return minimum;
	}

	/**
	 * Lineare Interpolation der Werte value_a und value_b zwischen a und b an
	 * der Stelle x, wobei x zwischen a und b liegen muss. Falls b-a <=0 wird
	 * value_a zur?ckgegeben
	 */
	public static double interpolate(double a, double b, double valuea,
			double valueb, double x) {
		if (b - a <= 0) {
			return valuea;
		} else {
			double k = (b - x) / (b - a);
			double l = (x - a) / (b - a);
			return k * valuea + l * valueb;
		}
	}

	/**
	 * Berechnet die 1-Norm eines Arrays von double Werten
	 * 
	 * @param t
	 *            ein Feld von double Werten
	 * @return die ein Norm des Feldes/Vektors
	 */
	public static double norm1(double[] t) {
		double norm = 0;
		for (int i = 0; i < t.length; i++) {
			norm += Math.abs(t[i]);
		}
		return norm;
	}

	/**
	 * Lineare Interpolation der Werte value_a und value_b zwischen zwei Punkten
	 * in der Ebene a und b an im Punkt x, wobei x auf der Strecke zwischen a
	 * und b liegen muss. Falls a==bwird value_a zur?ckgegeben
	 */
	public static double interpolate(Punkt2D a, Punkt2D b, double valuea,
			double valueb, Punkt2D p) {
		double lenght = (b.x - a.x) * (b.x - a.x) + (b.y - a.y) * (b.y - a.y);
		if (lenght == 0) {
			return valuea;
		}

		double dist = Math.sqrt(((p.x - a.x) * (p.x - a.x) + (p.y - a.y)
				* (p.y - a.y))
				/ lenght);
		return valuea * (1 - dist) + valueb * dist;
	}

	/**
	 * gibt das Vorzeichen eines double Wertes zur?ck
	 * 
	 * @param x
	 *            ein double Werte
	 * @return das Vorzeichen des Argumentes x
	 */
	public static double sign(double x) {
		if (Double.isNaN(x))
			return Double.NaN;

		if (x < 0) {
			return -1;
		} else {
			if (x > 0) {
				return 1;
			} else {
				return 0;
			}

		}
	}

	/** Berechnung der Norm des Punktes (a,b) ohne ?berlauf */
	public static double hypot(double a, double b) {
		double r;
		if (Math.abs(a) > Math.abs(b)) {
			r = b / a;
			r = Math.abs(a) * Math.sqrt(1 + r * r);
		} else if (b != 0) {
			r = a / b;
			r = Math.abs(b) * Math.sqrt(1 + r * r);
		} else {
			r = 0.0;
		}
		return r;
	}

}