package uibk.mtk.math;

import uibk.mtk.lang.ComputationException;
import uibk.mtk.lang.Messages;
import uibk.mtk.math.functions.Function1D;
import uibk.mtk.math.functions.ParsedFunction2D;
import uibk.mtk.geom.geom2d.Punkt2D;

/**
 * Klasse zum schrittweisen Ausf?hren des Newtonverfahrens f?r Funktionen von
 * R->R
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class NewtonMethod {

	/** akutelle N?herung/Approximation/Startwert */
	double x;

	/** akutelle Nherung/Approximation/Startwert */
	double[] coord = new double[2];

	/** die Funktion, auf die das Newtonverfahren angewendet wird */
	Function1D function;

	/** die Funktion, auf die das Newtonverfahren angewendet wird */
	ParsedFunction2D function2D;

	/**
	 * Typ des Newtonverfahrens: TYPE_SIMPLE: vereinfachtes Newtonverfahren
	 * TYPE_SIMPLE: Standard-Newtonverfahren
	 */
	int type;

	/** Code fr das Standard-Newtonverfahren */
	public static final int TYPE_STANDARD = 0;

	/** Code fr das vereinfachte Newtonverfahren */
	public static final int TYPE_SIMPLE = 1;

	/**
	 * Anfangswert/Startwert (f?r Berechnung der Jacobimatrix beim einfachen
	 * Verfahren)
	 */
	double x0;

	/** aktuelles Inkrement */
	double dx;

	/** aktueller Defekt */
	double defect;

	/** Name des Resourcebundles */
	static final String BUNDLE_NAME = "uibk.mtk.math.messages";//$NON-NLS-1$

	/**
	 * Erzeugt das Verfahren
	 * 
	 * @param function
	 *            Funktion auf die das Verfahren angewendet wird
	 * 
	 */
	public NewtonMethod(Function1D function) {
		setFunction(function);
	}

	/** Setzt die Funktion f?r das, das Verfahren durchgef?ht werden soll */
	public void setFunction(Function1D function) {
		this.function = function;
	}

	/**
	 * Erzeugt das Verfahren
	 * 
	 * @param function
	 *            Funktion auf die das Verfahren angewendet wird
	 * 
	 */
	public NewtonMethod(ParsedFunction2D function) {
		setFunction2D(function);
	}

	/** Setzt die Funktion f?r das, das Verfahren durchgef?ht werden soll */
	public void setFunction2D(ParsedFunction2D function) {
		this.function2D = function;
	}

	/**
	 * Erzeugt das Verfahren
	 */
	public NewtonMethod() {

	}

	/** gibt die aktuelle Nherung zur?ck */
	public double getX() {
		return x;
	}

	/** gibt den Absolutwert des aktuellen Inkrements zur?ck */
	public double getIncrement() {
		return Math.abs(dx);
	}

	/**
	 * gibt den Absolutwert des aktuellen Defekts (Wert der Fkt. an bei der
	 * aktuellen N?herung) zur?ck
	 */
	public double getDefect() {
		return Math.abs(defect);
	}

	/** setzt den Startwert f?r das Verfahren */
	public void setX0(double x) {
		this.x = x;
	}

	/**
	 * setzt Wert bei dem im einfachen Verfahren die Jacobimatrix berechnet
	 * werden soll
	 */
	public void setJacobiPostion(double x0) {
		this.x0 = x0;
	}

	/**
	 * Berechnet einen Schritt des Newtonverfahrens
	 * 
	 * @return n?chste N?herung
	 * @throws Exception
	 *             falls Punkt singul?r, oder ?berlauf
	 */
	public double nextApproximation() throws ComputationException {

		if (function == null)
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.0")); //$NON-NLS-1$

		// Ableitung berechnen
		double deriv;
		try {
			double tempx = type == TYPE_SIMPLE ? x0 : x;
			deriv = function.getDerivation(tempx);
		} catch (Exception ex) {
			throw new ComputationException(ex.getMessage());
		}

		// ist Ableitung 0
		if (deriv == 0)
			throw new ComputationException("f'(" + x //$NON-NLS-1$
					+ Messages.getString(BUNDLE_NAME, "NewtonMethod.2")); //$NON-NLS-1$

		// ?berlauf/NaN
		if (Double.isNaN(deriv) || Double.isInfinite(deriv))
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.3")); //$NON-NLS-1$

		// Defekt ausrechnen
		defect = function.getValue(x);

		// Inkrement ausrechnen
		dx = -defect / deriv;

		// neue N?herung
		x = x + dx;

		// ?berlauf/NaN
		if (Double.isNaN(x) || Double.isInfinite(x))
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.4")); //$NON-NLS-1$

		return x;

	}

	/**
	 * Berechnet einen Schritt des Newtonverfahrens
	 * 
	 * @Param valuevar
	 * @Param var
	 * @return n?chste N?herung
	 * @throws Exception
	 *             falls Punkt singul?r, oder ?berlauf
	 */
	public double nextApproximation(double valuevar, int var)
			throws ComputationException {

		if (function2D == null)
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.0")); //$NON-NLS-1$

		// Ableitung berechnen
		double deriv;
		try {
			Punkt2D tempx = new Punkt2D();
			if (var == 1) {
				tempx.x = var;
				tempx.y = type == TYPE_SIMPLE ? x0 : x;
				coord[0] = var;
				coord[1] = x;
				deriv = function2D.getDerivate2(tempx)[1].y;
			} else {
				tempx.y = var;
				tempx.x = type == TYPE_SIMPLE ? x0 : x;
				coord[1] = var;
				coord[0] = x;
				deriv = function2D.getDerivate2(tempx)[0].x;
			}

		} catch (Exception ex) {
			throw new ComputationException(ex.getMessage());
		}

		// ist Ableitung 0
		if (deriv == 0)
			throw new ComputationException("f'(" + x //$NON-NLS-1$
					+ Messages.getString(BUNDLE_NAME, "NewtonMethod.2")); //$NON-NLS-1$

		// ?berlauf/NaN
		if (Double.isNaN(deriv) || Double.isInfinite(deriv))
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.3")); //$NON-NLS-1$

		// Defekt ausrechnen
		if (var == 1) {
			defect = function2D.getDerivate1(new Punkt2D(coord[0], coord[1])).y;
		} else {
			defect = function2D.getDerivate1(new Punkt2D(coord[0], coord[1])).x;

		}

		// Inkrement ausrechnen
		dx = -defect / deriv;

		// neue N?herung
		x = x + dx;

		// ?berlauf/NaN
		if (Double.isNaN(x) || Double.isInfinite(x))
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.4")); //$NON-NLS-1$

		return x;

	}

	/**
	 * setzt den Typ des Newtonverfahrens
	 * 
	 * @param type
	 *            eine der zwei Konstanten TYPE_STANDARD/TYPE_STANDARD
	 */
	public void setType(int type) {
		if (type != TYPE_STANDARD && type != TYPE_SIMPLE)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.5")); //$NON-NLS-1$
		this.type = type;
	}

	/**
	 * gibt den Typ des Newtonverfahrens zur?ck
	 * 
	 * @return eine der zwei Konstanten TYPE_STANDARD/TYPE_STANDARD
	 */
	public int getType() {
		return type;
	}

	/** Gibt den Fehler zur?ck |f'(x(n-1))*f(x(n))| */
	public double getError() {
		return Math.abs(dx);
	}
}