package uibk.mtk.math;

import uibk.mtk.lang.ComputationException;
import uibk.mtk.lang.Messages;
import uibk.mtk.math.functions.ParsedFunction2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import Jama.Matrix;

/**
 * Klasse zum schrittweisen Ausfahren des Newtonverfahrens fr Funktionen von
 * R^2->R
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2006
 * </p>
 * 
 * @author Rudigier Andreas (csae7164@uibk.ac.at)
 */
public class NewtonMethod2D {

	/** akutelle Nherung/Approximation/Startwert */
	Matrix x = new Matrix(2, 1);

	/** die Funktion, auf die das Newtonverfahren angewendet wird */
	ParsedFunction2D function;

	/**
	 * Typ des Newtonverfahrens: TYPE_SIMPLE: vereinfachtes Newtonverfahren
	 * TYPE_SIMPLE: Standard-Newtonverfahren
	 */
	int type;

	/** Code fr das Standard-Newtonverfahren */
	public static final int TYPE_STANDARD = 0;

	/** Code fr das vereinfachte Newtonverfahren */
	public static final int TYPE_SIMPLE = 1;

	/**
	 * Anfangswert/Startwert (fr Berechnung der Jacobimatrix beim einfachen
	 * Verfahren)
	 */
	Matrix x0 = new Matrix(2, 1);

	/** aktuelles Inkrement */
	Matrix dx = new Matrix(2, 1);

	/** aktueller Defekt */
	Matrix defect = new Matrix(2, 1);

	/** Name des Resourcebundles */
	static final String BUNDLE_NAME = "uibk.mtk.math.messages";//$NON-NLS-1$

	/**
	 * Erzeugt das Verfahren
	 * 
	 * @param function
	 *            Funktion auf die das Verfahren angewendet wird
	 * 
	 */
	public NewtonMethod2D(ParsedFunction2D function) {
		setFunction(function);
	}

	/** Setzt die Funktion f?r das, das Verfahren durchgef?ht werden soll */
	public void setFunction(ParsedFunction2D function) {
		this.function = function;
	}

	/**
	 * Erzeugt das Verfahren
	 */
	public NewtonMethod2D() {

	}

	/** gibt die aktuelle Nherung der x-Kompenente zur?ck */
	public double getX() {
		return x.get(0, 0);
	}

	/** gibt die aktuelle Nherung der y-Kompenente zur?ck */
	public double getY() {
		return x.get(1, 0);
	}

	/** gibt die aktuelle Nherung zur?ck */
	public Matrix getXY() {
		return x;
	}

	/** gibt den Absolutwert des aktuellen Inkrements zur?ck */
	public double getIncrement() {
		return dx.normF();
	}

	/**
	 * gibt den Absolutwert des aktuellen Defekts (Wert der Fkt. an bei der
	 * aktuellen N?herung) zur?ck
	 */
	public double getDefect() {
		return defect.normF();
	}

	/** setzt den Startwert fr das Verfahren der x-Kompenente */
	public void setX0(double x) {
		this.x.set(0, 0, x);
	}

	/** setzt den Startwert fr das Verfahren der y-Kompenente */
	public void setY0(double x) {
		this.x.set(1, 0, x);
	}

	/** setzt den Startwert fr das Verfahren */
	public void setXY0(Matrix x) {
		this.x = x;
	}

	/**
	 * setzt Wert bei dem im einfachen Verfahren die Jacobimatrix berechnet
	 * werden soll
	 */
	public void setJacobiPostion(Matrix x0) {
		this.x0 = x0;
	}

	/**
	 * Berechnet einen Schritt des Newtonverfahrens
	 * 
	 * @return nchste Nherung
	 * @throws Exception
	 *             falls Punkt singulr, oder berlauf
	 */
	public Matrix nextApproximation() throws ComputationException {
		double lambda =0.5;
		Matrix deriviation = new Matrix(2, 2);

		if (function == null)
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.0")); //$NON-NLS-1$

		// Ableitung berechnen
		Punkt2D[] deriv;
		try {
			Matrix tempx = type == TYPE_SIMPLE ? x0 : x;
			Punkt2D punkt = new Punkt2D(tempx.get(0, 0), tempx.get(1, 0));
			deriv = function.getDerivate2(punkt);
			deriviation.set(0, 0, deriv[0].x);
			deriviation.set(0, 1, deriv[1].x);
			deriviation.set(1, 0, deriv[0].y);
			deriviation.set(1, 1, deriv[1].y);
		} catch (Exception ex) {
			throw new ComputationException(ex.getMessage());
		}

		// ist Ableitung 0
		if (deriviation.rank() != 2)
			throw new ComputationException("f'(" + x //$NON-NLS-1$
					+ Messages.getString(BUNDLE_NAME, "NewtonMethod.2")); //$NON-NLS-1$

		// berlauf/NaN
		if (Double.isNaN(deriviation.norm1())
				|| Double.isInfinite(deriviation.norm1()))
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.3")); //$NON-NLS-1$

		// Defekt ausrechnen
		Punkt2D def = function.getDerivate1(new Punkt2D(x.get(0, 0), x
				.get(1, 0)));
		defect.set(0, 0, -def.x);
		defect.set(1, 0, -def.y);

		// Inkrement ausrechnen
		dx = deriviation.solve(defect);

		// neue Nherung
		x = x.plus(dx);
		// Newton-Schritt
		Matrix x_lambda = x.plus(dx.times(lambda));
		// Newton-Schritt

		Matrix deltax_lambda = null;
		// Defekt ausrechnen
		def = function.getDerivate1(new Punkt2D(x_lambda.get(0, 0), x_lambda
				.get(1, 0)));
		defect.set(0, 0, -def.x);
		defect.set(1, 0, -def.y);
		deltax_lambda = (deriviation.solve(defect));
		double g_lambda = deltax_lambda.norm2();
		double g_0 = dx.norm2();

		if (g_lambda > g_0) {
			while (g_lambda > g_0) {
				lambda /= 2;
				x_lambda = x.plus(dx.times(lambda));
				// Newton-Schritt
				// Defekt ausrechnen
				def = function.getDerivate1(new Punkt2D(x_lambda.get(0, 0),
						x_lambda.get(1, 0)));
				defect.set(0, 0, -def.x);
				defect.set(1, 0, -def.y);
				deltax_lambda = (deriviation.solve(defect));
				g_lambda = deltax_lambda.norm2();

				if (lambda < 10e-6)
					throw new ComputationException(Messages.getString(
							NewtonMethod2D.BUNDLE_NAME, "Verfahren.13"));
			}
		} else if (lambda < 1) {
			while (g_lambda < g_0 && lambda <= 2) {
				g_0 = g_lambda;
				lambda = lambda * 2;
				x_lambda = x.plus(dx.times(lambda));

				// Newton-Schritt
				// Defekt ausrechnen
				def = function.getDerivate1(new Punkt2D(x_lambda.get(0, 0),
						x_lambda.get(1, 0)));
				defect.set(0, 0, -def.x);
				defect.set(1, 0, -def.y);
				deltax_lambda = (deriviation.solve(defect));
				g_lambda = deltax_lambda.norm2();

			}
			lambda = Math.min(1, lambda / 2);
		}
		x = x.plus(dx.times(lambda));
		// berlauf/NaN
		if (Double.isNaN(x.normF()) || Double.isInfinite(x.normF()))
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.4")); //$NON-NLS-1$

		return x;

	}

	/**
	 * setzt den Typ des Newtonverfahrens
	 * 
	 * @param type
	 *            eine der zwei Konstanten TYPE_STANDARD/TYPE_STANDARD
	 */
	public void setType(int type) {
		if (type != TYPE_STANDARD && type != TYPE_SIMPLE)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"NewtonMethod.5")); //$NON-NLS-1$
		this.type = type;
	}

	/**
	 * gibt den Typ des Newtonverfahrens zurck
	 * 
	 * @return eine der zwei Konstanten TYPE_STANDARD/TYPE_STANDARD
	 */
	public int getType() {
		return type;
	}

	/** Gibt den Fehler zurck |f'(x(n-1))*f(x(n))| */
	public double getError() {
		return dx.normF();
	}
}