package uibk.mtk.math.functions;

import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.geom2d.Vector2D;
import uibk.mtk.lang.Settings;
import uibk.mtk.parser.realparser.Expr;
import uibk.mtk.parser.realparser.Variable;

/**
 * Klasse die eine Funktion von R->R^2 implementiert und durch Benutzereingaben
 * definiert wird.
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class ParsedFunction2D extends ParsedFunction implements Function2D {

	/** Erzeugt die Funktion */
	public ParsedFunction2D(Variable varx, Variable vary, Expr function) {
		super(new Variable[] { varx, vary }, new Expr[] { function });

	}

	/**
	 * gibt den Wert der Funktion an der Stelle (0,y) zur?ck.
	 */
	public double getValue(double x, double y) {
		double[] v = { x, y };

		return super.getValue(v)[0];
	}

	/**
	 * gibt den Wert der Funktion an der Stelle (p.x,p.y) zur?ck.
	 */
	public double getValue(Punkt2D p) {

		return getValue(p.x, p.y);
	}

	/**
	 * Richtungsabeleitung im Punkt a in Richtung v Pr?fung auf Diffbarkeit wie
	 * in Klasse Derivations
	 * 
	 * @param a
	 *            Punkt an dem die Richtungsableitung berechnet werden soll
	 * @param v
	 *            Richtungsvektor
	 * @return
	 * @throws Exception
	 *             Exception beim Differenzieren
	 */
	public double getDerivate(Punkt2D a, Vector2D v) throws Exception {

		double[] deriv = directionalDerivate(new double[] { a.x, a.y },
				new double[] { v.x, v.y });
		return deriv[0];
	}
	/**
	 * Liefert Gradienten
	 *@param v
	 *            Richtungsvektor
	 * @return
	 */
	public Punkt2D getDerivate1( Punkt2D v){
		double h =gethSymm1((v.x+v.y)/2);
		Punkt2D deriv = new Punkt2D();
		deriv.x=(getValue(v.x+h,v.y)-getValue(v.x-h,v.y))/(2*h);
		deriv.y=(getValue(v.x,v.y-h)-getValue(v.x,v.y+h))/(2*h);
		return deriv;
	}
	/**
	 * gibt die Differenz h zum Berechnen der 1.Ableitung f?r die symmetrische
	 * Differenzenformel zur?ck
	 * 
	 * @param x
	 *            Stelle an der differenziert wird
	 *  
	 */
	private double gethSymm1(double x) {
		//h f?r die symmetrischen Diff. berechnen
		double hsym = Math.pow(Settings.getMachinePrecision(), 1.0 / 3);

		double s = Math.max(1, Math.abs(x));
		return hsym * s;
	}

	/**
	 * gibt die Differenz h zum Berechnen der 2.Ableitung f?r die symmetrische
	 * Differenzenformel zur?ck
	 * 
	 * @param x
	 *            Stelle an der differenziert wird
	 *  
	 */
	private double gethSymm2(double x) {
		//h f?r die symmetrischen Diff. berechnen
		double hsym = Math.pow(Settings.getMachinePrecision(), 1.0 / 4);

		double s = Math.max(1, Math.abs(x));
		return hsym * s;
	}

	/**
	 * Liefert die zweite Ableitung
	 *@param v
	 *           Array der Hesse Matrix
	 * @return
	 */
	public Punkt2D[] getDerivate2(Punkt2D v) {
		double h=gethSymm2((v.x+v.y)/2);
		Punkt2D[] deriv=new Punkt2D[2];
		deriv[0]= new Punkt2D();
		deriv[1]= new Punkt2D();
		deriv[0].x=(getValue(v.x-h,v.y)+getValue(v.x+h,v.y)-2*
				getValue(v.x,v.y))/(h*h);
		deriv[1].y=(getValue(v.x,v.y-h)+getValue(v.x,v.y+h)-2*
				getValue(v.x,v.y))/(h*h);
		deriv[0].y=(getValue(v.x+h,v.y+h)-getValue(v.x-h,v.y+h)
				-getValue(v.x+h,v.y-h)+getValue(v.x-h,v.y-h))/(4*h*h);
		deriv[1].x=deriv[0].y;
		return deriv;
	}
}