package uibk.mtk.math.integration;

import uibk.mtk.math.functions.Function1D;
import uibk.mtk.geom.geom2d.Punkt2D;

/**
 * Klasse zum Integrieren von Funktionen mit der Gau?-Legendre Quadraturformel
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2004
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class GaussIntegration {

	/** Genauigkeit mit der die Knoten und Gewichte berechnet werden */
	private static double EPS = 1.0E-11;

	public GaussIntegration() {
		super();
	}

	/**
	 * Berechnet die Knoten und Gewichte der Gau?-Legendre-Quadraturformel der
	 * Stufe <CODE>s</CODE> auf dem Intervall [a,b]. (vergl. Numerical
	 * Recipies in C)
	 * 
	 * @param a
	 *            untere Grenze des Intervalls
	 * @param b
	 *            obere Grenze des Intervalls
	 * @param knots
	 *            Feld der L?nge <CODE>s</CODE> zum Speichern der Knoten
	 * @param weights
	 *            Feld der L?nge <CODE>s</CODE> zum Speichern der Gewichte
	 * @param s
	 *            Stufenanzahl
	 */
	public void calcKnotsAndWeights(double a, double b, double knots[],
			double weights[], int s) {

		double z1, z, xl, pp, p3, p2, p1;

		// Die Nullstellen sind im Intervall symmetrisch ->
		// brechnen

		// Mitte des Intervalls
		double xm = 0.5 * (b + a);
		int m = (s + 1) / 2;

		for (int i = 1; i <= m; i++) {// Loop over the desired roots.

			// Approximation f?r die i-te Nullstelle
			z = Math.cos(Math.PI * (i - 0.25) / (s + 0.5));

			// Ausgehend von der Approximation: Newtoniteration
			do {
				p1 = 1.0;
				p2 = 0.0;
				for (int j = 1; j <= s; j++) {
					// Schleife ?ber die rekursive Darstellung der Legrende
					// Polynome
					// um den Wert des s-ten Legendre Polynoms an der Stelle z
					// zu berechnen
					p3 = p2;
					p2 = p1;
					p1 = ((2.0 * j - 1.0) * z * p2 - (j - 1.0) * p3) / j;
				}
				// Berechnung der Ableitung des s-ten Legendrepolynoms an der
				// Stelle z
				pp = s * (z * p1 - p2) / (z * z - 1.0);
				z1 = z;

				// Newtonschritt
				z = z1 - p1 / pp;
			} while (Math.abs(z - z1) > EPS);

			// Skalieren der Knoten auf das gew?nschte Intervall
			xl = 0.5 * (b - a);
			knots[i - 1] = xm - xl * z;
			knots[s - i] = xm + xl * z;

			// Berechnung der Gewichte nach der Formel von Christoffel
			// b(i)=1/((1-x(i)^2)*(P(s)(x(i))^2))
			weights[i - 1] = 2.0 * xl / ((1.0 - z * z) * pp * pp);
			// symmetrisches Gewicht
			weights[s - i] = weights[i - 1];
		}
	}

	/**
	 * Numerische Integration der reellen Funktion <CODE>f</CODE> mit der
	 * Gau?-Legendre Quadraturformel der Stufe</CODE> s</CODE> auf dem in
	 * <CODE>n</CODE> ?quidistante Teilintervalle zerlegten Intervall
	 * [lower,upper]
	 * 
	 * @param f
	 *            Funkton von R->R
	 * @param lower
	 *            untere Intervallgrenze
	 * @param upper
	 *            obere Intervallgrenze
	 * @param s
	 *            Stufenanzahl
	 * @param n
	 *            Anzahl der Teilintervalle, in das das Intervall [lower,upper]
	 *            zerlegt wird
	 * 
	 */
	public double quadrature(Function1D f, double lower, double upper, int s,
			int n) {

		// Knoten der Quadraturformel
		double[] knots = new double[s];
		// Gewichte der Quadraturformel
		double[] weights = new double[s];
		// Quadratursumme
		double sum = 0.0;
		// Zwischensumme
		double tempsum = 0.0;

		// Berechnung der Knoten und Gewichte f?r das Intervall [-1,1]
		calcKnotsAndWeights(-1, 1, knots, weights, s);

		// ?quidistante Unterteilung
		double h = (upper - lower) / n;

		for (int j = 0; j < n; j++) {

			// aktuelles Intervall
			double a = lower + j * h;
			double b = lower + (j + 1) * h;
			// Transformation des Intervalls
			double xplus = 0.5 * (a + b);
			double xminus = 0.5 * (b - a);
			// Intervall quadrieren
			tempsum = 0;
			for (int i = 0; i < s; i++) {
				double dx = xminus * knots[i];
				tempsum += weights[i] * f.getValue(xplus + dx);
			}
			// Skalieren
			tempsum *= xminus;
			// zur Gesamtsumme hinzuz?hlen
			sum += tempsum;

		}
		return sum;

	}

	/**
	 * Numerische Integration der reellen Funktion <CODE>f</CODE> mit der
	 * Gau?-Legendre Quadraturformel der Stufe</CODE> s</CODE> auf dem in
	 * <CODE>n</CODE> ?quidistante Teilintervalle zerlegten Intervall
	 * [lower,upper] und als Rckgabe die Punkte des Graphen an den
	 * Unterteilungspunkten und den Knotenpunkten.
	 * 
	 * @param anfang
	 *            Der Anfangswert der Punkte
	 * @param f
	 *            Funkton von R->R
	 * @param lower
	 *            untere Intervallgrenze
	 * @param upper
	 *            obere Intervallgrenze
	 * @param s
	 *            Stufenanzahl
	 * @param n
	 *            Anzahl der Teilintervalle, in das das Intervall [lower,upper]
	 *            zerlegt wird
	 * 
	 */
	public Punkt2D[] quadraturepoints(double anfang,Function1D f, double lower, double upper,
			int s, int n) {

		// Knoten der Quadraturformel
		double[] knots = new double[s];
		// Gewichte der Quadraturformel
		double[] weights = new double[s];
		// Punkte fr den Graphen
		Punkt2D[] punkt = new Punkt2D[1 + n * (s + 1)];
		// Zwischensumme
		double tempsum = 0.0;
		// zwischenwert der Knoten
		double low = lower;

		// Berechnung der Knoten und Gewichte f?r das Intervall [-1,1]
		calcKnotsAndWeights(-1, 1, knots, weights, s);

		// quidistante Unterteilung
		double h = (upper - lower) / n;

		punkt[0] = new Punkt2D();
		punkt[0].x = lower;
		punkt[0].y = anfang;
		for (int i = 0; i < s; i++) {
			if (i != 0) {
				low = lower + knots[i] * h / 2;
				double y = anfang+quadrature(f, lower, low, 6, 1);
				punkt[i] = new Punkt2D(low, y);
				n=i==1?n-1:n;
			}
			for (int j = 0; j < n; j++) {

				// aktuelles Intervall
				double a = low + j * h;
				double b = low + (j + 1) * h;
				// Transformation des Intervalls
				double xplus = 0.5 * (a + b);
				double xminus = 0.5 * (b - a);
				// Intervall quadrieren
				tempsum = 0;
				for (int k = 0; k < s; k++) {
					double dx = xminus * knots[k];
					tempsum += weights[k] * f.getValue(xplus + dx);
				}
				// Skalieren
				tempsum *= xminus;
				// zur Gesamtsumme hinzuz?hlen
				punkt[(j + 1) * (s + 1) + i] = new Punkt2D();
				punkt[(j + 1) * (s + 1) + i].x = b;
				punkt[(j + 1) * (s + 1) + i].y = punkt[(j) * (s + 1) + i].y
						+ tempsum;
			}
		}
		return punkt;
	}

	/**
	 * Numerische Integration der reellen Funktion <CODE>f</CODE> mit der
	 * Gau?-Legendre Quadraturformel der Stufe</CODE> s</CODE> auf dem in
	 * <CODE>n</CODE> ?quidistante Teilintervalle zerlegten Intervall
	 * [lower,upper] und als Rckgabe die Punkte des Graphen an den
	 * Unterteilungspunkten und den Knotenpunkten.
	 * @param anfang
	 *            Der Anfangswert der Punkte
	 *            
	 * @param f
	 *            Punkt2D[] berechnet mit quadraturepoints mit demselben
	 *            Interval,Unterteilung und Stufe.
	 * @param lower
	 *            untere Intervallgrenze
	 * @param upper
	 *            obere Intervallgrenze
	 * @param s
	 *            Stufenanzahl
	 * @param n
	 *            Anzahl der Teilintervalle, in das das Intervall [lower,upper]
	 *            zerlegt wird
	 * 
	 */
	public Punkt2D[] pointsquadrature(double anfang,Punkt2D[] f, double lower, double upper,
			int s, int n) {

		// Knoten der Quadraturformel
		double[] knots = new double[s];
		// Gewichte der Quadraturformel
		double[] weights = new double[s];
		// Punkte fr den Graphen
		Punkt2D[] points = new Punkt2D[1 + n * (s + 1)];
		// Zwischensumme
		double tempsum = 0.0;
		// zwischenwert der Knoten
		double low = lower;

		// Berechnung der Knoten und Gewichte f?r das Intervall [-1,1]
		calcKnotsAndWeights(-1, 1, knots, weights, s);

		// quidistante Unterteilung
		double h = (upper - lower) / n;

		points[0] = new Punkt2D();
		points[0].x = lower;
		points[0].y = anfang;
		for (int i = 0; i < s; i++) {
			if (i != 0) {
				low = lower + knots[i] * h / 2;
				// Kompromiss, da Funktion nicht mehr bekannt
				double y =anfang+(f[i].y+f[0].y)*(f[i].x-f[0].x) / 2;
				points[i] = new Punkt2D(low, y);
				n--;
			}
			for (int j = 0; j < n; j++) {

				// aktuelles Intervall
				double a = low + j * h;
				double b = low + (j + 1) * h;
				// Transformation des Intervalls
				double xminus = 0.5 * (b - a);
				// Intervall quadrieren
				tempsum = 0;
				for (int k = 0; k < s; k++) {
					tempsum += weights[k] * f[j * (s + 1) + i + k + 1].y;
				}
				// Skalieren
				tempsum *= xminus;
				// zur Gesamtsumme hinzuz?hlen
				points[(j + 1) * (s + 1) + i] = new Punkt2D();
				points[(j + 1) * (s + 1) + i].x = b;
				points[(j + 1) * (s + 1) + i].y = points[j * (s + 1) + i].y
						+ tempsum;
			}
		}
		return points;
	}
}