package uibk.mtk.math.integration;

import uibk.mtk.math.functions.Function2D;


/**
 * Klasse zum Integrieren von Funktionen R^2->R 
 * mit der Gau-Legendre Quadraturformel
 * 
 * <p>
 * Copyright: Institut fr Technische Mathematik, Geometrie und Bauinformatik,
 * 2006
 * </p>
 * 
 * @author Rudigier Andreas (csa7164@uibk.ac.at)
 */


public class GaussIntegration2D {

	/** Genauigkeit mit der die Knoten und Gewichte berechnet werden */
	private static double EPS = 1.0E-11;

	public GaussIntegration2D() {
		super();
	}
	/**
	 * Berechnet die Knoten und Gewichte der Gau?-Legendre-Quadraturformel der
	 * Stufe <CODE>s</CODE> auf dem Intervall [a,b]. (vergl. Numerical
	 * Recipies in C)
	 * 
	 * @param a
	 *            untere Grenze des Intervalls
	 * @param b
	 *            obere Grenze des Intervalls
	 * @param knots
	 *            Feld der L?nge <CODE>s</CODE> zum Speichern der Knoten
	 * @param weights
	 *            Feld der L?nge <CODE>s</CODE> zum Speichern der Gewichte
	 * @param s
	 *            Stufenanzahl
	 */
	public void calcKnotsAndWeights(double a, double b, double knots[],
			double weights[], int s) {

		double z1, z, xl, pp, p3, p2, p1;

		//Die Nullstellen sind im Intervall symmetrisch ->
		//brechnen

		//Mitte des Intervalls
		double xm = 0.5 * (b + a);
		int m = (s + 1) / 2;

		for (int i = 1; i <= m; i++) {// Loop over the desired roots.

			//Approximation f?r die i-te Nullstelle
			z = Math.cos(Math.PI * (i - 0.25) / (s + 0.5));

			//Ausgehend von der Approximation: Newtoniteration
			do {
				p1 = 1.0;
				p2 = 0.0;
				for (int j = 1; j <= s; j++) {
					//Schleife ?ber die rekursive Darstellung der Legrende
					// Polynome
					//um den Wert des s-ten Legendre Polynoms an der Stelle z
					//zu berechnen
					p3 = p2;
					p2 = p1;
					p1 = ((2.0 * j - 1.0) * z * p2 - (j - 1.0) * p3) / j;
				}
				//Berechnung der Ableitung des s-ten Legendrepolynoms an der
				// Stelle z
				pp = s * (z * p1 - p2) / (z * z - 1.0);
				z1 = z;

				//Newtonschritt
				z = z1 - p1 / pp;
			} while (Math.abs(z - z1) > EPS);

			//Skalieren der Knoten auf das gew?nschte Intervall
			xl = 0.5 * (b - a);
			knots[i - 1] = xm - xl * z;
			knots[s - i] = xm + xl * z;

			//Berechnung der Gewichte nach der Formel von Christoffel
			//b(i)=1/((1-x(i)^2)*(P(s)(x(i))^2))
			weights[i - 1] = 2.0 * xl / ((1.0 - z * z) * pp * pp);
			//symmetrisches Gewicht
			weights[s - i] = weights[i - 1];
		}
	}

	
	/**
	 * Berechnet die Knoten und Gewichte der Gau-Legendre-Quadraturformel der
	 * Stufe <CODE>s</CODE> auf dem Rechteck [a,b]x[c,d]. (vergl. Numerical
	 * Recipies in C)
	 * 
	 * @param a
	 *            untere Grenze des Rechtecks der x-Achse
	 * @param b
	 *            obere Grenze des Rechtecks der x-Achse
	 * @param c
	 *            untere Grenze des Rechtecks der y-Achse
	 * @param d
	 *            obere Grenze des Rechtecks der y-Achse
	 * @param knots
	 *            Arry der Gre <CODE>sxs</CODE> zum Speichern der Knoten
	 * @param weights
	 *            Array der Gre <CODE>sxs</CODE> zum Speichern der Gewichte
	 * @param s
	 *            Stufenanzahl
	 */
	public void calcKnotsAndWeights(double a, double b,
			double c, double d,double knots[][],
			double weights[][], int s) {
		double[][] xw=new double[2][s];
		double n=(b-a)/(d-c);
		int m= (s+1)/2;

			GaussIntegration integrate = new GaussIntegration();
			integrate.calcKnotsAndWeights(a,b,knots[0],xw[0],s);
			
			//Umskalieren der Knoten von dem Intervall [a,b]
			//auf das Intervall [c,d]	

			for (int i = 0; i < s; i++) {
				knots[1][i]=c+(knots[0][i]-a)/n;
				xw[1][i]=n*xw[0][i];
			}

			//Berechnung der Gewichte durch Multiplitkation
			//von einem Viertel des Rechtecks
			for (int i = 0; i <m; i++) {
				for (int j = 0; j <m; j++) {
					weights[i][j]=xw[0][i]*xw[1][j];
					weights[s-1-i][j]=weights[i][j];
					weights[i][s-1-j]=weights[i][j];
					weights[s-1-i][s-1-j]=weights[i][j];
				}
			}
	}
/**
 * Numerische Integration der reellen-zweidimensionalen Funktion
 *  <CODE>f</CODE> mit der
 * Gau-Legendre Quadraturformel der Stufe</CODE> s</CODE> auf dem in
 * <CODE>n</CODE> quidistante Rechtecke zerlegtem Rechteck
 * [lowerx,upperx]x[lowery,uppery]
 * 
 * @param f
 *            Funkton von R^2->R
 * @param lowerx
 *            untere Intervallgrenze der x-Achse
 * @param upperx
 *            obere Intervallgrenze der x-Achse
 *@param lowery
 *            untere Intervallgrenze der y-Achse
 * @param uppery
 *            obere Intervallgrenze der y-Achse
 * @param s
 *            Stufenanzahl
 * @param n
 *            Anzahl der Rechtecke in x und y-Richtung
 *  
 */
public double quadrature(Function2D f, double lowerx, double upperx,
		 double lowery, double uppery, int s, int n) {
	
	//Knoten der Quadraturformel
	double[][] knots = new double[2][s];
	//Gewichte der Quadraturformel
	double[][] weights = new double[s][s];
	//Quadratursumme
	double sum = 0.0;
	//Zwischensumme
	double tempsum = 0.0;

	//Berechnung der Knoten und Gewichte fr das Intervall [-1,1]
	calcKnotsAndWeights(-1, 1,-1,1 ,knots, weights, s);

	//quidistante Unterteilung
	double hx = (upperx - lowerx) / n;
	double hy = (uppery - lowery) / n;
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {

		//aktuelle Intervalle
		double ax = lowerx + i * hx;
		double bx = lowerx + (i + 1) * hx;
		double ay = lowery + j * hy;
		double by = lowery + (j + 1) * hy;
		//Transformation des Intervalls
		double xplus = 0.5 * (ax + bx);
		double xminus = 0.5 * (bx - ax);
		double yplus = 0.5 * (ay + by);
		double yminus = 0.5 * (by - ay);
		//Intervall quadrieren
		tempsum = 0;
		for (int m = 0;m < s; m++) {
			for (int k = 0; k < s; k++) {
				double dx = xminus * knots[0][m];
				double dy = yminus * knots[1][k];
				tempsum += weights[m][k] * f.getValue(xplus + dx,yplus + dy);
			}
		}
		//Skalieren
		tempsum= tempsum*xminus*yminus;
		//zur Gesamtsumme hinzuzhlen
		sum += tempsum;

	}
}
	
	return sum;
}
}
