
package uibk.mtk.parser.realparser;

import uibk.mtk.lang.Messages;
import uibk.mtk.math.MathUtil;

/**
 * Ein mathematischer Ausdruck aufgebaut aus konstanten Zahlen, Variablen,
 * arithmetischen und relationalen Operatoren und den elementraren Funktionen.
 * Der Ausdruck kann ausgewertet werden um den den Wert des Ausdrucks mit den
 * aktuellen Werten der Variablen zu brechnen
 * 
 * 
 * 
 * <p>
 * Diese Klasse ist dem frei verwendbaren Packet <i>expr </i> von Darius Bacon
 * (siehe http://www.accesscom.com/~darius/) entnommen und wurde ge?ndert.
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public abstract class Expr {

	/** Name des Resourcebundles */
	static final String BUNDLE_NAME = "uibk.mtk.parser.realparser.messages";//$NON-NLS-1$

	/** die Formel f?r den Ausdruck* */
	String formula;

	public String toString() {
		return formula;
	}

	/**
	 * Berechnung des Wertes des Ausdrucks
	 *  
	 */
	public abstract double value();

	/** Binary operator: Addition */
	public static final int ADD = 0;

	/** Bin?rer Operator: Subtration */
	public static final int SUB = 1;

	/** Bin?rer Operator: Multiplikation */
	public static final int MUL = 2;

	/** Bin?rer Operator: Division */
	public static final int DIV = 3;

	/** Bin?rer Operator: Exponentiation */
	public static final int POW = 4;

	/** Bin?rer Operator: arctangent */
	public static final int ATAN2 = 5;

	/** Bin?rer Operator: Maximum */
	public static final int MAX = 6;

	/** Bin?rer Operator: Minimum */
	public static final int MIN = 7;

	/** Bin?rer Operator: weniger */
	public static final int LT = 8;

	/** Bin?rer Operator: weniger oder glecih */
	public static final int LE = 9;

	/** Bin?rer Operator: gleich */
	public static final int EQ = 10;

	/** Bin?rer Operator: ungleich */
	public static final int NE = 11;

	/** Bin?rer Operator: gr??er gleich */
	public static final int GE = 12;

	/** Bin?rer Operator: gr??er */
	public static final int GT = 13;

	/** Bin?rer Operator: logisches und */
	public static final int AND = 14;

	/** Bin?rer Operator: logisches oder */
	public static final int OR = 15;

	/** Un?rer Opertor: Betrag */
	public static final int ABS = 100;

	/** Un?rer Opertor: ArcusCosinus */
	public static final int ACOS = 101;

	/** Un?rer Opertor: ArcusSinus */
	public static final int ASIN = 102;

	/** Un?rer Opertor: ArcusTangens */
	public static final int ATAN = 103;

	/** Un?rer Opertor: ceiling */
	public static final int CEIL = 104;

	/** Un?rer Opertor: Cosinus */
	public static final int COS = 105;

	/** Un?rer Opertor: e hoch x */
	public static final int EXP = 106;

	/** Un?rer Opertor: floor */
	public static final int FLOOR = 107;

	/** Un?rer Opertor: Lograithmus */
	public static final int LOG = 108;

	/** Un?rer Opertor: Negation */
	public static final int NEG = 109;

	/** Un?rer Opertor: Runden */
	public static final int ROUND = 110;

	/** Un?rer Opertor: Sinus */
	public static final int SIN = 111;

	/** Un?rer Opertor: Wurzel */
	public static final int SQRT = 112;

	/** Un?rer Opertor: Tanges */
	public static final int TAN = 113;

	/** Un?rer Opertor: Tanges */
	public static final int FACTORIAL = 114;

	/** Un?rer Opertor: Signum */
	public static final int SIGNUM = 115;

	/**
	 * Erzeugt einen Ausdruck mit einem konstanten Wert
	 * 
	 * @param v
	 *            der konstante Wert des Ausdrucks
	 * 
	 * @return ein Ausdruck des Wert immer gleich <CODE>v</CODE> ist
	 */
	public static Expr makeLiteral(double v) {
		return new LiteralExpr(v);
	}

	/**
	 * Erzeugt einen Ausdruck der einen un?ren Operator auf einen Operanden
	 * anwendet.
	 * 
	 * @param rator
	 *            Code eines un?ren Operators
	 * 
	 * @param rand
	 *            der Operand
	 * 
	 * @return ein Ausdruck der rator(op) bedeutet (z.b. -1)
	 */
	public static Expr makeApp1(int rator, Expr rand) {
		Expr app = new UnaryExpr(rator, rand);
		//      falls der Operand ein konstanter Ausdr?cke ist geben wir einen
		// konstanten Ausdruck zur?ck
		return rand instanceof LiteralExpr ? new LiteralExpr(app.value()) : app;
	}

	/**
	 * Erzeugt einen Ausdruck der einen un?ren Operator auf einen Operanden
	 * anwendet.
	 * 
	 * @param rator
	 *            Code eines un?ren Operators
	 * 
	 * @param rand
	 *            der Operand
	 * 
	 * @return ein Ausdruck der rator(op) bedeutet (z.b. -1)
	 */
	public static Expr makePost(int rator, Expr rand) {
		Expr app = new PostExpr(rator, rand);
		//      falls der Operand ein konstanter Ausdr?cke ist geben wir einen
		// konstanten Ausdruck zur?ck
		return rand instanceof LiteralExpr ? new LiteralExpr(app.value()) : app;
	}

	/**
	 * Erzeugt einen Ausdruck der einen bin?ren Operator auf zwei Operanden
	 * anwendet.
	 * 
	 * @param rator
	 *            Code eines un?ren Operators
	 * 
	 * @param rand0
	 *            der linke Operand
	 * 
	 * @param rand1
	 *            der rechte Operand
	 * 
	 * @return ein Ausdruck der rator(rand0,rand1) bedeutet (z.b. +(1,4)=1+4)
	 */
	public static Expr makeApp2(int rator, Expr rand0, Expr rand1) {
		Expr app = new BinaryExpr(rator, rand0, rand1);
		//falls beide Operanden konstante Ausdr?cke sind geben wir einen
		// konstanten Ausdruck zur?ck
		return rand0 instanceof LiteralExpr && rand1 instanceof LiteralExpr ? new LiteralExpr(
				app.value())
				: app;
	}

	/**
	 * Erzeugt einen bedingten Ausdruck
	 * 
	 * @param test
	 *            `if' Teil
	 * @param consequent
	 *            `then' Teil
	 * @param alternative
	 *            `else' Teil
	 * @return ein Ausdruck der bedeutet `if test, then consequent, else
	 *         alternative'
	 */
	public static Expr makeIfThenElse(Expr test, Expr consequent,
			Expr alternative) {
		Expr cond = new ConditionalExpr(test, consequent, alternative);
		//falls alle drei Argumente konstanten sind geben wir einen konstanten
		//Ausdruck zur?ck
		if (test instanceof LiteralExpr && consequent instanceof LiteralExpr
				&& alternative instanceof LiteralExpr)
			return new LiteralExpr(cond.value());
		else
			return cond;
	}
}

/**
 * Klasse f?r einen konstanten Ausdruck
 */

class LiteralExpr extends Expr {

	/** der Wert des Ausdrucks */
	double v;

	/**
	 * Erzeugt einen konstanten Ausdruck
	 * 
	 * @param v
	 *            der Wert des Ausdrucks
	 */
	LiteralExpr(double v) {
		this.v = v;
	}

	/** Gibt den Wert des Ausdrucks zur?ck */
	public double value() {
		return v;
	}
}

/**
 * Klasse f?r einen un?ren Ausdruck
 *  
 */

class PostExpr extends Expr {
	/** Code f?r einen Operator */
	int rator;

	/** der Operand */
	Expr rand;

	/**
	 * Erzeugt einen bin?ren Ausdruck
	 * 
	 * @param rator
	 *            Code f?r einen Operator
	 * @param rand
	 *            der Operand
	 */
	PostExpr(int rator, Expr rand) {
		this.rator = rator;
		this.rand = rand;
	}

	/** Gibt den Wert des un?ren Ausdrucks zur?ck */
	public double value() {
		double arg = rand.value();
		switch (rator) {
		default:
		case FACTORIAL:
			return Factorial.factorial(arg);
		}

	}
}

/**
 * Klasse f?r einen un?ren Ausdruck
 *  
 */

class UnaryExpr extends Expr {

	/** Code f?r einen Operator */
	int rator;

	/** der Operand */
	Expr rand;

	/**
	 * Erzeugt einen bin?ren Ausdruck
	 * 
	 * @param rator
	 *            Code f?r einen Operator
	 * @param rand
	 *            der Operand
	 */
	UnaryExpr(int rator, Expr rand) {
		this.rator = rator;
		this.rand = rand;
	}

	/** Gibt den Wert des un?ren Ausdrucks zur?ck */
	public double value() {
		double arg = rand.value();
		switch (rator) {
		case ABS:
			return Math.abs(arg);
		case ACOS:
			return Math.acos(arg);
		case ASIN:
			return Math.asin(arg);
		case ATAN:
			return Math.atan(arg);
		case CEIL:
			return Math.ceil(arg);
		case COS:
			return Math.cos(arg);
		case EXP:
			return Math.exp(arg);
		case FLOOR:
			return Math.floor(arg);
		case LOG:
			return Math.log(arg);
		case NEG:
			return -arg;
		case ROUND:
			return Math.round(arg);
		case SIN:
			return Math.sin(arg);
		case SQRT:
			return Math.sqrt(arg);
		case TAN:
			return Math.tan(arg);
		case FACTORIAL:
			return Factorial.factorial(arg);
		case SIGNUM:
			return MathUtil.sign(arg);

		default:
			throw new RuntimeException(Messages.getString(Expr.BUNDLE_NAME,
					"Expr.0")); //$NON-NLS-1$
		}
	}
}

/**
 * Klasse f?r einen bin?ren Ausdruck
 *  
 */

class BinaryExpr extends Expr {

	/** Code f?r einen Operator */
	int rator;

	/** linker und rechter Operand */
	Expr rand0, rand1;

	/**
	 * Erzeugt einen bin?ren Ausdruck
	 * 
	 * @param rator
	 *            Code f?r einen Operator
	 * @param rand0
	 *            linker Operand
	 * @param rand1
	 *            rechter Operand
	 */
	BinaryExpr(int rator, Expr rand0, Expr rand1) {
		this.rator = rator;
		this.rand0 = rand0;
		this.rand1 = rand1;
	}

	/** gibt den Wert des bin?ren Ausdrucks zur?ck */
	public double value() {
		double arg0 = rand0.value();
		double arg1 = rand1.value();
		switch (rator) {
		case ADD:
			return arg0 + arg1;
		case SUB:
			return arg0 - arg1;
		case MUL:
			return arg0 * arg1;
		case DIV:
			return arg0 / arg1;
		case POW:
			return Math.pow(arg0, arg1);
		case ATAN2:
			return Math.atan2(arg0, arg1);
		case MAX:
			return arg0 < arg1 ? arg1 : arg0;
		case MIN:
			return arg0 < arg1 ? arg0 : arg1;
		case LT:
			return arg0 < arg1 ? 1.0 : 0.0;
		case LE:
			return arg0 <= arg1 ? 1.0 : 0.0;
		case EQ:
			return arg0 == arg1 ? 1.0 : 0.0;
		case NE:
			return arg0 != arg1 ? 1.0 : 0.0;
		case GE:
			return arg0 >= arg1 ? 1.0 : 0.0;
		case GT:
			return arg0 > arg1 ? 1.0 : 0.0;
		case AND:
			return arg0 != 0 && arg1 != 0 ? 1.0 : 0.0;
		case OR:
			return arg0 != 0 || arg1 != 0 ? 1.0 : 0.0;
		default:
			throw new RuntimeException(Messages.getString(Expr.BUNDLE_NAME,
					"Expr.1")); //$NON-NLS-1$
		}
	}
}

/**
 * Klasse f?r einen bedingten Ausdruck
 *  
 */

class ConditionalExpr extends Expr {

	/** Testausdruck */
	Expr test;

	/** Ausdruck falls test wahr ist (d.h. test.value()!=0) */
	Expr consequent;

	/** Ausdruck falls test falsch ist (d.h. test.value()==0) */
	Expr alternative;

	/**
	 * Erzeugt einen bedingten Ausdruck
	 * 
	 * @param test
	 *            die Bedingung
	 * @param consequent
	 *            der Wert falls die Bedingung ungleich 0 ist
	 * @param alternative
	 *            die Alternative
	 */
	ConditionalExpr(Expr test, Expr consequent, Expr alternative) {
		this.test = test;
		this.consequent = consequent;
		this.alternative = alternative;
	}

	/** Gibt den Wert des bedingten Ausdrucks zur?ck */
	public double value() {
		return test.value() != 0 ? consequent.value() : alternative.value();
	}

}