package uibk.mtk.swing.appletbase;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Toolkit;

import javax.swing.BorderFactory;
import javax.swing.JApplet;
import javax.swing.JComponent;
import javax.swing.JSplitPane;
import javax.swing.border.CompoundBorder;

import uibk.mtk.lang.MathPanel;
import uibk.mtk.lang.Messages;
import uibk.mtk.lang.Options;
import uibk.mtk.lang.StandardOptions;
import uibk.mtk.swing.base.MPanel;

/**
 * Basisklasse f?r die Applets.
 * 
 * Diese Basisklasse realisiert den Grundaufbau der Applets. Rechts befindet
 * sich ein ein Panel (<CODE>panelcontrol</CODE>) der den Panel zur
 * Steuerung des Applets aufnimmt. Links ein Panel (<CODE>panelmain</CODE>)
 * der meist einen <CODE>MathPanel2D</CODE> oder <CODE>MathPanel3D</CODE>
 * aufnimmt. Beide befinden sich in einem JSlitPane, womit z.B. die Gr??e des
 * Kontrollpanels ge?ndert bzw. er ganz ausgeblendet werden kann. Unten befindet
 * sich ein Label f?r die Ausgaben des <CODE>MathPanel2D</CODE> bzw. <CODE>
 * MathPanel3D</CODE>
 * 
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
//TODO: schnittstellen doku + besser
public class AppletBase extends JApplet {

	/** wird das Applet als eingenst?ndige Applikation ausgef?hrt */
	protected boolean isStandalone = false;

	/** Gr??e des Dividers des JSplitPane */
	private static final int DIVIDERSIZE = 8;

	/** Label f?r die Ausgaben */
	protected ReportingLabel labelreports = new ReportingLabel(Messages
			.getString(BUNDLE_NAME, "AppletBase.1")); //$NON-NLS-1$

	/** JSplitPane zum Teilen des Panels */
	protected JSplitPane splitpanel;

	/** Panel mit */
	protected NavigationBar panelnav;

	/** Panel f?r die Hauptkomponente (immer ein MathPanel) */
	private MPanel panelmain = new MPanel();

	/** Panel f?r die Kontrollkomponente */
	private MPanel panelcontrol = new MPanel();

	/** Basis-Panel der alle anderen Komponenten enth?lt */
	MPanel base;

	/** Optionen f?r das Applet */
	Options options;

	static private float FRAME_SCALE_FACTOR = 0.75f;

	static private int FRAME_X_MIN = 650;

	static private int FRAME_Y_MIN = 550;

	/** Name des Resourcebundles */
	private static final String BUNDLE_NAME = "uibk.mtk.swing.appletbase.messages";//$NON-NLS-1$

	/** Erzeugt das Applet mit den Standardoptionen */
	public AppletBase() {
		options = new StandardOptions();
	}

	/** Erzeugt das Applet mit den ?bergebenen Optionen */
	public AppletBase(Options opt) {
		if (opt == null)
			options = new StandardOptions();
		else
			options = opt;
	}

	/** Initialisieren des Applets */
	public void init() {
		//Look and Feel setzen
		options.setLookAndFeel();
		initComponents();
		setOneTouchExpandable(true);
	}

	/** Initialisieren der Komponenten */
	protected void initComponents() {

		//Basispanel zu dem alle weiteren Elemente hinzugef?gt werden
		base = new MPanel();
		base.setBackground(options.getBackgroundColor());
		base.setBorder(new CompoundBorder(BorderFactory
				.createLineBorder(Color.black), BorderFactory
				.createEmptyBorder(5, 5, 5, 5)));

		//Ausgabelabel
		labelreports.setBorder(BorderFactory.createLoweredBevelBorder());
		Font font = labelreports.getFont();
		labelreports.setFont(font.deriveFont(Font.BOLD));

		//Navigationsleiste
		panelnav = new NavigationBar(this);
		panelnav.setReportingLabel(labelreports);
		if (isStandalone) {
			panelnav.enableInternetButtons(false);
		}

		//Panel f?r die Hauptkomponente
		panelmain = new MPanel();
		panelmain.setLayout(new BorderLayout(0, 0));
		panelmain.setBorder(BorderFactory.createLineBorder(Color.black));
		panelmain.setBackground(options.getBackgroundColor());

		//Panel f?r Kontrollkomponente
		panelcontrol = new MPanel();
		panelcontrol.setLayout(new BorderLayout(0, 0));
		panelcontrol.setBackground(options.getBackgroundColor());
		panelcontrol.setPreferredSize(new Dimension(140, 1000));

		//Splitpane zu der die Haupt und der Kontrolpanel hinzugef?gt werden
		splitpanel = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, panelmain,
				panelcontrol);
		splitpanel.setResizeWeight(0.8);
		splitpanel.setDividerSize(DIVIDERSIZE);
		splitpanel.setContinuousLayout(false);
		splitpanel.setBackground(options.getBackgroundColor());
		splitpanel.setBorder(BorderFactory.createEmptyBorder());

		//zusammensetzen der Komponenten
		base.setLayout(new BorderLayout(5, 5));
		base.add(panelnav, BorderLayout.NORTH);
		base.add(splitpanel, BorderLayout.CENTER);
		base.add(labelreports, BorderLayout.SOUTH);

		getContentPane().add(base, BorderLayout.CENTER);

	}

	/**
	 * Setzt fest ob der JSplitPane mit einem Klick ge?ffnet/geschlossen werden
	 * kann (vergleiche {@link JSplitPane#setOneTouchExpandable})
	 */
	public void setOneTouchExpandable(boolean enable) {
		splitpanel.setOneTouchExpandable(enable);

	}

	/** Setzt die Gr??e des "Trenners" des SplitPanels */
	public void setDividerSize(int size) {
		splitpanel.setDividerSize(size);
	}

	/**
	 * Setzt das Gewicht zur Bestimmung der Position des "Trenners" im
	 * SplitPanel
	 */
	public void setResizeWeight(double w) {
		splitpanel.setResizeWeight(w);
	}

	/** Gibt Applet-Information zur?ck */
	public String getAppletInfo() {
		return Messages.getString(BUNDLE_NAME, "AppletBase.0"); //$NON-NLS-1$
	}

	/** setzt die Komponente im Hauptpanel */
	public void setMainPanel(JComponent panel) {

		panelmain.removeAll();
		panelmain.add(panel);
		setInfosMathPanel(panel);

	}

	/**
	 * durchsucht die Komponente <CODE>component</CODE> rekursiv nach <CODE>
	 * MathPanel</CODE> und setzt die Optionen und den Reporter f?r die Ausgabe
	 *  
	 */
	private void setInfosMathPanel(JComponent component) {
		if (component instanceof MathPanel) {
			MathPanel mp = ((MathPanel) component);
			mp.setReporter(labelreports);
			mp.setOptions(this.options);
		}

		Component comp[] = component.getComponents();
		for (int i = 0; i < comp.length; i++) {
			if (comp[i] instanceof JComponent)
				setInfosMathPanel((JComponent) comp[i]);
		}
	}

	/** setzt die Komponente im Kontrollpanel */
	protected void setControlPanel(JComponent panel) {
		panelcontrol.removeAll();
		panelcontrol.add(panel);
	}

	/** setzt die Position des "Trenners" des JSplitPane */
	public void setDividerLocation(int l) {
		splitpanel.setDividerLocation(l);

	}

	/** setzt die Position des "Trenners" des JSplitPane */
	public void setDividerLocation(double d) {
		splitpanel.setDividerLocation(d);

	}
	
	/** liefert die Optionen des Pendels */
	public Options getOptions() {
		return options;
	}

	/**
	 * zum Ausf?hren des Applets als eigenst?ndige Anwendung
	 * 
	 * @param applet
	 *            das Applet das ausgef?hrt werden soll
	 * @param title
	 *            Fenstertitel
	 */
	public static void runmain(AppletBase applet, String title, float facx) {
		applet.isStandalone = true;

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		float ratio = 1.0f * screenSize.height / screenSize.width;
		int w = (int) (screenSize.width * facx);
		int h = (int) (w * ratio);

		if (w < FRAME_X_MIN || h < FRAME_Y_MIN) {
			w = FRAME_X_MIN;
			h = FRAME_Y_MIN;
		}

		AppletFrame frame = new AppletFrame(applet, w, h, w, h);
		frame.setTitle(title);
	}

	/**
	 * zum Ausf?hren des Applets als eigenst?ndige Anwendung
	 * 
	 * @param applet
	 *            das Applet das ausgef?hrt werden soll
	 * @param title
	 *            Fenstertitel
	 */
	public static void runmain(AppletBase applet, String title, float facx,
			float facy) {

		//Zentrieren des Fensters in der Mitte des Bildschirms
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

		applet.isStandalone = true;
		int w = (int) (screenSize.width * facx);
		int h = (int) (screenSize.height * facy);

		AppletFrame frame = new AppletFrame(applet, w, h, w, h);
		frame.setTitle(title);
	}

	/**
	 * zum Ausf?hren des Applets als eigenst?ndige Anwendung
	 * 
	 * @param applet
	 *            das Applet das ausgef?hrt werden soll
	 * @param title
	 *            Fenstertitel
	 */
	public static void runmain(AppletBase applet, String title) {

		runmain(applet, title, FRAME_SCALE_FACTOR);

	}
	
	/** Zum aktivieren/deaktivieren der DoubleBuffered des MainPanel */
	public void DoubleBufferedMainPanel(boolean b)
	{
		base.setDoubleBuffered(b);
	}
	

	/**
	 * zum Ausf?hren des Applets als eigenst?ndige Anwendung
	 * 
	 * @param applet
	 *            das Applet das ausgef?hrt werden soll
	 * @param title
	 *            Fenstertitel
	 * @param width
	 *            Breite des Fensters
	 * @param height
	 *            H?he des Fensters
	 */
	/*
	 * public static void runmain(AppletBase applet, String title, int width,
	 * int height) {
	 * 
	 * applet.isStandalone = true;
	 * 
	 * AppletFrame frame = new AppletFrame(applet, width, height, width,
	 * height); frame.setTitle(title); }
	 */

}