package uibk.mtk.util;

import uibk.mtk.geom.geom2d.Punkt2D;

/**
 * Utilityklasse die eine Funktion zum Erzeugen von Gitterpunkten in einem
 * Rechteck bereitstellt
 * 
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public final class Grid2D {

	private Grid2D() {
	}

	/**
	 * Erzeugt ein Gitter mit einem Array <CODE>grid</CODE>.
	 * Wobei grid[0][] das Feld mit den x-Werten enthlt,
	 * und grid[1][] das Feld mit den y-Werten enthlt.
	 * 
	 * @param grid
	 *            grid[0][] X-Werte des Grids 
	 *            grid[1][] Y-Werte des Grids 
	 * @return Gitter der Gre grid[0].length x grid[1].length
	 */
	public static Punkt2D[][] createGrid2D(Punkt2D[][] grid) {
		Punkt2D[][] gridpoints = new Punkt2D[grid[0].length][grid[1].length];
		for (int i = 0; i < grid[1].length; i++) {
			for (int j = 0; j < grid[0].length; j++) {
				gridpoints[j][i]=new Punkt2D();
				gridpoints[j][i].x=grid[0][j].x;
				gridpoints[j][i].y=grid[1][i].y;
			}
		}
		return gridpoints;
	}

	/**
	 * Erzeugt ein Gitter mit <CODE>gridx</CODE> x <CODE>gridy</CODE>
	 * Punkten und gibt sie in einem zweidimensionalen Array zur?ck. Die Punkte
	 * werden "spaltenweise" beginnend mit (xmin,ymin) angeordnet zur?ckgegeben.
	 * Falls xmax - xmin <= 0, ymax - ymin <= 0,gridx <2 oder gridy <2 wird eine
	 * <CODE>IllegalArgumentException</CODE> geworfen.
	 * 
	 * @param xmin
	 *            minimale x-Koordinate des Gitters
	 * @param xmax
	 *            maximale x-Koordinate des Gitters
	 * @param ymin
	 *            minimale y-Koordinate des Gitters
	 * @param ymax
	 *            maximale y-Koordinate des Gitters
	 * @param gridx
	 *            Anzahl der Gitterpunkte "parallel x-Achse"
	 * @param gridy
	 *            Anzahl der Gitterpunkte "parallel y-Achse"
	 * @return Gitter der Gr??e gridx x gridy
	 */
	public static Punkt2D[][] createGrid2D(double xmin, double xmax,
			double ymin, double ymax, int gridx, int gridy) {

		if (xmax - xmin <= 0 || ymax - ymin <= 0 || gridx < 2 || gridy < 2)
			throw new IllegalArgumentException();

		Punkt2D[][] gridpoints = new Punkt2D[gridx][gridy];

		IntervalIterator xIt = new IntervalIterator(xmin, xmax, gridx);
		IntervalIterator yIt = new IntervalIterator(ymin, ymax, gridy);

		double x, y;

		while (xIt.hasNext()) {
			x = xIt.nextdouble();

			while (yIt.hasNext()) {
				y = yIt.nextdouble();

				gridpoints[xIt.getPos() - 1][yIt.getPos() - 1] = new Punkt2D(x,
						y);

			}
			yIt.reset();

		}

		return gridpoints;
	}
}